
//
//  ram.cpp - Memory Module
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//      This is the memory management module header.
//
//  Created - 25 Jul, 2001 - James Leiterman
//
//  Modifications
//
//
//  stab = 4
//


#include    "KariType.h"                                // Master Definitions

    //  3rd Party Libraries

#include    <stdlib.h>
#include    <string.h>
#include    <assert.h>

#if defined( CC_VMP_MAC ) || defined( CC_VMP_WIN32 )
#include    <memory.h>
#endif

    //  Book Development Library Modules

#include    "ram.h"                                     // Memory Module
#include    "CpuAsm.h"



typedef struct
{
    int     Tally;                                      // Current # of bytes allocated
    int     Highest;                                    // Highest amount of memory ever allocated
} RamData;


static RamData ram;



//  Memory Header
//
//  Inserted at the beginning of every block and used for the purpose of tracking its size.

typedef struct RamHeadType
{
    uint32  nReqSize;                                   // Requested size
    uint32  extra[3];                                   // necessary for align16
} RamHead;




//
//  Ram Setup
//

void ramSetup( const CpuInfo * const pInfo )
{
    ASSERT_PTR4( pInfo );

    ramZero(&ram, sizeof(ram));
}



//
//  Allocate Memory (malloc wrapper)
//

void * ramAlloc( uint nReqSize )
{
    void *pMem;
    RamHead *pHead;
    uint nSize;

    ASSERT_ZERO(nReqSize);

        //  Force to 16byte block add room for pre-appended header

    nSize = ALIGN16( nReqSize ) + sizeof( RamHead );

    pHead = (RamHead *) malloc( nSize );
    pMem = pHead;

    if ( NULL == pHead )
      {
        //  Allocation Error
      }
    else
      {
        ram.Tally += (int)nSize;
        if(ram.Tally > ram.Highest)
          {
            ram.Highest = ram.Tally;
          }

        pHead->nReqSize = nReqSize + sizeof(RamHead);   // Save Requested Size
        pHead->extra[0] = 1;
        pHead->extra[1] = 2;
        pHead->extra[2] = 3;

           // Align by adjusting header to +4..+16 bytes

        pMem = (void *) ALIGN16( ((uint)pMem) + sizeof(uint32) );
      }

    return pMem;
}



//
//  Release memory (free wrapper)
//

void ramFree( const void * const pRaw )
{
    uint32 *pMem;

    ASSERT_PTR4( pRaw );

    pMem = (uint32 *)pRaw;
    if ( *(--pMem) < sizeof(RamHead) )
      {
        pMem -= *pMem;
      }

    ram.Tally -= ALIGN16( ((RamHead*)pMem)->nReqSize );

        //  pMem original (unadjusted) pointer
    free( pMem );
}



//
//  Allocate memory
//
//  A pointer is passed as 'ppMem' and set with allocation.
//
//  FUTURE: When enhanced, a pointer can own a block of memory and can be
//      part of a memory tracking/compaction system!
//

bool ramGet( byte ** const ppMem, uint nReqSize )
{
    ASSERT_PTR4( ppMem );

    *ppMem = (byte *) ramAlloc( nReqSize );
    return ( NULL != *ppMem ) ? true : false;
}



//
//  Allocate memory (cleared)
//

bool ramGetClr( byte ** const ppMem, uint nReqSize )
{
    bool ret;

    ASSERT_PTR4( ppMem );

    ret = false;

    *ppMem = (byte *)ramAlloc( nReqSize );
    if ( NULL != *ppMem )
      {
        ramZero( *ppMem, nReqSize );
        ret = true;
      }

    return ret;
}



//
//  Free memory
//
//  The memory associated with the pointer is passed released, and then the
//  pointer is set to NULL!

void ramRelease( byte ** const ppMem )
{
    ASSERT_PTR4( ppMem );

    ramFree( *ppMem );
    *ppMem = NULL;
}



                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   
