
//
//  Bench.cpp - Packed Rotation
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 2 Jun 98 - James Leiterman
//
//  Last Modification
//
//      19 Dec 2001 - JCL - Brought up to date.
//
//  stab=4
//



#include "KariType.h"                                   // Master Definitions

    //  3rd Party Libraries
    
#include <iostream.h>
#include <assert.h>

#include <conio.h>
#include <stdio.h>
#include <malloc.h>
#include <string.h>


    //  Book Development Library Modules
    
#include "CpuAsm.h"                                     // CPU Module
#include "vmp3D.h"
#include "PRot.h"
#include "PRotEx.h"



typedef enum
{
    iTEST_PSRAB,            // sra 8bit  count
    iTEST_PSRAH,            // sra 16bit count
    iTEST_PSRAW,            // sra 32bit count

    iTEST_PSRA1B,           // sra 8bit  1
    iTEST_PSRA1H,           // sra 16bit 1
    iTEST_PSRA1W,           // sra 32bit 1

    iTEST_PSLLB,            // sll 8bit  count
    iTEST_PSLLH,            // sll 16bit count
    iTEST_PSLLW,            // sll 32bit count

    iTEST_PSLL1B,           // sll 8bit  1
    iTEST_PSLL1H,           // sll 16bit 1
    iTEST_PSLL1W,           // sll 32bit 1

    iTEST_PSRLB,            // srl 8bit  count
    iTEST_PSRLH,            // srl 16bit count
    iTEST_PSRLW,            // srl 32bit count

    iTEST_PSRL1B,           // srl 8bit  1
    iTEST_PSRL1H,           // srl 16bit 1
    iTEST_PSRL1W,           // srl 32bit 1

    iTEST_PROLB,            // rol 8bit  count
    iTEST_PROLH,            // rol 16bit count
    iTEST_PROLW,            // rol 32bit count

    iTEST_PROL1B,           // rol 8bit  1
    iTEST_PROL1H,           // rol 16bit 1
    iTEST_PROL1W,           // rol 32bit 1

} iTEST;




//
//  Compare two 128bit bit blocks
//

bool vmp_IsPVEqual( void * const pvA, void * const pvB )
{
    ASSERT_PTR4(pvA);
    ASSERT_PTR4(pvB);
    
    if ( memcmp( pvA, pvB, 16 ) )
      {
        return false;
      }

    return true;
}                                               



//
//  Test Packed Shift 128bit
//
static void test_pV( bool bAlign, iTEST iFunc )
{
    uint    count;
    int8    v, buf[3*16 +15], *bt, *bu, *bv;

    if (bAlign)
      {
        bt = (int8*) ALIGN16((int)buf);
      }
    else
      {
        bt = buf;
        if (0==(0x3 & (int)buf))
          {
            bt += 4;
          }
      }

    bu = bt + (128/8);
    bv = bu + (128/8);

    v = 0;
    do {
        for ( count=0; count < 8+1; count++ )
          {
            bt[0]=v;     bt[1]=v+1;   bt[2]=v+2;   bt[3]=v+3;
            bt[4]=v+4;   bt[5]=v+5;   bt[6]=v+6;   bt[7]=v+7;
            bt[8]=v+8;   bt[9]=v+9;   bt[10]=v+10; bt[11]=v+11;
            bt[12]=v+12; bt[13]=v+13; bt[14]=v+14; bt[15]=v+15;

            switch (iFunc)
              {
                case iTEST_PSRAB:                       // sra n 8bit
                    vmp_psraBGeneric(bu, bt, count);
                    if ( bAlign )
                        vmp_psraB(   bv, bt, count);
                    else
                        vmp_psraB0(  bv, bt, count);
                    break;

                case iTEST_PSRAH:                       // sra n 16bit
                    vmp_psraHGeneric((int16*)bu, (int16*)bt, count);
                    if ( bAlign )
                        vmp_psraH(   (int16*)bv, (int16*)bt, count);
                    else
                        vmp_psraH0(  (int16*)bv, (int16*)bt, count);
                    break;

                case iTEST_PSRAW:                       // sra n 32bit
                    vmp_psraWGeneric((int32*)bu, (int32*)bt, count);
                    if ( bAlign )
                        vmp_psraW(   (int32*)bv, (int32*)bt, count);
                    else
                        vmp_psraW0(  (int32*)bv, (int32*)bt, count);
                    break;

                case iTEST_PSRA1B:                      // sra 1 8bit   
                    vmp_psra1BGeneric(bu, bt);
                    if ( bAlign )
                        vmp_psra1B(   bv, bt);
                    else
                        vmp_psra1B0(  bv, bt);
                    break;

                case iTEST_PSRA1H:                      // sra 1 16bit
                    vmp_psra1HGeneric((int16*)bu, (int16*)bt);
                    if ( bAlign )
                        vmp_psra1H(   (int16*)bv, (int16*)bt);
                    else
                        vmp_psra1H0(  (int16*)bv, (int16*)bt);
                    break;

                case iTEST_PSRA1W:                      // sra 1 32bit
                    vmp_psra1WGeneric((int32*)bu, (int32*)bt);
                    if ( bAlign )
                        vmp_psra1W(   (int32*)bv, (int32*)bt);
                    else
                        vmp_psra1W0(  (int32*)bv, (int32*)bt);
                    break;


                case iTEST_PSLLB:                       // sll n 8bit
                    vmp_psllBGeneric((uint8*)bu, (uint8*)bt, count);
                    if ( bAlign )
                        vmp_psllB(   (uint8*)bv, (uint8*)bt, count);
                    else
                        vmp_psllB0(  (uint8*)bv, (uint8*)bt, count);
                    break;

                case iTEST_PSLLH:                       // sll n 16bit
                    vmp_psllHGeneric((uint16*)bu, (uint16*)bt, count);
                    if ( bAlign )
                        vmp_psllH(   (uint16*)bv, (uint16*)bt, count);
                    else
                        vmp_psllH0(  (uint16*)bv, (uint16*)bt, count);
                    break;

                case iTEST_PSLLW:                       // sll n 32bit
                    vmp_psllWGeneric((uint32*)bu, (uint32*)bt, count);
                    if ( bAlign )
                        vmp_psllW(   (uint32*)bv, (uint32*)bt, count);
                    else
                        vmp_psllW0(  (uint32*)bv, (uint32*)bt, count);
                    break;

                case iTEST_PSLL1B:                      // sll 1 8bit   
                    vmp_psll1BGeneric((uint8*)bu, (uint8*)bt);
                    if ( bAlign )
                        vmp_psll1B(   (uint8*)bv, (uint8*)bt);
                    else
                        vmp_psll1B0(  (uint8*)bv, (uint8*)bt);
                    break;

                case iTEST_PSLL1H:                      // sll 1 16bit
                    vmp_psll1HGeneric((uint16*)bu, (uint16*)bt);
                    if ( bAlign )
                        vmp_psll1H(   (uint16*)bv, (uint16*)bt);
                    else
                        vmp_psll1H0(  (uint16*)bv, (uint16*)bt);
                    break;

                case iTEST_PSLL1W:                      // sll 1 32bit
                    vmp_psll1WGeneric((uint32*)bu, (uint32*)bt);
                    if ( bAlign )
                        vmp_psll1W(   (uint32*)bv, (uint32*)bt);
                    else
                        vmp_psll1W0(  (uint32*)bv, (uint32*)bt);
                    break;


                case iTEST_PSRLB:                       // srl n 8bit
                    vmp_psrlBGeneric((uint8*)bu, (uint8*)bt, count);
                    if ( bAlign )
                        vmp_psrlB(   (uint8*)bv, (uint8*)bt, count);
                    else
                        vmp_psrlB0(  (uint8*)bv, (uint8*)bt, count);
                    break;

                case iTEST_PSRLH:                       // srl n 16bit
                    vmp_psrlHGeneric((uint16*)bu, (uint16*)bt, count);
                    if ( bAlign )
                        vmp_psrlH(   (uint16*)bv, (uint16*)bt, count);
                    else
                        vmp_psrlH0(  (uint16*)bv, (uint16*)bt, count);
                    break;

                case iTEST_PSRLW:                       // srl n 32bit
                    vmp_psrlWGeneric((uint32*)bu, (uint32*)bt, count);
                    if ( bAlign )
                        vmp_psrlW(   (uint32*)bv, (uint32*)bt, count);
                    else
                        vmp_psrlW0(  (uint32*)bv, (uint32*)bt, count);
                    break;

                case iTEST_PSRL1B:                      // srl 1 8bit   
                    vmp_psrl1BGeneric((uint8*)bu, (uint8*)bt);
                    if ( bAlign )
                        vmp_psrl1B(   (uint8*)bv, (uint8*)bt);
                    else
                        vmp_psrl1B0(  (uint8*)bv, (uint8*)bt);
                    break;

                case iTEST_PSRL1H:                      // srl 1 16bit
                    vmp_psrl1HGeneric((uint16*)bu, (uint16*)bt);
                    if ( bAlign )
                        vmp_psrl1H(   (uint16*)bv, (uint16*)bt);
                    else
                        vmp_psrl1H0(  (uint16*)bv, (uint16*)bt);
                    break;

                case iTEST_PSRL1W:                      // srl 1 32bit
                    vmp_psrl1WGeneric((uint32*)bu, (uint32*)bt);
                    if ( bAlign )
                        vmp_psrl1W(   (uint32*)bv, (uint32*)bt);
                    else
                        vmp_psrl1W0(  (uint32*)bv, (uint32*)bt);
                    break;

                case iTEST_PROLB:                       // rol n 8bit
                    vmp_prolBGeneric((uint8*)bu, (uint8*)bt, count);
                    if ( bAlign )
                        vmp_prolB(   (uint8*)bv, (uint8*)bt, count);
                    else
                        vmp_prolB0(  (uint8*)bv, (uint8*)bt, count);
                    break;

                case iTEST_PROLH:                       // rol n 16bit
                    vmp_prolHGeneric((uint16*)bu, (uint16*)bt, count);
                    if ( bAlign )
                        vmp_prolH(   (uint16*)bv, (uint16*)bt, count);
                    else
                        vmp_prolH0(  (uint16*)bv, (uint16*)bt, count);
                    break;

                case iTEST_PROLW:                       // rol n 32bit
                    vmp_prolWGeneric((uint32*)bu, (uint32*)bt, count);
                    if ( bAlign )
                        vmp_prolW(   (uint32*)bv, (uint32*)bt, count);
                    else
                        vmp_prolW0(  (uint32*)bv, (uint32*)bt, count);
                    break;

                case iTEST_PROL1B:                      // rol 1 8bit   
                    vmp_prol1BGeneric((uint8*)bu, (uint8*)bt);
                    if ( bAlign )
                        vmp_prol1B(   (uint8*)bv, (uint8*)bt);
                    else
                        vmp_prol1B0(  (uint8*)bv, (uint8*)bt);
                    break;

                case iTEST_PROL1H:                      // rol 1 16bit
                    vmp_prol1HGeneric((uint16*)bu, (uint16*)bt);
                    if ( bAlign )
                        vmp_prol1H(   (uint16*)bv, (uint16*)bt);
                    else
                        vmp_prol1H0(  (uint16*)bv, (uint16*)bt);
                    break;

                case iTEST_PROL1W:                      // rol 1 32bit
                    vmp_prol1WGeneric((uint32*)bu, (uint32*)bt);
                    if ( bAlign )
                        vmp_prol1W(   (uint32*)bv, (uint32*)bt);
                    else
                        vmp_prol1W0(  (uint32*)bv, (uint32*)bt);
                    break;
              }

            if ( !vmp_IsPVEqual( bu, bv ) )
              {
                cout << "Mismatch on " << v << " and count " << count << endl;
              }
          }
      } while( 0 != ++v);
}







int main( int argc, char *argv[] )
{
    CpuInfo cinfo;
    char szBuf[ CPU_SZBUF_MAX ];

    CpuDetect( &cinfo );                                // Detect CPU
    
    cout << "\nPacked (Vector) Shift/Rotation Test\n\n";
    cout << CpuInfoStr( szBuf, &cinfo ) << endl;
    
    PRotSetup( &cinfo );                                // Setup our 3D Vector stuff

    cout << "Testing (Vector) Shift Arithmetic Right 16x8bit by 'n'" << endl;
    test_pV( false, iTEST_PSRAB );

    cout << "Testing (Vector) Shift Arithmetic Right 16x8bit by 'n' (Aligned)" << endl;
    test_pV( true,  iTEST_PSRAB );

    cout << "Testing (Vector) Shift Arithmetic Right 8x16bit by 'n'" << endl;
    test_pV( false, iTEST_PSRAH );

    cout << "Testing (Vector) Shift Arithmetic Right 8x16bit by 'n' (Aligned)" << endl;
    test_pV( true,  iTEST_PSRAH );

    cout << "Testing (Vector) Shift Arithmetic Right 4x32bit by 'n'" << endl;
    test_pV( false, iTEST_PSRAW );

    cout << "Testing (Vector) Shift Arithmetic Right 4x32bit by 'n' (Aligned)" << endl;
    test_pV( true,  iTEST_PSRAW );
    cout << endl;

    cout << "Testing (Vector) Shift Arithmetic Right 16x8bit by 1" << endl;
    test_pV( false, iTEST_PSRA1B );

    cout << "Testing (Vector) Shift Arithmetic Right 16x8bit by 1 (Aligned)" << endl;
    test_pV( true,  iTEST_PSRA1B );

    cout << "Testing (Vector) Shift Arithmetic Right 8x16bit by 1" << endl;
    test_pV( false, iTEST_PSRA1H );

    cout << "Testing (Vector) Shift Arithmetic Right 8x16bit by 1 (Aligned)" << endl;
    test_pV( true,  iTEST_PSRA1H );

    cout << "Testing (Vector) Shift Arithmetic Right 4x32bit by 1" << endl;
    test_pV( false, iTEST_PSRA1W );

    cout << "Testing (Vector) Shift Arithmetic Right 4x32bit by 1 (Aligned)" << endl;
    test_pV( true,  iTEST_PSRA1W );
    cout << endl;


    cout << "Testing (Vector) Shift Left Logical 16x8bit by 'n'" << endl;
    test_pV( false, iTEST_PSLLB );

    cout << "Testing (Vector) Shift Left Logical 16x8bit by 'n' (Aligned)" << endl;
    test_pV( true,  iTEST_PSLLB );

    cout << "Testing (Vector) Shift Left Logical 8x16bit by 'n'" << endl;
    test_pV( false, iTEST_PSLLH );

    cout << "Testing (Vector) Shift Left Logical 8x16bit by 'n' (Aligned)" << endl;
    test_pV( true,  iTEST_PSLLH );

    cout << "Testing (Vector) Shift Left Logical 4x32bit by 'n'" << endl;
    test_pV( false, iTEST_PSLLW );

    cout << "Testing (Vector) Shift Left Logical 4x32bit by 'n' (Aligned)" << endl;
    test_pV( true,  iTEST_PSLLW );
    cout << endl;

    cout << "Testing (Vector) Shift Left Logical 16x8bit by 1" << endl;
    test_pV( false, iTEST_PSLL1B );

    cout << "Testing (Vector) Shift Left Logical 16x8bit by 1 (Aligned)" << endl;
    test_pV( true,  iTEST_PSLL1B );

    cout << "Testing (Vector) Shift Left Logical 8x16bit by 1" << endl;
    test_pV( false, iTEST_PSLL1H );

    cout << "Testing (Vector) Shift Left Logical 8x16bit by 1 (Aligned)" << endl;
    test_pV( true,  iTEST_PSLL1H );

    cout << "Testing (Vector) Shift Left Logical 4x32bit by 1" << endl;
    test_pV( false, iTEST_PSLL1W );

    cout << "Testing (Vector) Shift Left Logical 4x32bit by 1 (Aligned)" << endl;
    test_pV( true,  iTEST_PSLL1W );
    cout << endl;


    cout << "Testing (Vector) Shift Right Logical 16x8bit by 'n'" << endl;
    test_pV( false, iTEST_PSRLB );

    cout << "Testing (Vector) Shift Right Logical 16x8bit by 'n' (Aligned)" << endl;
    test_pV( true,  iTEST_PSRLB );

    cout << "Testing (Vector) Shift Right Logical 8x16bit by 'n'" << endl;
    test_pV( false, iTEST_PSRLH );

    cout << "Testing (Vector) Shift Right Logical 8x16bit by 'n' (Aligned)" << endl;
    test_pV( true,  iTEST_PSRLH );

    cout << "Testing (Vector) Shift Right Logical 4x32bit by 'n'" << endl;
    test_pV( false, iTEST_PSRLW );

    cout << "Testing (Vector) Shift Right Logical 4x32bit by 'n' (Aligned)" << endl;
    test_pV( true,  iTEST_PSRLW );
    cout << endl;

    cout << "Testing (Vector) Shift Right Logical 16x8bit by 1" << endl;
    test_pV( false, iTEST_PSRL1B );

    cout << "Testing (Vector) Shift Right Logical 16x8bit by 1 (Aligned)" << endl;
    test_pV( true,  iTEST_PSRL1B );

    cout << "Testing (Vector) Shift Right Logical 8x16bit by 1" << endl;
    test_pV( false, iTEST_PSRL1H );

    cout << "Testing (Vector) Shift Right Logical 8x16bit by 1 (Aligned)" << endl;
    test_pV( true,  iTEST_PSRL1H );

    cout << "Testing (Vector) Shift Right Logical 4x32bit by 1" << endl;
    test_pV( false, iTEST_PSRL1W );

    cout << "Testing (Vector) Shift Right Logical 4x32bit by 1 (Aligned)" << endl;
    test_pV( true,  iTEST_PSRL1W );
    cout << endl;


    cout << "Testing (Vector) Rotate Left 16x8bit by 'n'" << endl;
    test_pV( false, iTEST_PROLB );

    cout << "Testing (Vector) Rotate Left 16x8bit by 'n' (Aligned)" << endl;
    test_pV( true,  iTEST_PROLB );

    cout << "Testing (Vector) Rotate Left 8x16bit by 'n'" << endl;
    test_pV( false, iTEST_PROLH );

    cout << "Testing (Vector) Rotate Left 8x16bit by 'n' (Aligned)" << endl;
    test_pV( true,  iTEST_PROLH );

    cout << "Testing (Vector) Rotate Left 4x32bit by 'n'" << endl;
    test_pV( false, iTEST_PROLW );

    cout << "Testing (Vector) Rotate Left 4x32bit by 'n' (Aligned)" << endl;
    test_pV( true,  iTEST_PROLW );
    cout << endl;

    cout << "Testing (Vector) Rotate Left 16x8bit by 1" << endl;
    test_pV( false, iTEST_PROL1B );

    cout << "Testing (Vector) Rotate Left 16x8bit by 1 (Aligned)" << endl;
    test_pV( true,  iTEST_PROL1B );

    cout << "Testing (Vector) Rotate Left 8x16bit by 1" << endl;
    test_pV( false, iTEST_PROL1H );

    cout << "Testing (Vector) Rotate Left 8x16bit by 1 (Aligned)" << endl;
    test_pV( true,  iTEST_PROL1H );

    cout << "Testing (Vector) Rotate Left 4x32bit by 1" << endl;
    test_pV( false, iTEST_PROL1W );

    cout << "Testing (Vector) Rotate Left 4x32bit by 1 (Aligned)" << endl;
    test_pV( true,  iTEST_PROL1W );
    cout << endl;

    cout << "Fin\n\n" << flush;

    return 0;
}

