        TITLE   Book PSubX86.asm - Packed Subtraction
        PAGE    53,160

;
; PSubX86.asm - 
;
; This document contains file that you may use at your own risk in your own programs.
;
; This module is designed to accomodate X86 processors in the book.
;
;       NOTE: Needs masm 6.14 or later!
;
;       Created - 31 Dec 2001 - J.Leiterman
;
;       Modifications
;
;               
;
;       Tabs = 8



        .XLIST

        .387
        .686P
        .MMX
        .model flat, C

        include x86.inc
        include KariType.inc
 
        .LIST



        .data



PSubX86  SEGMENT USE32 PUBLIC 'CODE'



;_________________________________________________________________________________________________
;       (Quad) Vector 16x8bit Subtraction
;
;void vmp_psubB( uint8 * const pbD, const uint8 * const pbA, const uint8 * const pbB )
;{
;       *(pbD+0) = *(pbA+0) - *(pbB+0);
;       *(pbD+1) = *(pbA+1) - *(pbB+1);
;       *(pbD+2) = *(pbA+2) - *(pbB+2);
;       *(pbD+3) = *(pbA+3) - *(pbB+3);
;       *(pbD+4) = *(pbA+4) - *(pbB+4);
;       *(pbD+5) = *(pbA+5) - *(pbB+5);
;       *(pbD+6) = *(pbA+6) - *(pbB+6);
;       *(pbD+7) = *(pbA+7) - *(pbB+7);
;       *(pbD+8) = *(pbA+8) - *(pbB+8);
;       *(pbD+9) = *(pbA+9) - *(pbB+9);
;       *(pbD+10) = *(pbA+10) - *(pbB+10);
;       *(pbD+11) = *(pbA+11) - *(pbB+11);
;       *(pbD+12) = *(pbA+12) - *(pbB+12);
;       *(pbD+13) = *(pbA+13) - *(pbB+13);
;       *(pbD+14) = *(pbA+14) - *(pbB+14);
;       *(pbD+15) = *(pbA+15) - *(pbB+15);
;}                                                                      

        align   16
        .MMX
        public  vmp_psubBAsmMMX
vmp_psubBAsmMMX       proc    near pbD:ptr, pbA:ptr, pbB:ptr
;;;        emms
ASSERT_PTR4(pbA)
ASSERT_PTR4(pbB)
ASSERT_PTR4(pbD)

        mov     eax,pbA                         ; Vector A
        mov     edx,pbB                         ; Vector B
        mov     ecx,pbD                         ; Vector Destination

        movq    mm0,[eax+0]                     ; Read A Data
        movq    mm1,[eax+8]
        movq    mm2,[edx+0]                     ; Read B Data
        movq    mm3,[edx+8]

        psubb   mm0,mm2                         ;       {Al-Bl}
        psubb   mm1,mm3                         ;       {Au-Bu}

        movq    [ecx+0],mm0                     ;       {Al-Bl}
        movq    [ecx+8],mm1                     ;       {Au-Bu}
        
;;;        emms
        ret
vmp_psubBAsmMMX endp



;_________________________________________________________________________________________________
;       *** SSE2 version *** 128bit unaligned

        align   16
        .XMM
        public  vmp_psubBAsmSSE20
vmp_psubBAsmSSE20       proc    near  pbD:ptr, pbA:ptr, pbB:ptr

ASSERT_PTR4(pbA)
ASSERT_PTR4(pbB)
ASSERT_PTR4(pbD)

        mov     eax,pbA                         ; Vector A
        mov     edx,pbB                         ; Vector B

        movdqu  xmm0,[eax]                      ; Read A Data
        movdqu  xmm1,[edx]                      ; Read B Data

        mov     eax,pbD                         ; Vector Destination
        psubb   xmm0,xmm1                       ;       {vA-vB}

        movdqu  [eax],xmm0                      ; Write D Data
        ret
vmp_psubBAsmSSE20 endp



;_________________________________________________________________________________________________
;       *** SSE2 version *** 128bit aligned

        align   16
        .XMM
        public  vmp_psubBAsmSSE2
vmp_psubBAsmSSE2       proc    near pbD:ptr, pbA:ptr, pbB:ptr

ASSERT_PTR16(pbA)
ASSERT_PTR16(pbB)
ASSERT_PTR16(pbD)

        mov     eax,pbA                         ; Vector A
        mov     edx,pbB                         ; Vector B

        movdqa  xmm0,[eax]                      ; Read A Data
        movdqa  xmm1,[edx]                      ; Read B Data

        mov     eax,pbD                         ; Vector Destination
        psubb   xmm0,xmm1                       ;       {vA-vB}

        movdqa  [eax],xmm0                      ; Write D Data
        ret
vmp_psubBAsmSSE2 endp



;_________________________________________________________________________________________________
;=================================================================================================
;       Packed (un)signed Subtraction 8x16bit
;
;void vmp_psubH( uint16 * const phD, const uint16 * const phA, const uint16 * const phB )
;{
;       *(phD+0) = *(phA+0) - *(phB+0);
;       *(phD+1) = *(phA+1) - *(phB+1);
;       *(phD+2) = *(phA+2) - *(phB+2);
;       *(phD+3) = *(phA+3) - *(phB+3);
;       *(phD+4) = *(phA+4) - *(phB+4);
;       *(phD+5) = *(phA+5) - *(phB+5);
;       *(phD+6) = *(phA+6) - *(phB+6);
;       *(phD+7) = *(phA+7) - *(phB+7);
;}                                                                      

        align   16
        .MMX
        public  vmp_psubHAsmMMX
vmp_psubHAsmMMX       proc    near phD:ptr, phA:ptr, phB:ptr
;;;        emms
ASSERT_PTR4(phA)
ASSERT_PTR4(phB)
ASSERT_PTR4(phD)

        mov     eax,phA                         ; Vector A
        mov     edx,phB                         ; Vector B
        mov     ecx,phD                         ; Vector Destination

        movq    mm0,[eax+0]                     ; Read A Data
        movq    mm1,[eax+8]
        movq    mm2,[edx+0]                     ; Read B Data
        movq    mm3,[edx+8]

        psubw   mm0,mm2                         ;       {Al-Bl}
        psubw   mm1,mm3                         ;       {Au-Bu}

        movq    [ecx+0],mm0                     ; Write Destination
        movq    [ecx+8],mm1
        
;;;        emms
        ret
vmp_psubHAsmMMX endp



;_________________________________________________________________________________________________
;       *** SSE2 version *** 128bit unaligned

        align   16
        .XMM
        public  vmp_psubHAsmSSE20
vmp_psubHAsmSSE20       proc    near  pbD:ptr, pbA:ptr, pbB:ptr

ASSERT_PTR4(pbA)
ASSERT_PTR4(pbB)
ASSERT_PTR4(pbD)

        mov     eax,pbA                         ; Vector A
        mov     edx,pbB                         ; Vector B

        movdqu  xmm0,[eax]                      ; Read A Data
        movdqu  xmm1,[edx]                      ; Read B Data

        mov     eax,pbD                         ; Vector Destination
        psubw   xmm0,xmm1                       ;       {vA-vB}

        movdqu  [eax],xmm0                      ; Write D Data
        ret
vmp_psubHAsmSSE20 endp



;_________________________________________________________________________________________________
;       *** SSE2 version *** 128bit aligned

        align   16
        .XMM
        public  vmp_psubHAsmSSE2
vmp_psubHAsmSSE2       proc    near pbD:ptr, pbA:ptr, pbB:ptr

ASSERT_PTR16(pbA)
ASSERT_PTR16(pbB)
ASSERT_PTR16(pbD)

        mov     eax,pbA                         ; Vector A
        mov     edx,pbB                         ; Vector B

        movdqa  xmm0,[eax]                      ; Read A Data
        movdqa  xmm1,[edx]                      ; Read B Data

        mov     eax,pbD                         ; Vector Destination
        psubw   xmm0,xmm1                       ;       {vA-vB}

        movdqa  [eax],xmm0                      ; Write D Data
        ret
vmp_psubHAsmSSE2 endp



;_________________________________________________________________________________________________
;=================================================================================================
;       Packed (un)signed Subtraction 4x32bit
;
;void vmp_psubW( uint32 * const pwD, const uint32 * const pwA, const uint32 * const pwB )
;{
;       *(pwD+0) = *(pwA+0) - *(pwB+0);
;       *(pwD+1) = *(pwA+1) - *(pwB+1);
;       *(pwD+2) = *(pwA+2) - *(pwB+2);
;       *(pwD+3) = *(pwA+3) - *(pwB+3);
;}                                                                      

        align   16
        public  vmp_psubWAsmMMX
vmp_psubWAsmMMX       proc    near phD:ptr, phA:ptr, phB:ptr
        .MMX
;;;        emms
ASSERT_PTR4(phA)
ASSERT_PTR4(phB)
ASSERT_PTR4(phD)

        mov     eax,phA                         ; Vector A
        mov     edx,phB                         ; Vector B
        mov     ecx,phD                         ; Vector Destination

        movq    mm0,[eax+0]                     ;vA[0...1] {a1 a0}
        movq    mm1,[eax+8]                     ;vA[2...3] {a3 a2}
        movq    mm2,[edx+0]                     ;vB[0...1] {b1 b0}
        movq    mm3,[edx+8]                     ;vB[2...3] {b3 b2}

        psubd   mm0,mm2                         ;       {Al-Bl}
        psubd   mm1,mm3                         ;       {Au-Bu}

        movq    [ecx+0],mm0                     ;       {Al-Bl}
        movq    [ecx+8],mm1                     ;       {Au-Bu}
        
;;;        emms
        ret
vmp_psubWAsmMMX endp


;_________________________________________________________________________________________________
;       *** SSE2 version *** 128bit unaligned

        align   16
        .XMM
        public  vmp_psubWAsmSSE20
vmp_psubWAsmSSE20       proc    near  pbD:ptr, pbA:ptr, pbB:ptr

ASSERT_PTR4(pbA)
ASSERT_PTR4(pbB)
ASSERT_PTR4(pbD)

        mov     eax,pbA                         ; Vector A
        mov     edx,pbB                         ; Vector B

        movdqu  xmm0,[eax]                      ; Read A Data
        movdqu  xmm1,[edx]                      ; Read B Data

        mov     eax,pbD                         ; Vector Destination
        psubd   xmm0,xmm1                       ;       {vA-vB}

        movdqu  [eax],xmm0                      ; Write D Data
        ret
vmp_psubWAsmSSE20 endp



;_________________________________________________________________________________________________
;       *** SSE2 version *** 128bit aligned

        align   16
        .XMM
        public  vmp_psubWAsmSSE2
vmp_psubWAsmSSE2       proc    near pbD:ptr, pbA:ptr, pbB:ptr

ASSERT_PTR16(pbA)
ASSERT_PTR16(pbB)
ASSERT_PTR16(pbD)

        mov     eax,pbA                         ; Vector A
        mov     edx,pbB                         ; Vector B

        movdqa  xmm0,[eax]                      ; Read A Data
        movdqa  xmm1,[edx]                      ; Read B Data

        mov     eax,pbD                         ; Vector Destination
        psubd   xmm0,xmm1                       ;       {vA-vB}

        movdqa  [eax],xmm0                      ; Write D Data
        ret
vmp_psubWAsmSSE2 endp



;_________________________________________________________________________________________________
;=================================================================================================
;       (Quad) Vector Signed 16x8 Subtraction w/Saturation  {-128...127}

        align   16
        public  vmp_psubsBAsmMMX
vmp_psubsBAsmMMX       proc    near , pbD:ptr, pbA:ptr, pbB:ptr
        .MMX
;;;        emms
ASSERT_PTR4(pbA)
ASSERT_PTR4(pbB)
ASSERT_PTR4(pbD)

        mov     eax,pbA                         ; Vector A
        mov     edx,pbB                         ; Vector B
        mov     ecx,pbD                         ; Vector Destination

        movq    mm0,[eax+0]                     ;vA[0...7] {a7 a6 a5 a4 a3 a2 a1 a0}
        movq    mm1,[eax+8]                     ;vA[8...15] {a15 a14 a13 a12 a11 a10 a9 a8}
        movq    mm2,[edx+0]                     ;vB[0...7] {b7 b6 b5 b4 b3 b2 b1 b0}
        movq    mm3,[edx+8]                     ;vB[8...15] {b15 b14 b13 b12 b11 b10 b9 b8}

        psubsb  mm0,mm2                         ;       {Al-Bl}
        psubsb  mm1,mm3                         ;       {Au-Bu}

        movq    [ecx+0],mm0                     ;       {Al-Bl}
        movq    [ecx+8],mm1                     ;       {Au-Bu}
        
;;;        emms
        ret
vmp_psubsBAsmMMX endp




;_________________________________________________________________________________________________
;       *** SSE2 version *** 128bit unaligned

        align   16
        .XMM
        public  vmp_psubsBAsmSSE20
vmp_psubsBAsmSSE20       proc    near  pbD:ptr, pbA:ptr, pbB:ptr

ASSERT_PTR4(pbA)
ASSERT_PTR4(pbB)
ASSERT_PTR4(pbD)

        mov     eax,pbA                         ; Vector A
        mov     edx,pbB                         ; Vector B

        movdqu  xmm0,[eax]                      ; Read A Data
        movdqu  xmm1,[edx]                      ; Read B Data

        mov     eax,pbD                         ; Vector Destination
        psubsb  xmm0,xmm1                       ;       {vA-vB}

        movdqu  [eax],xmm0                      ; Write D Data
        ret
vmp_psubsBAsmSSE20 endp



;_________________________________________________________________________________________________
;       *** SSE2 version *** 128bit aligned

        align   16
        .XMM
        public  vmp_psubsBAsmSSE2
vmp_psubsBAsmSSE2       proc    near pbD:ptr, pbA:ptr, pbB:ptr

ASSERT_PTR16(pbA)
ASSERT_PTR16(pbB)
ASSERT_PTR16(pbD)

        mov     eax,pbA                         ; Vector A
        mov     edx,pbB                         ; Vector B

        movdqa  xmm0,[eax]                      ; Read A Data
        movdqa  xmm1,[edx]                      ; Read B Data

        mov     eax,pbD                         ; Vector Destination
        psubsb  xmm0,xmm1                       ;       {vA-vB}

        movdqa  [eax],xmm0                      ; Write D Data
        ret
vmp_psubsBAsmSSE2 endp



;_________________________________________________________________________________________________
;=================================================================================================
;       (Quad) Vector Unsigned 16x8 Subtraction w/Saturation  {0...255}

        align   16
        public  vmp_psubusBAsmMMX
vmp_psubusBAsmMMX       proc    near pbD:ptr, pbA:ptr, pbB:ptr
        .MMX
;;;        emms
ASSERT_PTR4(pbA)
ASSERT_PTR4(pbB)
ASSERT_PTR4(pbD)

        mov     eax,pbA                         ; Vector A
        mov     edx,pbB                         ; Vector B
        mov     ecx,pbD                         ; Vector Destination

        movq    mm0,[eax+0]                     ;vA[0...7] {a7 a6 a5 a4 a3 a2 a1 a0}
        movq    mm1,[eax+8]                     ;vA[8...15] {a15 a14 a13 a12 a11 a10 a9 a8}
        movq    mm2,[edx+0]                     ;vB[0...7] {b7 b6 b5 b4 b3 b2 b1 b0}
        movq    mm3,[edx+8]                     ;vB[8...15] {b15 b14 b13 b12 b11 b10 b9 b8}

        psubusb mm0,mm2                         ;       {Al-Bl}
        psubusb mm1,mm3                         ;       {Au-Bu}

        movq    [ecx+0],mm0                     ;       {Al-Bl}
        movq    [ecx+8],mm1                     ;       {Au-Bu}
        
;;;        emms
        ret
vmp_psubusBAsmMMX endp



;_________________________________________________________________________________________________
;       *** SSE2 version *** 128bit unaligned

        align   16
        .XMM
        public  vmp_psubusBAsmSSE20
vmp_psubusBAsmSSE20       proc    near  pbD:ptr, pbA:ptr, pbB:ptr

ASSERT_PTR4(pbA)
ASSERT_PTR4(pbB)
ASSERT_PTR4(pbD)

        mov     eax,pbA                         ; Vector A
        mov     edx,pbB                         ; Vector B

        movdqu  xmm0,[eax]                      ; Read A Data
        movdqu  xmm1,[edx]                      ; Read B Data

        mov     eax,pbD                         ; Vector Destination
        psubusb xmm0,xmm1                       ;       {vA-vB}

        movdqu  [eax],xmm0                      ; Write D Data
        ret
vmp_psubusBAsmSSE20 endp



;_________________________________________________________________________________________________
;       *** SSE2 version *** 128bit aligned

        align   16
        .XMM
        public  vmp_psubusBAsmSSE2
vmp_psubusBAsmSSE2       proc    near pbD:ptr, pbA:ptr, pbB:ptr

ASSERT_PTR16(pbA)
ASSERT_PTR16(pbB)
ASSERT_PTR16(pbD)

        mov     eax,pbA                         ; Vector A
        mov     edx,pbB                         ; Vector B

        movdqa  xmm0,[eax]                      ; Read A Data
        movdqa  xmm1,[edx]                      ; Read B Data

        mov     eax,pbD                         ; Vector Destination
        psubusb xmm0,xmm1                       ;       {vA-vB}

        movdqa  [eax],xmm0                      ; Write D Data
        ret
vmp_psubusBAsmSSE2 endp



;_________________________________________________________________________________________________
;=================================================================================================
;       (Quad) Vector Signed 8x16 Subtraction w/Saturation {-32768...32767}
;
        align   16
        public  vmp_psubsHAsmMMX
vmp_psubsHAsmMMX       proc    near phD:ptr, phA:ptr, phB:ptr
        .MMX
;;;        emms
ASSERT_PTR4(phA)
ASSERT_PTR4(phB)
ASSERT_PTR4(phD)

        mov     eax,phA                         ; Vector A
        mov     edx,phB                         ; Vector B
        mov     ecx,phD                         ; Vector Destination

        movq    mm0,[eax+0]                     ;vA[0...3] {a3 a2 a1 a0}
        movq    mm1,[eax+8]                     ;vA[4...7] {a7 a6 a5 a4}
        movq    mm2,[edx+0]                     ;vB[0...3] {b3 b2 b1 b0}
        movq    mm3,[edx+8]                     ;vB[4...7] {b7 b6 b5 b4}

        psubsw  mm0,mm2                         ;       {Al-Bl}
        psubsw  mm1,mm3                         ;       {Au-Bu}

        movq    [ecx+0],mm0                     ;       {Al-Bl}
        movq    [ecx+8],mm1                     ;       {Au-Bu}
        
;;;        emms
        ret
vmp_psubsHAsmMMX endp



;_________________________________________________________________________________________________
;       *** SSE2 version *** 128bit unaligned

        align   16
        .XMM
        public  vmp_psubsHAsmSSE20
vmp_psubsHAsmSSE20       proc    near  pbD:ptr, pbA:ptr, pbB:ptr

ASSERT_PTR4(pbA)
ASSERT_PTR4(pbB)
ASSERT_PTR4(pbD)

        mov     eax,pbA                         ; Vector A
        mov     edx,pbB                         ; Vector B

        movdqu  xmm0,[eax]                      ; Read A Data
        movdqu  xmm1,[edx]                      ; Read B Data

        mov     eax,pbD                         ; Vector Destination
        psubsw  xmm0,xmm1                       ;       {vA-vB}

        movdqu  [eax],xmm0                      ; Write D Data
        ret
vmp_psubsHAsmSSE20 endp



;_________________________________________________________________________________________________
;       *** SSE2 version *** 128bit aligned

        align   16
        .XMM
        public  vmp_psubsHAsmSSE2
vmp_psubsHAsmSSE2       proc    near pbD:ptr, pbA:ptr, pbB:ptr

ASSERT_PTR16(pbA)
ASSERT_PTR16(pbB)
ASSERT_PTR16(pbD)

        mov     eax,pbA                         ; Vector A
        mov     edx,pbB                         ; Vector B

        movdqa  xmm0,[eax]                      ; Read A Data
        movdqa  xmm1,[edx]                      ; Read B Data

        mov     eax,pbD                         ; Vector Destination
        psubsw  xmm0,xmm1                       ;       {vA-vB}

        movdqa  [eax],xmm0                      ; Write D Data
        ret
vmp_psubsHAsmSSE2 endp



;_________________________________________________________________________________________________
;=================================================================================================
;       (Quad) Vector Unsigned 8x16 Subtraction w/Saturation {0...65535}
;
        align   16
        public  vmp_psubusHAsmMMX
vmp_psubusHAsmMMX       proc    near phD:ptr, phA:ptr, phB:ptr
        .MMX
;;;        emms
ASSERT_PTR4(phA)
ASSERT_PTR4(phB)
ASSERT_PTR4(phD)

        mov     eax,phA                         ; Vector A
        mov     edx,phB                         ; Vector B
        mov     ecx,phD                         ; Vector Destination

        movq    mm0,[eax+0]                     ;vA[0...3] {a3 a2 a1 a0}
        movq    mm1,[eax+8]                     ;vA[4...7] {a7 a6 a5 a4}
        movq    mm2,[edx+0]                     ;vB[0...3] {b3 b2 b1 b0}
        movq    mm3,[edx+8]                     ;vB[4...7] {b7 b6 b5 b4}

        psubusw mm0,mm2                         ;       {Al-Bl}
        psubusw mm1,mm3                         ;       {Au-Bu}

        movq    [ecx+0],mm0                     ;       {Al-Bl}
        movq    [ecx+8],mm1                     ;       {Au-Bu}
        
;;;        emms
        ret
vmp_psubusHAsmMMX endp



;_________________________________________________________________________________________________
;       *** SSE2 version *** 128bit unaligned

        align   16
        .XMM
        public  vmp_psubusHAsmSSE20
vmp_psubusHAsmSSE20       proc    near  pbD:ptr, pbA:ptr, pbB:ptr

ASSERT_PTR4(pbA)
ASSERT_PTR4(pbB)
ASSERT_PTR4(pbD)

        mov     eax,pbA                         ; Vector A
        mov     edx,pbB                         ; Vector B

        movdqu  xmm0,[eax]                      ; Read A Data
        movdqu  xmm1,[edx]                      ; Read B Data

        mov     eax,pbD                         ; Vector Destination
        psubusw xmm0,xmm1                       ;       {vA-vB}

        movdqu  [eax],xmm0                      ; Write D Data
        ret
vmp_psubusHAsmSSE20 endp



;_________________________________________________________________________________________________
;       *** SSE2 version *** 128bit aligned

        align   16
        .XMM
        public  vmp_psubusHAsmSSE2
vmp_psubusHAsmSSE2       proc    near pbD:ptr, pbA:ptr, pbB:ptr

ASSERT_PTR16(pbA)
ASSERT_PTR16(pbB)
ASSERT_PTR16(pbD)

        mov     eax,pbA                         ; Vector A
        mov     edx,pbB                         ; Vector B

        movdqa  xmm0,[eax]                      ; Read A Data
        movdqa  xmm1,[edx]                      ; Read B Data

        mov     eax,pbD                         ; Vector Destination
        psubusw xmm0,xmm1                       ;       {vA-vB}

        movdqa  [eax],xmm0                      ; Write D Data
        ret
vmp_psubusHAsmSSE2 endp



PSubX86  ends
        end

