        TITLE   Book PMulX86.asm - Packed Multiplication
        PAGE    53,160

;
; PMulX86.asm - 
;
; This document contains file that you may use at your own risk in your own programs.
;
; This module is designed to accomodate X86 processors in the book.
;
;       NOTE: Needs masm 6.14 or later!
;
;       Created - 31 Dec 2001 - J.Leiterman
;
;       Modifications
;
;               
;
;       Tabs = 8



        .XLIST

        .387
        .686P
        .MMX
        .model flat, C

        include x86.inc
        include KariType.inc
 
        .LIST



        .data

lmask   dq      00ff00ff00ff00ffh,00ff00ff00ff00ffh


PMulX86  SEGMENT USE32 PUBLIC 'CODE'


;_________________________________________________________________________________________________
;=================================================================================================
;       Packed signed Multiplication High(16bit) = 8x16bit
;
;void vmp_pmulhH( int16 * const phD, const int16 * const phA, const int16 * const phB )
;{
;       *(phD+0) = (int16)((((int)*(phA+0)) * ((int)*(phB+0))) >> 16);
;       *(phD+1) = (int16)((((int)*(phA+1)) * ((int)*(phB+1))) >> 16);
;       *(phD+2) = (int16)((((int)*(phA+2)) * ((int)*(phB+2))) >> 16);
;       *(phD+3) = (int16)((((int)*(phA+3)) * ((int)*(phB+3))) >> 16);
;       *(phD+4) = (int16)((((int)*(phA+4)) * ((int)*(phB+4))) >> 16);
;       *(phD+5) = (int16)((((int)*(phA+5)) * ((int)*(phB+5))) >> 16);
;       *(phD+6) = (int16)((((int)*(phA+6)) * ((int)*(phB+6))) >> 16);
;       *(phD+7) = (int16)((((int)*(phA+7)) * ((int)*(phB+7))) >> 16);
;}                                                                      

;       Packed signed Multiplication High(16bit) = 8x16bit
        align   16
        .MMX
        public  vmp_pmulhHAsmMMX
vmp_pmulhHAsmMMX       proc    near phD:ptr, phA:ptr, phB:ptr
;;;        emms
ASSERT_PTR4(phD)
ASSERT_PTR4(phA)
ASSERT_PTR4(phB)

        mov     edx,phB                         ; Vector B
        mov     eax,phA                         ; Vector A
        mov     ecx,phD                         ; Vector Destination

        movq    mm0,[edx+0]                     ; Read B Data
        movq    mm1,[edx+8]
        movq    mm2,[eax+0]                     ; Read A Data
        movq    mm3,[eax+8]

        pmulhw  mm0,mm2
        pmulhw  mm1,mm3

        movq    [ecx+0],mm0                     ; Write D upper 16bits
        movq    [ecx+8],mm1                     ;
        
;;;        emms
        ret
vmp_pmulhHAsmMMX endp



;_________________________________________________________________________________________________
;       *** SSE2 version *** 128bit unaligned

        align   16
        .XMM
        public  vmp_pmulhHAsmSSE20
vmp_pmulhHAsmSSE20       proc    near phD:ptr, phA:ptr, phB:ptr

ASSERT_PTR4(phD)
ASSERT_PTR4(phA)
ASSERT_PTR4(phB)

        mov     edx,phB                         ; Vector B
        mov     eax,phA                         ; Vector A

        movdqu  xmm0,[edx]                      ; Read B Data
        movdqu  xmm1,[eax]                      ; Read A Data

        mov     eax,phD                         ; Vector Destination
        pmulhw  xmm0,xmm1

        movdqu  [eax],xmm0                      ; Write D upper 16bits
        ret
vmp_pmulhHAsmSSE20 endp


;_________________________________________________________________________________________________
;       *** SSE2 version *** 128bit aligned

        align   16
        .XMM
        public  vmp_pmulhHAsmSSE2
vmp_pmulhHAsmSSE2       proc    near phD:ptr, phA:ptr, phB:ptr

ASSERT_PTR16(phD)
ASSERT_PTR16(phA)
ASSERT_PTR16(phB)

        mov     edx,phB                         ; Vector B
        mov     eax,phA                         ; Vector A

        movdqa  xmm0,[edx]                      ; Read B Data
        movdqa  xmm1,[eax]                      ; Read A Data

        mov     eax,phD                         ; Vector Destination
        pmulhw  xmm0,xmm1

        movdqa  [eax],xmm0                      ; Write D upper 16bits
        ret
vmp_pmulhHAsmSSE2 endp



;_________________________________________________________________________________________________
;=================================================================================================
;       Packed unsigned Multiplication High(16bit) = 8x16bit
;
;void vmp_pmulhuHGeneric( uint16 * const phD, const uint16 * const phA, const uint16 * const phB )
;{
;       *(phD+0) = (uint16)((((uint)*(phA+0)) * ((uint)*(phB+0))) >> 16);
;       *(phD+1) = (uint16)((((uint)*(phA+1)) * ((uint)*(phB+1))) >> 16);
;       *(phD+2) = (uint16)((((uint)*(phA+2)) * ((uint)*(phB+2))) >> 16);
;       *(phD+3) = (uint16)((((uint)*(phA+3)) * ((uint)*(phB+3))) >> 16);
;       *(phD+4) = (uint16)((((uint)*(phA+4)) * ((uint)*(phB+4))) >> 16);
;       *(phD+5) = (uint16)((((uint)*(phA+5)) * ((uint)*(phB+5))) >> 16);
;       *(phD+6) = (uint16)((((uint)*(phA+6)) * ((uint)*(phB+6))) >> 16);
;       *(phD+7) = (uint16)((((uint)*(phA+7)) * ((uint)*(phB+7))) >> 16);
;}                                                                      
        align   16
        .XMM
        public  vmp_pmulhuHAsmMMX2
        public  vmp_pmulhuHAsmSSE
vmp_pmulhuHAsmMMX2:
vmp_pmulhuHAsmSSE       proc    near phD:ptr, phA:ptr, phB:ptr
;;;        emms
ASSERT_PTR4(phD)
ASSERT_PTR4(phA)
ASSERT_PTR4(phB)

        mov     edx,phB                         ; Vector B
        mov     eax,phA                         ; Vector A
        mov     ecx,phD                         ; Vector Destination

        movq    mm0,[edx+0]                     ; Read B Data
        movq    mm1,[edx+8]
        movq    mm2,[eax+0]                     ; Read A Data
        movq    mm3,[eax+8]

        pmulhuw mm0,mm2
        pmulhuw mm1,mm3

        movq    [ecx+0],mm0                     ; Write D upper 16bits
        movq    [ecx+8],mm1                     ;
        
;;;        emms
        ret
vmp_pmulhuHAsmSSE endp


;_________________________________________________________________________________________________
;       *** SSE2 version *** 128bit unaligned

        align   16
        .XMM
        public  vmp_pmulhuHAsmSSE20
vmp_pmulhuHAsmSSE20       proc    near phD:ptr, phA:ptr, phB:ptr

ASSERT_PTR4(phD)
ASSERT_PTR4(phA)
ASSERT_PTR4(phB)

        mov     edx,phB                         ; Vector B
        mov     eax,phA                         ; Vector A

        movdqu  xmm0,[edx]                      ; Read B Data
        movdqu  xmm1,[eax]                      ; Read A Data

        mov     eax,phD                         ; Vector Destination
        pmulhuw xmm0,xmm1

        movdqu  [eax],xmm0                      ; Write D upper 16bits
        ret
vmp_pmulhuHAsmSSE20 endp


;_________________________________________________________________________________________________
;       *** SSE2 version *** 128bit aligned

        align   16
        .XMM
        public  vmp_pmulhuHAsmSSE2
vmp_pmulhuHAsmSSE2       proc    near phD:ptr, phA:ptr, phB:ptr

ASSERT_PTR16(phD)
ASSERT_PTR16(phA)
ASSERT_PTR16(phB)

        mov     edx,phB                         ; Vector B
        mov     eax,phA                         ; Vector A

        movdqa  xmm0,[edx]                      ; Read B Data
        movdqa  xmm1,[eax]                      ; Read A Data

        mov     eax,phD                         ; Vector Destination
        pmulhuw xmm0,xmm1

        movdqa  [eax],xmm0                      ; Write D upper 16bits
        ret
vmp_pmulhuHAsmSSE2 endp



;_________________________________________________________________________________________________
;=================================================================================================
;       Packed (un)signed Multiplication Low(16bit) = 8x16bit
;
;void vmp_pmulluHGeneric( uint16 * const phD, const uint16 * const phA, const uint16 * const phB )
;{
;       *(phD+0) = (uint16)((((uint)*(phA+0)) * ((uint)*(phB+0))) & 0x0000ffff);
;       *(phD+1) = (uint16)((((uint)*(phA+1)) * ((uint)*(phB+1))) & 0x0000ffff);
;       *(phD+2) = (uint16)((((uint)*(phA+2)) * ((uint)*(phB+2))) & 0x0000ffff);
;       *(phD+3) = (uint16)((((uint)*(phA+3)) * ((uint)*(phB+3))) & 0x0000ffff);
;       *(phD+4) = (uint16)((((uint)*(phA+4)) * ((uint)*(phB+4))) & 0x0000ffff);
;       *(phD+5) = (uint16)((((uint)*(phA+5)) * ((uint)*(phB+5))) & 0x0000ffff);
;       *(phD+6) = (uint16)((((uint)*(phA+6)) * ((uint)*(phB+6))) & 0x0000ffff);
;       *(phD+7) = (uint16)((((uint)*(phA+7)) * ((uint)*(phB+7))) & 0x0000ffff);
;}                                                                      
        align   16
        .MMX
        public  vmp_pmulluHAsmMMX
vmp_pmulluHAsmMMX       proc    near phD:ptr, phA:ptr, phB:ptr
;;;        emms
ASSERT_PTR4(phD)
ASSERT_PTR4(phA)
ASSERT_PTR4(phB)

        mov     edx,phB                         ; Vector B
        mov     eax,phA                         ; Vector A
        mov     ecx,phD                         ; Vector Destination

        movq    mm0,[edx+0]                     ; Read B Data
        movq    mm1,[edx+8]
        movq    mm2,[eax+0]                     ; Read A Data
        movq    mm3,[eax+8]

        pmullw  mm0,mm2
        pmullw  mm1,mm3

        movq    [ecx+0],mm0                     ; Write D upper 16bits
        movq    [ecx+8],mm1                     ;
        
;;;        emms
        ret
vmp_pmulluHAsmMMX endp



;_________________________________________________________________________________________________
;       *** SSE2 version *** 128bit unaligned

        align   16
        .XMM
        public  vmp_pmulluHAsmSSE20
vmp_pmulluHAsmSSE20       proc    near phD:ptr, phA:ptr, phB:ptr

ASSERT_PTR4(phD)
ASSERT_PTR4(phA)
ASSERT_PTR4(phB)

        mov     edx,phB                         ; Vector B
        mov     eax,phA                         ; Vector A

        movdqu  xmm0,[edx]                      ; Read B Data
        movdqu  xmm1,[eax]                      ; Read A Data

        mov     eax,phD                         ; Vector Destination
        pmullw  xmm0,xmm1

        movdqu  [eax],xmm0                      ; Write D upper 16bits
        ret
vmp_pmulluHAsmSSE20 endp


;_________________________________________________________________________________________________
;       *** SSE2 version *** 128bit aligned

        align   16
        .XMM
        public  vmp_pmulluHAsmSSE2
vmp_pmulluHAsmSSE2       proc    near phD:ptr, phA:ptr, phB:ptr

ASSERT_PTR16(phD)
ASSERT_PTR16(phA)
ASSERT_PTR16(phB)

        mov     edx,phB                         ; Vector B
        mov     eax,phA                         ; Vector A

        movdqa  xmm0,[edx]                      ; Read B Data
        movdqa  xmm1,[eax]                      ; Read A Data

        mov     eax,phD                         ; Vector Destination
        pmullw  xmm0,xmm1

        movdqa  [eax],xmm0                      ; Write D upper 16bits
        ret
vmp_pmulluHAsmSSE2 endp



;_________________________________________________________________________________________________
;=================================================================================================
;       Packed signed Multiplication Hi/Lo(2x16bit) = 8x16bit
;
;void vmp_pmulhlHGeneric( int16 * const phHi, int16 * const phLo, const int16 * const phA, const int16 * const phB )
;{
;       int hilo[8];
;
;       hilo[0] = ((int)*(phA+0)) * ((int)*(phB+0));
;       hilo[1] = ((int)*(phA+1)) * ((int)*(phB+1));
;       hilo[2] = ((int)*(phA+2)) * ((int)*(phB+2));
;       hilo[3] = ((int)*(phA+3)) * ((int)*(phB+3));
;       hilo[4] = ((int)*(phA+4)) * ((int)*(phB+4));
;       hilo[5] = ((int)*(phA+5)) * ((int)*(phB+5));
;       hilo[6] = ((int)*(phA+6)) * ((int)*(phB+6));
;       hilo[7] = ((int)*(phA+7)) * ((int)*(phB+7));
;
;       *(phHi+0) = (int16)(hilo[0] >> 16);
;       *(phHi+1) = (int16)(hilo[1] >> 16);
;       *(phHi+2) = (int16)(hilo[2] >> 16);
;       *(phHi+3) = (int16)(hilo[3] >> 16);
;       *(phHi+4) = (int16)(hilo[4] >> 16);
;       *(phHi+5) = (int16)(hilo[5] >> 16);
;       *(phHi+6) = (int16)(hilo[6] >> 16);
;       *(phHi+7) = (int16)(hilo[7] >> 16);
;
;       *(phLo+0) = (int16)(hilo[0] & 0x0000ffff);
;       *(phLo+1) = (int16)(hilo[1] & 0x0000ffff);
;       *(phLo+2) = (int16)(hilo[2] & 0x0000ffff);
;       *(phLo+3) = (int16)(hilo[3] & 0x0000ffff);
;       *(phLo+4) = (int16)(hilo[4] & 0x0000ffff);
;       *(phLo+5) = (int16)(hilo[5] & 0x0000ffff);
;       *(phLo+6) = (int16)(hilo[6] & 0x0000ffff);
;       *(phLo+7) = (int16)(hilo[7] & 0x0000ffff);
;}                                                                      
        align   16
        .MMX
        public  vmp_pmulhlHAsmMMX
vmp_pmulhlHAsmMMX       proc    near phHi:ptr, phLo:ptr, phA:ptr, phB:ptr
;;;        emms
ASSERT_PTR4(phHi)
ASSERT_PTR4(phLo)
ASSERT_PTR4(phA)
ASSERT_PTR4(phB)

        mov     edx,phB                         ; Vector B
        mov     eax,phA                         ; Vector A

        movq    mm0,[edx+0]                     ; Read B Data
        movq    mm1,[edx+8]
        movq    mm2,[eax+0]                     ; Read A Data
        movq    mm3,[eax+8]

        mov     edx,phHi
        mov     eax,phLo

        movq    mm4,mm0
        movq    mm5,mm1

        pmulhw  mm0,mm2                         ; upper 16bits
        pmulhw  mm1,mm3
        pmullw  mm2,mm4                         ; lower 16bits
        pmullw  mm3,mm5

        movq    [edx+0],mm0                     ; Write Upper 16bits
        movq    [edx+8],mm1                     ;
        movq    [eax+0],mm2                     ; Write Lower 16bits    
        movq    [eax+8],mm3                     ;

;;;        emms
        ret
vmp_pmulhlHAsmMMX endp


;_________________________________________________________________________________________________
;       *** SSE2 version *** 128bit unaligned

        align   16
        .XMM
        public  vmp_pmulhlHAsmSSE20
vmp_pmulhlHAsmSSE20       proc    near phHi:ptr, phLo:ptr, phA:ptr, phB:ptr

ASSERT_PTR4(phA)
ASSERT_PTR4(phB)
ASSERT_PTR4(phLo)
ASSERT_PTR4(phHi)

        mov     eax,phA                         ; Vector A
        mov     edx,phB                         ; Vector B

        movdqu  xmm0,[eax]                      ; Read A Data
        movdqu  xmm1,[edx]                      ; Read B Data

        mov     eax,phLo
        mov     edx,phHi                                ; Vector Destination

        movdqa  xmm2,xmm0

        pmullw  xmm0,xmm1                       ; Lower 16bits
        pmulhw  xmm1,xmm2                       ; Upper 16bits (signed)

        movdqu  [eax],xmm0                      ; Write D lower 16bits
        movdqu  [edx],xmm1                      ;         upper 16bits
        ret
vmp_pmulhlHAsmSSE20 endp


;_________________________________________________________________________________________________
;       *** SSE2 version *** 128bit aligned

        align   16
        .XMM
        public  vmp_pmulhlHAsmSSE2
vmp_pmulhlHAsmSSE2       proc    near phHi:ptr, phLo:ptr, phA:ptr, phB:ptr

ASSERT_PTR16(phA)
ASSERT_PTR16(phB)
ASSERT_PTR16(phLo)
ASSERT_PTR16(phHi)

        mov     eax,phA                         ; Vector A
        mov     edx,phB                         ; Vector B

        movdqa  xmm0,[eax]                      ; Read A Data
        movdqa  xmm1,[edx]                      ; Read B Data

        mov     eax,phLo
        mov     edx,phHi                                ; Vector Destination

        movdqa  xmm2,xmm0

        pmullw  xmm0,xmm1                       ; Lower 16bits
        pmulhw  xmm1,xmm2                       ; Upper 16bits (signed)

        movdqa  [eax],xmm0                      ; Write D lower 16bits
        movdqa  [edx],xmm1                      ;         upper 16bits
        ret
vmp_pmulhlHAsmSSE2 endp




;_________________________________________________________________________________________________
;       Packed unsigned Multiplication Hi/Lo(2x16bit) = 8x16bit
;
;void vmp_pmulhluHGeneric( uint16 * const phHi, uint16 * const phLo, const uint16 * const phA, const uint16 * const phB )
;{
;       uint hilo[8];
;
;       hilo[0] = ((uint)*(phA+0)) * ((uint)*(phB+0));
;       hilo[1] = ((uint)*(phA+1)) * ((uint)*(phB+1));
;       hilo[2] = ((uint)*(phA+2)) * ((uint)*(phB+2));
;       hilo[3] = ((uint)*(phA+3)) * ((uint)*(phB+3));
;       hilo[4] = ((uint)*(phA+4)) * ((uint)*(phB+4));
;       hilo[5] = ((uint)*(phA+5)) * ((uint)*(phB+5));
;       hilo[6] = ((uint)*(phA+6)) * ((uint)*(phB+6));
;       hilo[7] = ((uint)*(phA+7)) * ((uint)*(phB+7));
;
;       *(phHi+0) = (uint16)(hilo[0] >> 16);
;       *(phHi+1) = (uint16)(hilo[1] >> 16);
;       *(phHi+2) = (uint16)(hilo[2] >> 16);
;       *(phHi+3) = (uint16)(hilo[3] >> 16);
;       *(phHi+4) = (uint16)(hilo[4] >> 16);
;       *(phHi+5) = (uint16)(hilo[5] >> 16);
;       *(phHi+6) = (uint16)(hilo[6] >> 16);
;       *(phHi+7) = (uint16)(hilo[7] >> 16);
;
;       *(phLo+0) = (uint16)(hilo[0] & 0x0000ffff);
;       *(phLo+1) = (uint16)(hilo[1] & 0x0000ffff);
;       *(phLo+2) = (uint16)(hilo[2] & 0x0000ffff);
;       *(phLo+3) = (uint16)(hilo[3] & 0x0000ffff);
;       *(phLo+4) = (uint16)(hilo[4] & 0x0000ffff);
;       *(phLo+5) = (uint16)(hilo[5] & 0x0000ffff);
;       *(phLo+6) = (uint16)(hilo[6] & 0x0000ffff);
;       *(phLo+7) = (uint16)(hilo[7] & 0x0000ffff);
;}                                                                      
        align   16
        .XMM
        public vmp_pmulhluHAsmMMX2
        public  vmp_pmulhluHAsmSSE
vmp_pmulhluHAsmMMX2:
vmp_pmulhluHAsmSSE       proc    near phHi:ptr, phLo:ptr, phA:ptr, phB:ptr
;;;        emms
ASSERT_PTR4(phHi)
ASSERT_PTR4(phLo)
ASSERT_PTR4(phA)
ASSERT_PTR4(phB)

        mov     edx,phB                         ; Vector B
        mov     eax,phA                         ; Vector A

        movq    mm0,[edx+0]                     ; Read B Data
        movq    mm1,[edx+8]
        movq    mm2,[eax+0]                     ; Read A Data
        movq    mm3,[eax+8]
        mov     edx,phHi
        mov     eax,phLo

        movq    mm4,mm0
        movq    mm5,mm1

        pmulhuw mm0,mm2         ;MMX+ SSE       ; upper 16bits
        pmulhuw mm1,mm3         ;MMX+ SSE
        pmullw  mm2,mm4                         ; lower 16bits
        pmullw  mm3,mm5

        movq    [edx+0],mm0                     ; Write Upper 16bits
        movq    [edx+8],mm1                     ;
        movq    [eax+0],mm2                     ; Write Lower 16bits    
        movq    [eax+8],mm3                     ;

;;;        emms
        ret
vmp_pmulhluHAsmSSE endp



;_________________________________________________________________________________________________
;       *** SSE2 version *** 128bit unaligned

        align   16
        .XMM
        public  vmp_pmulhluHAsmSSE20
vmp_pmulhluHAsmSSE20       proc    near phHi:ptr, phLo:ptr, phA:ptr, phB:ptr

ASSERT_PTR4(phA)
ASSERT_PTR4(phB)
ASSERT_PTR4(phLo)
ASSERT_PTR4(phHi)

        mov     eax,phA                         ; Vector A
        mov     edx,phB                         ; Vector B

        movdqu  xmm0,[eax]                      ; Read A Data
        movdqu  xmm1,[edx]                      ; Read B Data

        mov     eax,phLo
        mov     edx,phHi                                ; Vector Destination

        movdqa  xmm2,xmm0

        pmullw  xmm0,xmm1                       ; Lower 16bits
        pmulhuw xmm1,xmm2                       ; Upper 16bits (unsigned)

        movdqu  [eax],xmm0                      ; Write D lower 16bits
        movdqu  [edx],xmm1                      ;         upper 16bits
        ret
vmp_pmulhluHAsmSSE20 endp


;_________________________________________________________________________________________________
;       *** SSE2 version *** 128bit aligned

        align   16
        .XMM
        public  vmp_pmulhluHAsmSSE2
vmp_pmulhluHAsmSSE2       proc    near phHi:ptr, phLo:ptr, phA:ptr, phB:ptr

ASSERT_PTR16(phA)
ASSERT_PTR16(phB)
ASSERT_PTR16(phLo)
ASSERT_PTR16(phHi)

        mov     eax,phA                         ; Vector A
        mov     edx,phB                         ; Vector B

        movdqa  xmm0,[eax]                      ; Read A Data
        movdqa  xmm1,[edx]                      ; Read B Data

        mov     eax,phLo
        mov     edx,phHi                                ; Vector Destination

        movdqa  xmm2,xmm0

        pmullw  xmm0,xmm1                       ; Lower 16bits
        pmulhuw xmm1,xmm2                       ; Upper 16bits (unsigned)

        movdqa  [eax],xmm0                      ; Write D lower 16bits
        movdqa  [edx],xmm1                      ;         upper 16bits
        ret
vmp_pmulhluHAsmSSE2 endp






;_________________________________________________________________________________________________
;=================================================================================================
;       Packed unsigned Multiplication 32bit Even = 8x16bit
;
;void vmp_pmullueHGeneric( uint32 * const pwD, const uint16 * const phA, const uint16 * const phB )
;{
;       *(pwD+0) = ((uint)*(phA+0)) * ((uint)*(phB+0));
;       *(pwD+1) = ((uint)*(phA+2)) * ((uint)*(phB+2));
;       *(pwD+2) = ((uint)*(phA+4)) * ((uint)*(phB+4));
;       *(pwD+3) = ((uint)*(phA+6)) * ((uint)*(phB+6));
;}                                                                      
        align   16
        .XMM
        public vmp_pmullueHAsmMMX2
        public  vmp_pmullueHAsmSSE
vmp_pmullueHAsmMMX2:
vmp_pmullueHAsmSSE       proc    near pwD:ptr, phA:ptr, phB:ptr
;;;        emms
ASSERT_PTR4(pwD)
ASSERT_PTR4(phA)
ASSERT_PTR4(phB)

        mov     edx,phB                         ; Vector B
        mov     eax,phA                         ; Vector A
        mov     ecx,pwD

        movq    mm0,[edx+0]                     ; B3 B2 B1 B0 - Read B Data
        movq    mm3,[edx+8]                     ; B7 B6 B5 B4
        movq    mm2,[eax+0]                     ; A3 A2 A1 A0 - Read A Data
        movq    mm1,[eax+8]                     ; A7 A6 A5 A4

        movq    mm4,mm0
        movq    mm5,mm2

        punpcklwd mm0,mm3                               ; B5 B1 B4 B0
        punpckhwd mm4,mm3                               ; B7 B3 B6 B2
        punpcklwd mm2,mm1                               ; A5 A1 A4 A0
        punpckhwd mm5,mm1                               ; A7 A3 A6 A2

        punpcklwd mm0,mm4                               ; B6 B4 B2 B0
        punpcklwd mm2,mm5                               ; A6 A4 A2 A0

        movq    mm1,mm0
        pmullw  mm0,mm2                         ; A6*B6 A4*B4 A2*B2 A0*B0 lower 16bits
        pmulhuw mm2,mm1         ;MMX+ SSE       ; A6*B6 A4*B4 A2*B2 A0*B0 upper 16bits (unsigned)

        movq    mm1,mm0
        punpcklwd mm0,mm2                               ; H_a2b2 l_a2b2 h_a0b0 l_a0b0
        punpckhwd mm1,mm2                               ; h_a6b6 l_a6b6 h_a4b4 l_a4b4

        movq    [ecx+0],mm0                     ; Write 4 32bit
        movq    [ecx+8],mm1                     ;

;;;        emms
        ret
vmp_pmullueHAsmSSE endp



;_________________________________________________________________________________________________
;       *** SSE2 version *** 128bit unaligned

        align   16
        .XMM
        public  vmp_pmullueHAsmSSE20
vmp_pmullueHAsmSSE20       proc    near phD:ptr, phA:ptr, phB:ptr

ASSERT_PTR4(phD)
ASSERT_PTR4(phA)
ASSERT_PTR4(phB)

        mov     edx,phB                         ; Vector B
        mov     eax,phA                         ; Vector A

        movdqu  xmm0,[eax]                      ; A7 A6 A5 A4 A3 A2 A1 A0 - Read A Data
        movdqu  xmm1,[edx]                      ; B7 B6 B5 B4 B3 B2 B1 B0 - Read B Data

        movdqa  xmm2,xmm0
        pmullw  xmm0,xmm1                       ;  A7*B7 A6*B6 ... A1*B1 A0*B0 lower 16bits
        pmulhuw xmm1,xmm2                       ;  A7*B7 A6*B6 ... A1*B1 A0*B0 upper 16bits (unsigned)

        pand    xmm0,lmask                      ;   0  l_A6B6 ...   0  l_A0B0
        pslld   xmm1,16                         ;h_A6B6  0    ... h_A0B0 0                              ; 

        mov     eax,phD                         ; Vector Destination
        por     xmm0,xmm1                       

        movdqu  [eax],xmm0                      ; Write D 32bits  {even}
        ret
vmp_pmullueHAsmSSE20 endp


;_________________________________________________________________________________________________
;       *** SSE2 version *** 128bit aligned

        align   16
        .XMM
        public  vmp_pmullueHAsmSSE2
vmp_pmullueHAsmSSE2       proc    near phD:ptr, phA:ptr, phB:ptr

ASSERT_PTR16(phD)
ASSERT_PTR16(phA)
ASSERT_PTR16(phB)

        mov     edx,phB                         ; Vector B
        mov     eax,phA                         ; Vector A

        movdqa  xmm0,[eax]                      ; A7 A6 A5 A4 A3 A2 A1 A0 - Read A Data
        movdqa  xmm1,[edx]                      ; B7 B6 B5 B4 B3 B2 B1 B0 - Read B Data

        movdqa  xmm2,xmm0
        pmullw  xmm0,xmm1                       ;  A7*B7 A6*B6 ... A1*B1 A0*B0 lower 16bits
        pmulhuw xmm1,xmm2                       ;  A7*B7 A6*B6 ... A1*B1 A0*B0 upper 16bits (unsigned)

        pand    xmm0,lmask                      ;   0  l_A6B6 ...   0  l_A0B0
        pslld   xmm1,16                         ;h_A6B6  0    ... h_A0B0 0                              ; 

        mov     eax,phD                         ; Vector Destination
        por     xmm0,xmm1                       

        movdqa  [eax],xmm0                      ; Write D 32bits  {even}
        ret
vmp_pmullueHAsmSSE2 endp



;_________________________________________________________________________________________________
;=================================================================================================
;       Packed unsigned Multiplication 32bit Odd = 8x16bit
;
;void vmp_pmulluoHGeneric( uint32 * const pwD, const uint16 * const phA, const uint16 * const phB )
;{
;       *(pwD+0) = ((uint)*(phA+1)) * ((uint)*(phB+1));
;       *(pwD+1) = ((uint)*(phA+3)) * ((uint)*(phB+3));
;       *(pwD+2) = ((uint)*(phA+5)) * ((uint)*(phB+5));
;       *(pwD+3) = ((uint)*(phA+7)) * ((uint)*(phB+7));
;}                                                                      
        align   16
        .XMM
        public vmp_pmulluoHAsmMMX2
        public  vmp_pmulluoHAsmSSE
vmp_pmulluoHAsmMMX2:
vmp_pmulluoHAsmSSE       proc    near pwD:ptr, phA:ptr, phB:ptr
;;;        emms
ASSERT_PTR4(pwD)
ASSERT_PTR4(phA)
ASSERT_PTR4(phB)

        mov     edx,phB                         ; Vector B
        mov     eax,phA                         ; Vector A
        mov     ecx,pwD

        movq    mm0,[edx+0]                     ; B3 B2 B1 B0 - Read B Data
        movq    mm3,[edx+8]                     ; B7 B6 B5 B4
        movq    mm2,[eax+0]                     ; A3 A2 A1 A0 - Read A Data
        movq    mm1,[eax+8]                     ; A7 A6 A5 A4

        movq    mm4,mm0
        movq    mm5,mm2

        punpcklwd mm0,mm3                               ; B5 B1 B4 B0
        punpckhwd mm4,mm3                               ; B7 B3 B6 B2
        punpcklwd mm2,mm1                               ; A5 A1 A4 A0
        punpckhwd mm5,mm1                               ; A7 A3 A6 A2

        punpckhwd mm0,mm4                               ; B7 B5 B3 B1
        punpckhwd mm2,mm5                               ; A7 A5 A3 A1

        movq    mm1,mm0
        pmullw  mm0,mm2                         ; A7*B7 A5*B5 A3*B3 A1*B1 lower 16bits
        pmulhuw mm2,mm1         ;MMX+ SSE       ; A7*B7 A5*B5 A3*B3 A1*B1 upper 16bits (unsigned)

        movq    mm1,mm0
        punpcklwd mm0,mm2                               ; H_a3b3 l_a3b3 h_a1b1 l_a1b1
        punpckhwd mm1,mm2                               ; h_a7b7 l_a7b7 h_a5b5 l_a5b5

        movq    [ecx+0],mm0                     ;
        movq    [ecx+8],mm1                     ;

;;;        emms
        ret
vmp_pmulluoHAsmSSE endp



;_________________________________________________________________________________________________
;       *** SSE2 version *** 128bit unaligned

        align   16
        .XMM
        public  vmp_pmulluoHAsmSSE20
vmp_pmulluoHAsmSSE20       proc    near phD:ptr, phA:ptr, phB:ptr

ASSERT_PTR4(phD)
ASSERT_PTR4(phA)
ASSERT_PTR4(phB)

        mov     edx,phB                         ; Vector B
        mov     eax,phA                         ; Vector A

        movdqu  xmm0,[eax]                      ; A7 A6 A5 A4 A3 A2 A1 A0 - Read A Data
        movdqu  xmm1,[edx]                      ; B7 B6 B5 B4 B3 B2 B1 B0 - Read B Data

        movdqa  xmm2,xmm0
        pmullw  xmm0,xmm1                       ;  A7*B7 A6*B6 ... A1*B1 A0*B0 lower 16bits
        pmulhuw xmm1,xmm2                       ;  A7*B7 A6*B6 ... A1*B1 A0*B0 upper 16bits (unsigned)

        psrld   xmm0,16                         ;   0  l_A7B7 ...   0  l_A1B1
        pand    xmm1,lmask                      ;h_A7B7  0    ... h_A1B1 0                              ; 

        mov     eax,phD                         ; Vector Destination
        por     xmm0,xmm1                       

        movdqu  [eax],xmm0                      ; Write D 32bits  {even}
        ret
vmp_pmulluoHAsmSSE20 endp


;_________________________________________________________________________________________________
;       *** SSE2 version *** 128bit aligned

        align   16
        .XMM
        public  vmp_pmulluoHAsmSSE2
vmp_pmulluoHAsmSSE2       proc    near phD:ptr, phA:ptr, phB:ptr

ASSERT_PTR16(phD)
ASSERT_PTR16(phA)
ASSERT_PTR16(phB)

        mov     edx,phB                         ; Vector B
        mov     eax,phA                         ; Vector A

        movdqa  xmm0,[eax]                      ; A7 A6 A5 A4 A3 A2 A1 A0 - Read A Data
        movdqa  xmm1,[edx]                      ; B7 B6 B5 B4 B3 B2 B1 B0 - Read B Data

        movdqa  xmm2,xmm0
        pmullw  xmm0,xmm1                       ;  A7*B7 A6*B6 ... A1*B1 A0*B0 lower 16bits
        pmulhuw xmm1,xmm2                       ;  A7*B7 A6*B6 ... A1*B1 A0*B0 upper 16bits (unsigned)

        psrld   xmm0,16                         ;   0  l_A7B7 ...   0  l_A1B1
        pand    xmm1,lmask                      ;h_A7B7  0    ... h_A1B1 0                              ; 

        mov     eax,phD                         ; Vector Destination
        por     xmm0,xmm1                       

        movdqa  [eax],xmm0                      ; Write D 32bits  {even}
        ret
vmp_pmulluoHAsmSSE2 endp



;_________________________________________________________________________________________________
;=================================================================================================
;       Packed signed Multiplication 32bit Even = 8x16bit
;
;void vmp_pmulleHGeneric( int32 * const pwD, const int16 * const phA, const int16 * const phB )
;{
;       *(pwD+0) = ((int)*(phA+0)) * ((int)*(phB+0));
;       *(pwD+1) = ((int)*(phA+2)) * ((int)*(phB+2));
;       *(pwD+2) = ((int)*(phA+4)) * ((int)*(phB+4));
;       *(pwD+3) = ((int)*(phA+6)) * ((int)*(phB+6));
;}                                                                      
        align   16
        .MMX
        public  vmp_pmulleHAsmMMX
vmp_pmulleHAsmMMX       proc    near pwD:ptr, phA:ptr, phB:ptr
;;;        emms
ASSERT_PTR4(pwD)
ASSERT_PTR4(phA)
ASSERT_PTR4(phB)

        mov     edx,phB                         ; Vector B
        mov     eax,phA                         ; Vector A
        mov     ecx,pwD

        movq    mm0,[edx+0]                     ; B3 B2 B1 B0 - Read B Data
        movq    mm3,[edx+8]                     ; B7 B6 B5 B4
        movq    mm2,[eax+0]                     ; A3 A2 A1 A0 - Read A Data
        movq    mm1,[eax+8]                     ; A7 A6 A5 A4

        movq    mm4,mm0
        movq    mm5,mm2

        punpcklwd mm0,mm3                               ; B5 B1 B4 B0
        punpckhwd mm4,mm3                               ; B7 B3 B6 B2
        punpcklwd mm2,mm1                               ; A5 A1 A4 A0
        punpckhwd mm5,mm1                               ; A7 A3 A6 A2

        punpcklwd mm0,mm4                               ; B6 B4 B2 B0
        punpcklwd mm2,mm5                               ; A6 A4 A2 A0

        movq    mm1,mm0
        pmullw  mm0,mm2                         ; A6*B6 A4*B4 A2*B2 A0*B0 lower 16bits
        pmulhw  mm2,mm1                         ; A6*B6 A4*B4 A2*B2 A0*B0 upper 16bits (signed)

        movq    mm1,mm0
        punpcklwd mm0,mm2                               ; H_a2b2 l_a2b2 h_a0b0 l_a0b0
        punpckhwd mm1,mm2                               ; h_a6b6 l_a6b6 h_a4b4 l_a4b4

        movq    [ecx+0],mm0                     ; Write 4 32bit
        movq    [ecx+8],mm1                     ;

;;;        emms
        ret
vmp_pmulleHAsmMMX endp



;_________________________________________________________________________________________________
;       *** SSE2 version *** 128bit unaligned

        align   16
        .XMM
        public  vmp_pmulleHAsmSSE20
vmp_pmulleHAsmSSE20       proc    near phD:ptr, phA:ptr, phB:ptr

ASSERT_PTR4(phD)
ASSERT_PTR4(phA)
ASSERT_PTR4(phB)

        mov     edx,phB                         ; Vector B
        mov     eax,phA                         ; Vector A

        movdqu  xmm0,[eax]                      ; A7 A6 A5 A4 A3 A2 A1 A0 - Read A Data
        movdqu  xmm1,[edx]                      ; B7 B6 B5 B4 B3 B2 B1 B0 - Read B Data

        movdqa  xmm2,xmm0
        pmullw  xmm0,xmm1                       ;  A7*B7 A6*B6 ... A1*B1 A0*B0 lower 16bits
        pmulhw  xmm1,xmm2                       ;  A7*B7 A6*B6 ... A1*B1 A0*B0 upper 16bits (signed)

        pand    xmm0,lmask                      ;   0  l_A6B6 ...   0  l_A0B0
        pslld   xmm1,16                         ;h_A6B6  0    ... h_A0B0 0                              ; 

        mov     eax,phD                         ; Vector Destination
        por     xmm0,xmm1                       

        movdqu  [eax],xmm0                      ; Write D 32bits  {even}
        ret
vmp_pmulleHAsmSSE20 endp


;_________________________________________________________________________________________________
;       *** SSE2 version *** 128bit aligned

        align   16
        .XMM
        public  vmp_pmulleHAsmSSE2
vmp_pmulleHAsmSSE2       proc    near phD:ptr, phA:ptr, phB:ptr

ASSERT_PTR16(phD)
ASSERT_PTR16(phA)
ASSERT_PTR16(phB)

        mov     edx,phB                         ; Vector B
        mov     eax,phA                         ; Vector A

        movdqa  xmm0,[eax]                      ; A7 A6 A5 A4 A3 A2 A1 A0 - Read A Data
        movdqa  xmm1,[edx]                      ; B7 B6 B5 B4 B3 B2 B1 B0 - Read B Data

        movdqa  xmm2,xmm0
        pmullw  xmm0,xmm1                       ;  A7*B7 A6*B6 ... A1*B1 A0*B0 lower 16bits
        pmulhw  xmm1,xmm2                       ;  A7*B7 A6*B6 ... A1*B1 A0*B0 upper 16bits (signed)

        pand    xmm0,lmask                      ;   0  l_A6B6 ...   0  l_A0B0
        pslld   xmm1,16                         ;h_A6B6  0    ... h_A0B0 0                              ; 

        mov     eax,phD                         ; Vector Destination
        por     xmm0,xmm1                       

        movdqa  [eax],xmm0                      ; Write D 32bits  {even}
        ret
vmp_pmulleHAsmSSE2 endp


;_________________________________________________________________________________________________
;=================================================================================================
;       Packed signed Multiplication 32bit Odd = 8x16bit
;
;void vmp_pmulloHGeneric( int32 * const pwD, const int16 * const phA, const int16 * const phB )
;{
;       *(pwD+0) = ((int)*(phA+1)) * ((int)*(phB+1));
;       *(pwD+1) = ((int)*(phA+3)) * ((int)*(phB+3));
;       *(pwD+2) = ((int)*(phA+5)) * ((int)*(phB+5));
;       *(pwD+3) = ((int)*(phA+7)) * ((int)*(phB+7));
;}                                                                      
        align   16
        .MMX
        public  vmp_pmulloHAsmMMX
vmp_pmulloHAsmMMX       proc    near pwD:ptr, phA:ptr, phB:ptr
;;;        emms
ASSERT_PTR4(pwD)
ASSERT_PTR4(phA)
ASSERT_PTR4(phB)

        mov     edx,phB                         ; Vector B
        mov     eax,phA                         ; Vector A
        mov     ecx,pwD

        movq    mm0,[edx+0]                     ; B3 B2 B1 B0 - Read B Data
        movq    mm3,[edx+8]                     ; B7 B6 B5 B4
        movq    mm2,[eax+0]                     ; A3 A2 A1 A0 - Read A Data
        movq    mm1,[eax+8]                     ; A7 A6 A5 A4

        movq    mm4,mm0
        movq    mm5,mm2

        punpcklwd mm0,mm3                               ; B5 B1 B4 B0
        punpckhwd mm4,mm3                               ; B7 B3 B6 B2
        punpcklwd mm2,mm1                               ; A5 A1 A4 A0
        punpckhwd mm5,mm1                               ; A7 A3 A6 A2

        punpckhwd mm0,mm4                               ; B7 B5 B3 B1
        punpckhwd mm2,mm5                               ; A7 A5 A3 A1

        movq    mm1,mm0
        pmullw  mm0,mm2                         ; A7*B7 A5*B5 A3*B3 A1*B1 lower 16bits
        pmulhw  mm2,mm1                         ; A7*B7 A5*B5 A3*B3 A1*B1 upper 16bits (signed)

        movq    mm1,mm0
        punpcklwd mm0,mm2                               ; H_a3b3 l_a3b3 h_a1b1 l_a1b1
        punpckhwd mm1,mm2                               ; h_a7b7 l_a7b7 h_a5b5 l_a5b5

        movq    [ecx+0],mm0                     ;
        movq    [ecx+8],mm1                     ;

;;;        emms
        ret
vmp_pmulloHAsmMMX endp                                                           



;_________________________________________________________________________________________________
;       *** SSE2 version *** 128bit unaligned

        align   16
        .XMM
        public  vmp_pmulloHAsmSSE20
vmp_pmulloHAsmSSE20       proc    near phD:ptr, phA:ptr, phB:ptr

ASSERT_PTR4(phD)
ASSERT_PTR4(phA)
ASSERT_PTR4(phB)

        mov     edx,phB                         ; Vector B
        mov     eax,phA                         ; Vector A

        movdqu  xmm0,[eax]                      ; A7 A6 A5 A4 A3 A2 A1 A0 - Read A Data
        movdqu  xmm1,[edx]                      ; B7 B6 B5 B4 B3 B2 B1 B0 - Read B Data

        movdqa  xmm2,xmm0
        pmullw  xmm0,xmm1                       ;  A7*B7 A6*B6 ... A1*B1 A0*B0 lower 16bits
        pmulhw  xmm1,xmm2                       ;  A7*B7 A6*B6 ... A1*B1 A0*B0 upper 16bits (signed)

        psrld   xmm0,16                         ;   0  l_A7B7 ...   0  l_A1B1
        pand    xmm1,lmask                      ;h_A7B7  0    ... h_A1B1 0                              ; 

        mov     eax,phD                         ; Vector Destination
        por     xmm0,xmm1                       

        movdqu  [eax],xmm0                      ; Write D 32bits  {even}
        ret
vmp_pmulloHAsmSSE20 endp


;_________________________________________________________________________________________________
;       *** SSE2 version *** 128bit aligned

        align   16
        .XMM
        public  vmp_pmulloHAsmSSE2
vmp_pmulloHAsmSSE2       proc    near phD:ptr, phA:ptr, phB:ptr

ASSERT_PTR16(phD)
ASSERT_PTR16(phA)
ASSERT_PTR16(phB)

        mov     edx,phB                         ; Vector B
        mov     eax,phA                         ; Vector A

        movdqa  xmm0,[eax]                      ; A7 A6 A5 A4 A3 A2 A1 A0 - Read A Data
        movdqa  xmm1,[edx]                      ; B7 B6 B5 B4 B3 B2 B1 B0 - Read B Data

        movdqa  xmm2,xmm0
        pmullw  xmm0,xmm1                       ;  A7*B7 A6*B6 ... A1*B1 A0*B0 lower 16bits
        pmulhw  xmm1,xmm2                       ;  A7*B7 A6*B6 ... A1*B1 A0*B0 upper 16bits (signed)

        psrld   xmm0,16                         ;   0  l_A7B7 ...   0  l_A1B1
        pand    xmm1,lmask                      ;h_A7B7  0    ... h_A1B1 0                              ; 

        mov     eax,phD                         ; Vector Destination
        por     xmm0,xmm1                       

        movdqa  [eax],xmm0                      ; Write D 32bits  {even}
        ret
vmp_pmulloHAsmSSE2 endp




PMulX86  ends
        end

