        TITLE   Vmd3DX86.asm - 80(x)86 3D Math Multiplication & Division
        PAGE    53,160

;  Vmd3D.asm - 3D Math Multiplication & Division
;
;  This document contains file that you may use at your own risk in your own programs.
;
;  This module is designed to accomodate the book.
;
;               This module is designed for fast 3D math
;
;  Created - 3 Oct 2001 - James Leiterman
;
;  Last Modification
;
;       Tabs = 8



        .XLIST

;;;        .387
;;;     .586P   .586    .686   .686P
;;;      .K3D    .XMM    .MMX

        .686
        .XMM

        .model flat, C


        include KariType.inc
        include vmp3DX86M.inc
 
        .LIST


        .data
        align   16

himsk32 DWORD   000000000h,000000000h,000000000h,0FFFFFFFFh
lomsk96 DWORD   0FFFFFFFFh,0FFFFFFFFh,0FFFFFFFFh,000000000h

;------------------------------------------------------------------------------------------------


Vmd3DX86        SEGMENT USE32 PUBLIC 'CODE'

;-------------------------------------------------------------------------------------------------



;_________________________________________________________________________________________________
;=================================================================================================
;       Vector Product (Multiplication)
;
;void vmp_VecMul(vmp3DVector * const pvD, const vmp3DVector * const pvA, const vmp3DVector * const pvB)
;{
;       pvD->x = pvA->x * pvB->x;
;       pvD->y = pvA->y * pvB->y;
;       pvD->z = pvA->z * pvB->z;
;
;       NOTE: vA, vB most likely will be misaligned especially if a packed structure
;       such as VMP3DVector is used so each field must be loaded individually!

;_________________________________________________________________________________________________

; *** 3DNow version ***
;
; NOTE: If arguments not aligned, a misaligned memory stall of 1 clock will occur!
        align   16
        public  vmp_VecMulAsm3DNow
vmp_VecMulAsm3DNow       proc    near , vD:ptr, vA:ptr, vB:ptr
        .K3D
;;;        femms
ASSERT_PTR(vD)
ASSERT_PTR(vA)
ASSERT_PTR(vB)

        mov     edx,vB                          ; Vector B
        mov     ecx,vA                          ; Vector A
        mov     eax,vD                          ; Vector Destination

        movq    mm0,[edx]                       ;vB.xy   {By Bx}
        movq    mm1,[ecx]                       ;vA.xy   {Ay Ax}
        movd    mm2,(vmp3DVector PTR [edx]).z   ;vB.z0   {0 Bz}
        movd    mm3,(vmp3DVector PTR [ecx]).z   ;vA.z0   {0 Az}

        pfmul   mm1,mm0                         ;       {AyBy AxBx}
        pfmul   mm3,mm2                         ;       {0*0  AzBz}

        movq    [eax],mm1                       ;       {AyBy AxBx}
        movd    (vmp3DVector PTR [eax]).z,mm3   ;       {0    AzBz}
        
;;;        femms
        ret
vmp_VecMulAsm3DNow endp


;_________________________________________________________________________________________________


; *** SSE version *** non aligned
;
        align   16
        public  vmp_VecMulAsmSSE0
vmp_VecMulAsmSSE0       proc    near , vD:ptr, vA:ptr, vB:ptr
        .XMM
;;;        emms
ASSERT_PTR(vD)
ASSERT_PTR(vA)
ASSERT_PTR(vB)

        mov     eax,vD                          ; Vector Destination
        mov     edx,vB                          ; Vector B
        mov     ecx,vA                          ; Vector A

        movups  xmm2,[eax]
        movups  xmm1,[edx]                      ;vB.xyz# {# Bz By Bx}
        movups  xmm0,[ecx]                      ;vA.xyz# {# Az Ay Ax}
        andps   xmm1,OWORD PTR lomsk96          ;        {0 Bz By Bx}

        andps   xmm2,OWORD PTR himsk32          ;        {# 0  0  0}
        mulps   xmm0,xmm1                       ;        {0 AzBz AyBy AxBx}
                ; Sometimes leaves 80000000 (-0)
        andps   xmm0,OWORD PTR lomsk96          ;        {0  #  #  #}
        orps    xmm0,xmm2                       ;        {# AzBz AyBy AxBx}

        movups  [eax],xmm0                      ;        {# AzBz AyBy AxBx}     
        
;;;        emms
        ret
vmp_VecMulAsmSSE0 endp


;_________________________________________________________________________________________________

; *** SSE version *** 128bit aligned
;
        align   16
        public  vmp_VecMulAsmSSE
vmp_VecMulAsmSSE       proc    near , vD:ptr, vA:ptr, vB:ptr
        .XMM
;;;        emms
ASSERT_PTR16(vD)
ASSERT_PTR16(vA)
ASSERT_PTR16(vB)

        mov     eax,vD                          ; Vector Destination
        mov     edx,vB                          ; Vector B
        mov     ecx,vA                          ; Vector A

        movaps  xmm2,[eax]
        movaps  xmm1,[edx]                      ;vB.xyz# {# Bz By Bx}
        movaps  xmm0,[ecx]                      ;vA.xyz# {# Az Ay Ax}
        andps   xmm1,OWORD PTR lomsk96          ;        {0 Bz By Bx}

        andps   xmm2,OWORD PTR himsk32          ;        {# 0  0  0}
        mulps   xmm0,xmm1                       ;        {0 AzBz AyBy AxBx}
                ; Sometimes leaves 80000000 (-0)
        andps   xmm0,OWORD PTR lomsk96          ;        {0  #  #  #}
        orps    xmm0,xmm2                       ;        {# AzBz AyBy AxBx}

        movaps  [eax],xmm0                      ;        {# AzBz AyBy AxBx}     
        
;;;        emms
        ret
vmp_VecMulAsmSSE endp


;_________________________________________________________________________________________________
;=================================================================================================
;       Vector Product Sum (Multiplication-Add)
;
;void vmp_VecMAdd(vmp3DVector * const pvD, const vmp3DVector * const pvA, 
;          const vmp3DVector * const pvB, const vmp3DVector * const pvC)
;{
;       pvD->x = pvA->x * pvB->x + pvC->x;
;       pvD->y = pvA->y * pvB->y + pvC->y;
;       pvD->z = pvA->z * pvB->z + pvC->z;
;
;       NOTE: vA, vB most likely will be misaligned especially if a packed structure
;       such as VMP3DVector is used so each field must be loaded individually!

;_________________________________________________________________________________________________

; *** 3DNow version ***
;
; NOTE: If arguments not aligned, a misaligned memory stall of 1 clock will occur!
        align   16
        public  vmp_VecMAddAsm3DNow
vmp_VecMAddAsm3DNow       proc    near , vD:ptr, vA:ptr, vB:ptr, vC:ptr
        .K3D
;;;        femms
ASSERT_PTR(vD)
ASSERT_PTR(vA)
ASSERT_PTR(vB)
ASSERT_PTR(vC)

        mov     edx,vB                          ; Vector B
        mov     ecx,vA                          ; Vector A
        mov     eax,vC                          ; Vector C

        movq    mm0,[edx]                       ;vB.xy   {By Bx}
        movq    mm2,[ecx]                       ;vA.xy   {Ay Ax}
        movq    mm4,[eax]                       ;vC.xy   {Cy Cx}
        movd    mm1,(vmp3DVector PTR [edx]).z   ;vB.z0   {0 Bz}
        movd    mm3,(vmp3DVector PTR [ecx]).z   ;vA.z0   {0 Az}
        movd    mm5,(vmp3DVector PTR [eax]).z   ;vC.z0  {0 Cz}

        mov     eax,vD                          ; Vector Destination
        pfmul   mm0,mm2                         ;       {AyBy AxBx}
        pfmul   mm1,mm3                         ;       {0*0  AzBz}
        pfadd   mm0,mm4                         ;       {0    AzBz+Cz}
        pfadd   mm1,mm5                         ;       {AyBy+Cy AxBx+Cx}
        movq    [eax],mm0                       ;       {AyBy+Cy AxBx+Cx}
        movd    (vmp3DVector PTR [eax]).z,mm1   ;       {0       AzBz+Cz}
        
;;;        femms
        ret
vmp_VecMAddAsm3DNow endp


;_________________________________________________________________________________________________


; *** SSE version *** non aligned
;
        align   16
        public  vmp_VecMAddAsmSSE0
vmp_VecMAddAsmSSE0       proc    near , vD:ptr, vA:ptr, vB:ptr, vC:ptr
        .XMM
;;;        emms
ASSERT_PTR(vC)
ASSERT_PTR(vB)
ASSERT_PTR(vA)
ASSERT_PTR(vD)

        mov     eax,vC                          ; Vector C
        mov     edx,vB                          ; Vector B
        mov     ecx,vA                          ; Vector A

        movups  xmm2,[eax]                      ;vC.xyz# {# Cz Cy Cx}
        mov     eax,vD                          ; Vector Destination
        movups  xmm1,[edx]                      ;vB.xyz# {# Bz By Bx}
        movups  xmm0,[ecx]                      ;vA.xyz# {# Az Ay Ax}

        movups  xmm3,[eax]
        andps   xmm1,OWORD PTR lomsk96          ;        {0 Bz By Bx}
        andps   xmm3,OWORD PTR himsk32          ;        {# 0  0  0}

        mulps   xmm0,xmm1                       ;        {0 AzBz AyBy AxBx}
        addps   xmm0,xmm2                       ;       {Cw AzBz+Cz AyBy+Cy AxBx+Cx}
                ; Sometimes leaves 80000000 (-0)
        andps   xmm0,OWORD PTR lomsk96          ;        {0  #  #  #}
        orps    xmm0,xmm3                       ;        {# AzBz+Cz AyBy+Cy AxBx+Cx}

        movups  [eax],xmm0                      ;        {# AzBz+Cz AyBy+Cy AxBx+Cx}    
        
;;;        emms
        ret
vmp_VecMAddAsmSSE0 endp


;_________________________________________________________________________________________________

; *** SSE version *** 128bit aligned
;
        align   16
        public  vmp_VecMAddAsmSSE
vmp_VecMAddAsmSSE       proc    near , vD:ptr, vA:ptr, vB:ptr, vC:ptr
        .XMM
;;;        emms
ASSERT_PTR16(vC)
ASSERT_PTR16(vB)
ASSERT_PTR16(vA)
ASSERT_PTR16(vD)

        mov     eax,vC                          ; Vector C
        mov     edx,vB                          ; Vector B
        mov     ecx,vA                          ; Vector A

        movaps  xmm2,[eax]                      ;vC.xyz# {# Cz Cy Cx}
        mov     eax,vD                          ; Vector Destination
        movaps  xmm1,[edx]                      ;vB.xyz# {# Bz By Bx}
        movaps  xmm0,[ecx]                      ;vA.xyz# {# Az Ay Ax}

        movaps  xmm3,[eax]
        andps   xmm1,OWORD PTR lomsk96          ;        {0 Bz By Bx}
        andps   xmm3,OWORD PTR himsk32          ;        {# 0  0  0}

        mulps   xmm0,xmm1                       ;        {0 AzBz AyBy AxBx}
        addps   xmm0,xmm2                       ;       {Cw AzBz+Cz AyBy+Cy AxBx+Cx}
                ; Sometimes leaves 80000000 (-0)
        andps   xmm0,OWORD PTR lomsk96          ;        {0  #  #  #}
        orps    xmm0,xmm3                       ;        {# AzBz+Cz AyBy+Cy AxBx+Cx}

        movaps  [eax],xmm0                      ;        {# AzBz+Cz AyBy+Cy AxBx+Cx}    
        
;;;        emms
        ret
vmp_VecMAddAsmSSE endp


;_________________________________________________________________________________________________
;=================================================================================================
;       Vector Product Dif (Multiplication-Sub)
;
;void vmp_VecMSub(vmp3DVector * const pvD, const vmp3DVector * const pvA, 
;         const vmp3DVector * const pvB, const vmp3DVector * const pvC)
;{
;       pvD->x = pvA->x * pvB->x - pvC->x;
;       pvD->y = pvA->y * pvB->y - pvC->y;
;       pvD->z = pvA->z * pvB->z - pvC->z;
;
;       NOTE: vA, vB most likely will be misaligned especially if a packed structure
;       such as VMP3DVector is used so each field must be loaded individually!

;_________________________________________________________________________________________________

; *** 3DNow version ***
;
; NOTE: If arguments not aligned, a misaligned memory stall of 1 clock will occur!
        align   16
        public  vmp_VecMSubAsm3DNow
vmp_VecMSubAsm3DNow       proc    near , vD:ptr, vA:ptr, vB:ptr, vC:ptr
        .K3D
;;;        femms
ASSERT_PTR(vD)
ASSERT_PTR(vA)
ASSERT_PTR(vB)
ASSERT_PTR(vC)

        mov     edx,vB                          ; Vector B
        mov     ecx,vA                          ; Vector A
        mov     eax,vC                          ; Vector C

        movq    mm0,[edx]                       ;vB.xy   {By Bx}
        movq    mm2,[ecx]                       ;vA.xy   {Ay Ax}
        movq    mm4,[eax]                       ;vC.xy   {Cy Cx}
        movd    mm1,(vmp3DVector PTR [edx]).z   ;vB.z0   {0 Bz}
        movd    mm3,(vmp3DVector PTR [ecx]).z   ;vA.z0   {0 Az}
        movd    mm5,(vmp3DVector PTR [eax]).z   ;vC.z0  {0 Cz}

        mov     eax,vD                          ; Vector Destination
        pfmul   mm0,mm2                         ;       {AyBy AxBx}
        pfmul   mm1,mm3                         ;       {0*0  AzBz}
        pfsub   mm0,mm4                         ;       {0    AzBz-Cz}
        pfsub   mm1,mm5                         ;       {AyBy-Cy AxBx-Cx}
        movq    [eax],mm0                       ;       {AyBy-Cy AxBx-Cx}
        movd    (vmp3DVector PTR [eax]).z,mm1   ;       {0       AzBz-Cz}
        
;;;        femms
        ret
vmp_VecMSubAsm3DNow endp


;_________________________________________________________________________________________________


; *** SSE version *** non aligned
;
        align   16
        public  vmp_VecMSubAsmSSE0
vmp_VecMSubAsmSSE0       proc    near , vD:ptr, vA:ptr, vB:ptr, vC:ptr
        .XMM
;;;        emms
ASSERT_PTR(vC)
ASSERT_PTR(vB)
ASSERT_PTR(vA)
ASSERT_PTR(vD)

        mov     eax,vC                          ; Vector C
        mov     edx,vB                          ; Vector B
        mov     ecx,vA                          ; Vector A

        movups  xmm2,[eax]                      ;vC.xyz# {# Cz Cy Cx}
        mov     eax,vD                          ; Vector Destination
        movups  xmm1,[edx]                      ;vB.xyz# {# Bz By Bx}
        movups  xmm0,[ecx]                      ;vA.xyz# {# Az Ay Ax}

        movups  xmm3,[eax]
        andps   xmm1,OWORD PTR lomsk96          ;        {0 Bz By Bx}
        andps   xmm3,OWORD PTR himsk32          ;        {# 0  0  0}

        mulps   xmm0,xmm1                       ;        {0 AzBz AyBy AxBx}
        subps   xmm0,xmm2                       ;       {Cw AzBz-Cz AyBy-Cy AxBx-Cx}
                ; Sometimes leaves 80000000 (-0)
        andps   xmm0,OWORD PTR lomsk96          ;        {0  #  #  #}
        orps    xmm0,xmm3                       ;        {# AzBz-Cz AyBy-Cy AxBx-Cx}

        movups  [eax],xmm0                      ;        {# AzBz-Cz AyBy-Cy AxBx-Cx}    
        
;;;        emms
        ret
vmp_VecMSubAsmSSE0 endp


;_________________________________________________________________________________________________

; *** SSE version *** 128bit aligned
;
        align   16
        public  vmp_VecMSubAsmSSE
vmp_VecMSubAsmSSE       proc    near , vD:ptr, vA:ptr, vB:ptr, vC:ptr
        .XMM
;;;        emms
ASSERT_PTR16(vC)
ASSERT_PTR16(vB)
ASSERT_PTR16(vA)
ASSERT_PTR16(vD)

        mov     eax,vC                          ; Vector C
        mov     edx,vB                          ; Vector B
        mov     ecx,vA                          ; Vector A

        movaps  xmm2,[eax]                      ;vC.xyz# {# Cz Cy Cx}
        mov     eax,vD                          ; Vector Destination
        movaps  xmm1,[edx]                      ;vB.xyz# {# Bz By Bx}
        movaps  xmm0,[ecx]                      ;vA.xyz# {# Az Ay Ax}

        movaps  xmm3,[eax]
        andps   xmm1,OWORD PTR lomsk96          ;        {0 Bz By Bx}
        andps   xmm3,OWORD PTR himsk32          ;        {# 0  0  0}

        mulps   xmm0,xmm1                       ;        {0 AzBz AyBy AxBx}
        subps   xmm0,xmm2                       ;       {Cw AzBz-Cz AyBy-Cy AxBx-Cx}
                ; Sometimes leaves 80000000 (-0)
        andps   xmm0,OWORD PTR lomsk96          ;        {0  #  #  #}
        orps    xmm0,xmm3                       ;        {# AzBz-Cz AyBy-Cy AxBx-Cx}

        movaps  [eax],xmm0                      ;        {# AzBz-Cz AyBy-Cy AxBx-Cx}    
        
;;;        emms
        ret
vmp_VecMSubAsmSSE endp


;_________________________________________________________________________________________________
;=================================================================================================
;       Scale Vector (w/Scalar)
;
;void vmp_VecScale(vmp3DVector * const pvD, const vmp3DVector * const pvA, float fScalar)
;{
;       pvD->x = pvA->x * fScalar;
;       pvD->y = pvA->y * fScalar;
;       pvD->z = pvA->z * fScalar;

;
;       NOTE: vA, vB most likely will be misaligned especially if a packed structure
;       such as VMP3DVector is used so each field must be loaded individually!

;_________________________________________________________________________________________________

; *** 3DNow version ***
;
; NOTE: If arguments not aligned, a misaligned memory stall of 1 clock will occur!

        align   16
        public  vmp_VecScaleAsm3DNow
vmp_VecScaleAsm3DNow       proc    near , vD:ptr, vA:ptr, fScalar:REAL4
        .K3D
;;;        femms
ASSERT_PTR(vD)
ASSERT_PTR(vA)

        movd    mm0,fScalar                     ;         {0  s}
        mov     ecx,vA                          ; Vector A
        punpckldq mm0,mm0                               ;         {s  s}

        mov     eax,vD                          ; Vector Destination

        movq    mm1,[ecx]                       ;vA.xy     {Ay Ax}
        movd    mm2,(vmp3DVector PTR [ecx]).z   ;vA.z0     {0  Az}
                                                        
        pfmul   mm1,mm0                         ;         {Ays Axs}
        pfmul   mm2,mm0                         ;         {0*s Azs}

        movq    [eax],mm1                       ;         {Ays Axs}
        movd    (vmp3DVector PTR [eax]).z,mm2   ;         {0   Azs}
        
;;;        femms
        ret
vmp_VecScaleAsm3DNow endp


;_________________________________________________________________________________________________


; *** SSE version *** non aligned
;
        align   16
        public  vmp_VecScaleAsmSSE0
vmp_VecScaleAsmSSE0       proc    near , vD:ptr, vA:ptr, fScalar:REAL4
        .XMM
;;;        emms

ASSERT_PTR(vD)
ASSERT_PTR(vA)

        pxor    xmm1,xmm1                       ;        (0 0 0 0)
        mov     eax,vD                          ; Vector Destination
        mov     ecx,vA                          ; Vector A

        movss   xmm1,fScalar                    ;        {0 0 0 s}

        movups  xmm2,[eax]                      ; Get Dest {# # # #}
        movups  xmm0,[ecx]                      ;vA.xyz# {# Az Ay Ax}
        shufps  xmm1,xmm1,11000000b             ; 3 0 0 0         {0  s  s  s}

        andps   xmm2,OWORD PTR himsk32          ;        {# 0  0  0}

        mulps   xmm0,xmm1                       ;        {0 Azs Ays Axs}
                ; Sometimes leaves 80000000 (-0)
        andps   xmm0,OWORD PTR lomsk96    ;limit -0
        orps    xmm0,xmm2                       ;        {# Azs Ays Axs}

        movups  [eax],xmm0                      ;        {# Azs Ays Axs}        
        
;;;        emms
        ret
vmp_VecScaleAsmSSE0 endp


;_________________________________________________________________________________________________

; *** SSE version *** 128bit aligned
;
        align   16
        public  vmp_VecScaleAsmSSE
vmp_VecScaleAsmSSE       proc    near , vD:ptr, vA:ptr, fScalar:REAL4
        .XMM
;;;        emms
ASSERT_PTR16(vD)
ASSERT_PTR16(vA)

        pxor    xmm1,xmm1                       ;        (0 0 0 0)
        mov     eax,vD                          ; Vector Destination
        mov     ecx,vA                          ; Vector A

        movss   xmm1,fScalar                    ;        {0 0 0 s}

        movaps  xmm2,[eax]
        movaps  xmm0,[ecx]                      ;vA.xyz# {# Az Ay Ax}
        shufps  xmm1,xmm1,11000000b             ; 3 0 0 0         {0  s  s  s}

        andps   xmm2,OWORD PTR himsk32          ;        {# 0  0  0}

        mulps   xmm0,xmm1                       ;        {0 Azs Ays Axs}
                ; Sometimes leaves 80000000 (-0)
        andps   xmm0,OWORD PTR lomsk96          ;        {0  #  #  #}
        orps    xmm0,xmm2                       ;        {# Azs Ays Axs}

        movaps  [eax],xmm0                      ;        {# Azs Ays Axs}        
;;;        emms
        ret
vmp_VecScaleAsmSSE endp

;_________________________________________________________________________________________________
;=================================================================================================
;       Vector Division (w/Scalar)
;
;void vmp_VecDiv(vmp3DVector * const pvD, const vmp3DVector * const pvA, float fScalar)
;{
;       pvD->x = pvA->x * fScalar;
;       pvD->y = pvA->y * fScalar;
;       pvD->z = pvA->z * fScalar;

;
;       NOTE: vA, vB most likely will be misaligned especially if a packed structure
;       such as VMP3DVector is used so each field must be loaded individually!

;_________________________________________________________________________________________________

; *** 3DNow version *** Slower 24bit precision
;
; NOTE: If arguments not aligned, a misaligned memory stall of 1 clock will occur!

        align   16
        public  vmp_VecDivAsm3DNow
vmp_VecDivAsm3DNow       proc    near , vD:ptr, vA:ptr, fScalar:REAL4
        .K3D
;;;        femms
ASSERT_PTR(vD)
ASSERT_PTR(vA)
ASSERT_ZERO(fScalar)

        movd    mm2,fScalar                     ;         {0  s}
        mov     ecx,vA                          ; Vector A

        pfrcp   mm3,mm2                         ;         {1/s 1/s}  14bit

        mov     eax,vD                          ; Vector Destination
                ; start 24bit conversion
        punpckldq mm2,mm2                               ;         {s  s}                

        movq    mm0,[ecx]                       ;vA.xy     {Ay Ax}
        pfrcpit1        mm2,mm3                         ;         {1/s 1/s}

        movd    mm1,(vmp3DVector PTR [ecx]).z   ;vA.z0     {0  Az}
        pfrcpit2        mm2,mm3

        pfmul   mm0,mm2                         ;         {Ay/s Ax/s}
        pfmul   mm1,mm2                         ;         {0/s Az/s}

        movq    [eax],mm0                       ;         {Ay/s Ax/s}
        movd    (vmp3DVector PTR [eax]).z,mm1   ;         {0/s   Az/s}
        
;;;        femms
        ret
vmp_VecDivAsm3DNow endp


;_________________________________________________________________________________________________

; *** 3DNow version *** FAST *** 14bit precision
;
; NOTE: If arguments not aligned, a misaligned memory stall of 1 clock will occur!

        align   16
        public  vmp_VecDivFastAsm3DNow
vmp_VecDivFastAsm3DNow       proc    near , vD:ptr, vA:ptr, fScalar:REAL4
        .K3D
;;;        femms
ASSERT_PTR(vD)
ASSERT_PTR(vA)
ASSERT_ZERO(fScalar)

        movd    mm2,fScalar                     ;         {0  s}
        mov     ecx,vA                          ; Vector A

        pfrcp   mm2,mm2                         ;         {1/s 1/s}  14bit
        mov     eax,vD                          ; Vector Destination

        movq    mm0,[ecx]                       ;vA.xy     {Ay Ax}
        movd    mm1,(vmp3DVector PTR [ecx]).z   ;vA.z0     {0  Az}

        pfmul   mm0,mm2                         ;         {Ay/s Ax/s}
        pfmul   mm1,mm2                         ;         {0/s Az/s}

        movq    [eax],mm0                       ;         {Ay/s Ax/s}
        movd    (vmp3DVector PTR [eax]).z,mm1   ;         {0/s   Az/s}
        
;;;        femms
        ret
vmp_VecDivFastAsm3DNow endp


;_________________________________________________________________________________________________


; *** SSE version *** non aligned
;
        align   16
        public  vmp_VecDivAsmSSE0
vmp_VecDivAsmSSE0       proc    near , vD:ptr, vA:ptr, fScalar:REAL4
        .XMM
;;;        emms

ASSERT_PTR(vD)
ASSERT_PTR(vA)
ASSERT_ZERO(fScalar)

        mov     eax,vD                          ; Vector Destination
        mov     ecx,vA                          ; Vector A

        movss   xmm1,fScalar                    ;        {0 0 0 s}

        movups  xmm2,[eax]                      ; Get Dest {# # # #}
        movups  xmm0,[ecx]                      ;vA.xyz# {# Az Ay Ax}
        shufps  xmm1,xmm1,00000000b             ; 0 0 0 0         {s  s  s  s}

        andps   xmm2,OWORD PTR himsk32          ;        {# 0  0  0}

        divps   xmm0,xmm1                       ;        {0/s Az/s Ay/s Ax/s}
                ; Sometimes leaves 80000000 (-0)
        andps   xmm0,OWORD PTR lomsk96          ;limit -0
        orps    xmm0,xmm2                       ;        {# Az/s Ay/s Ax/s}

        movups  [eax],xmm0                      ;        {# Az/s Ay/s Ax/s}     
        
;;;        emms
        ret
vmp_VecDivAsmSSE0 endp


;_________________________________________________________________________________________________

; *** SSE version *** 128bit aligned
;
        align   16
        public  vmp_VecDivAsmSSE
vmp_VecDivAsmSSE       proc    near , vD:ptr, vA:ptr, fScalar:REAL4
        .XMM
;;;        emms
ASSERT_PTR16(vD)
ASSERT_PTR16(vA)
ASSERT_ZERO(fScalar)

        mov     eax,vD                          ; Vector Destination
        mov     ecx,vA                          ; Vector A

        movss   xmm1,fScalar                    ;        {0 0 0 s}

        movaps  xmm2,[eax]
        movaps  xmm0,[ecx]                      ;vA.xyz# {# Az Ay Ax}
        shufps  xmm1,xmm1,00000000b             ; 0 0 0 0         {s  s  s  s}

        andps   xmm2,OWORD PTR himsk32          ;        {# 0  0  0}

        divps   xmm0,xmm1                       ;        {0/s Az/s Ay/s Ax/s}
                ; Sometimes leaves 80000000 (-0)
        andps   xmm0,OWORD PTR lomsk96          ;limit -0
        orps    xmm0,xmm2                       ;        {# Az/s Ay/s Ax/s}

        movaps  [eax],xmm0                      ;        {# Az/s Ay/s Ax/s}     
;;;        emms
        ret
vmp_VecDivAsmSSE endp

;_________________________________________________________________________________________________


; *** SSE version *** non aligned
;
        align   16
        public  vmp_VecDivFastAsmSSE0
vmp_VecDivFastAsmSSE0       proc    near , vD:ptr, vA:ptr, fScalar:REAL4
        .XMM
;;;        emms

ASSERT_PTR(vD)
ASSERT_PTR(vA)
ASSERT_ZERO(fScalar)

        mov     eax,vD                          ; Vector Destination
        mov     ecx,vA                          ; Vector A

        movss   xmm1,fScalar                    ;        {0 0 0 s}

        movups  xmm2,[eax]                      ; Get Dest {# # # #}
        movups  xmm0,[ecx]                      ;vA.xyz# {# Az Ay Ax}
        shufps  xmm1,xmm1,00000000b             ; 0 0 0 0         {s  s  s  s}

        andps   xmm2,OWORD PTR himsk32          ;        {# 0  0  0}

        rcpps   xmm1,xmm1                       ;        {1/s 1/s 1/s 1/s}
        mulps   xmm0,xmm1                       ;        {#*1/s Az*1/s Ay*1/s Ax*1/s}

                ; Sometimes leaves 80000000 (-0)
        andps   xmm0,OWORD PTR lomsk96          ;limit -0
        orps    xmm0,xmm2                       ;        {# Az/s Ay/s Ax/s}

        movups  [eax],xmm0                      ;        {# Az/s Ay/s Ax/s}     
        
;;;        emms
        ret
vmp_VecDivFastAsmSSE0 endp


;_________________________________________________________________________________________________

; *** SSE version *** 128bit aligned
;
        align   16
        public  vmp_VecDivFastAsmSSE
vmp_VecDivFastAsmSSE       proc    near , vD:ptr, vA:ptr, fScalar:REAL4
        .XMM
;;;        emms
ASSERT_PTR16(vD)
ASSERT_PTR16(vA)
ASSERT_ZERO(fScalar)

        mov     eax,vD                          ; Vector Destination
        mov     ecx,vA                          ; Vector A

        movss   xmm1,fScalar                    ;        {0 0 0 s}

        movaps  xmm2,[eax]
        movaps  xmm0,[ecx]                      ;vA.xyz# {# Az Ay Ax}
        shufps  xmm1,xmm1,00000000b             ; 0 0 0 0         {s  s  s  s}

        andps   xmm2,OWORD PTR himsk32          ;        {# 0  0  0}

        rcpps   xmm1,xmm1                       ;        {1/s 1/s 1/s 1/s}
        mulps   xmm0,xmm1                       ;        {#*1/s Az*1/s Ay*1/s Ax*1/s}

                ; Sometimes leaves 80000000 (-0)
        andps   xmm0,OWORD PTR lomsk96          ;limit -0
        orps    xmm0,xmm2                       ;        {# Az/s Ay/s Ax/s}

        movaps  [eax],xmm0                      ;        {# Az/s Ay/s Ax/s}     
;;;        emms
        ret
vmp_VecDivFastAsmSSE endp

;_________________________________________________________________________________________________
;=================================================================================================
;       Dot Product
;
;void vmp_DotProduct(float * const pfD, const vmp3DVector * const vA, const vmp3DVector * const vB)
;{      
;       = vA->x * vB->x + vA->y * vB->y + vA->z * vB->z;   =AxBx+AyBy+AzBz
;
;       NOTE: vA, vB most likely will be misaligned especially if a packed structure
;       such as VMP3DVector is used so each field must be loaded individually!
;

;_________________________________________________________________________________________________

; *** 3DNow version ***
;
; NOTE: If arguments not aligned, a misaligned memory stall of 1 clock will occur!

        align   16
        public  vmp_DotProductAsm3DNow
vmp_DotProductAsm3DNow       proc    near , vD:ptr, vA:ptr, vB:ptr
        .K3D
;;;        femms
ASSERT_PTR(vD)
ASSERT_PTR(vA)
ASSERT_PTR(vB)

        mov     edx,vB                          ; Vector B
        mov     ecx,vA                          ; Vector A

        movd    mm0,(vmp3DVector PTR [edx]).z   ;vB.z   {0 Bz}
        movd    mm1,(vmp3DVector PTR [ecx]).z   ;vA.z   {0 Az}
        movq    mm2,[edx]                       ;vB.x   {By Bx}

        pfmul   mm0,mm1                         ;*vB.z  {00 BzAz}
        movq    mm3,[ecx]                       ;vA.x   {Ay Ax}

        mov     eax,vD                          ; Vector Destination
        pfmul   mm2,mm3                         ;       {ByAy BxAx}

        pfacc   mm2,mm2                         ;       {ByAy+BxAx   ByAy+BxAx}

        pfadd   mm0,mm2                         ;       {ByAy+BxAx+0 ByAy+BxAx+BzAz}

        movd    [eax],mm0                       ; Save resulting scalar
        
;;;        femms
        ret
vmp_DotProductAsm3DNow endp


;_________________________________________________________________________________________________

; *** NON-Aligned SSE version ***

        align   16
        public  vmp_DotProductAsmSSE0
vmp_DotProductAsmSSE0       proc    near , vD:ptr, vA:ptr, vB:ptr
        .XMM
;;;        emms
ASSERT_PTR(vD)
ASSERT_PTR(vA)
ASSERT_PTR(vB)

        mov     edx,vB                          ; Vector B
        mov     ecx,vA                          ; Vector A

        movups  xmm1,[edx]                      ;vB.xyz# {# Bz By Bx}
        movups  xmm0,[ecx]                      ;vA.xyz# {# Az Ay Ax}

        mov     eax,vD                          ; float D
        mulps   xmm0,xmm1                       ; {A#B# AzBz AyBy AxBx}

        movaps  xmm1,xmm0
        movaps  xmm2,xmm0

        unpckhps        xmm0,xmm0                       ; {A#B# A#B# AzBz AzBz}
        shufps  xmm1,xmm1,0e1h    ;11100001 3 2 0 1     ; {A#B# AzBz AxBx AyBy}
        addss   xmm2,xmm0                       ; {A#B# AzBz AxBx AzBz+AxBx}
        addss   xmm2,xmm1                       ; {A#B# AzBz AxBx AzBz+AxBx+AyBy}

        movss   [eax],xmm2                      ; Save Scalar

;;;        emms
        ret
vmp_DotProductAsmSSE0 endp

;_________________________________________________________________________________________________

; *** 128bit Aligned SSE version ***

        align   16
        public  vmp_DotProductAsmSSE
vmp_DotProductAsmSSE       proc    near , vD:ptr, vA:ptr, vB:ptr
        .XMM
;;;        emms
ASSERT_PTR(vD)
ASSERT_PTR16(vA)
ASSERT_PTR16(vB)

        mov     edx,vB                          ; Vector B
        mov     ecx,vA                          ; Vector A

        movaps  xmm1,[edx]                      ;vB.xyz# {# Bz By Bx}
        movaps  xmm0,[ecx]                      ;vA.xyz# {# Az Ay Ax}

        mov     eax,vD                          ; float D
        mulps   xmm0,xmm1                       ; {A#B# AzBz AyBy AxBx}

        movaps  xmm1,xmm0
        movaps  xmm2,xmm0

        unpckhps        xmm0,xmm0                       ; {A#B# A#B# AzBz AzBz}
        shufps  xmm1,xmm1,0e1h    ;11100001 3 2 0 1     ; {A#B# AzBz AxBx AyBy}
        addss   xmm2,xmm0                       ; {A#B# AzBz AxBx AzBz+AxBx}
        addss   xmm2,xmm1                       ; {A#B# AzBz AxBx AzBz+AxBx+AyBy}

        movss   [eax],xmm2                      ; Save Scalar

;;;        emms
        ret
vmp_DotProductAsmSSE endp


;_________________________________________________________________________________________________
;=================================================================================================

;void vmp_CrossProduct(vmp3DVector * pvD, const vmp3DVector* pvA, const vmp3DVector* pvB)
;{      
;       float x,y;
;            x = pvA->y * pvB->z - pvA->z * pvB->y;        AyBz-AzBy
;            y = pvA->z * pvB->x - pvA->x * pvB->z;        AzBx-AxBz
;       pvD->z = pvA->x * pvB->y - pvA->y * pvB->x;        AxBy-AyBx
;       pvD->x = x;
;       pvD->y = y;
;       return vD;
;
; WARNING: vD isn't saved until all calculations completed just incase vD is also vA or vB!
;
;       NOTE: vA, vB, vD most likely will be misaligned especially if a packed structure
;       such as VMP3DVector is used so each field must be loaded individually!
;

;_________________________________________________________________________________________________

; *** 3D Now! Version ***

        align   16
        public  vmp_CrossProductAsm3DNow
vmp_CrossProductAsm3DNow       proc    near , vD:ptr, vA:ptr, vB:ptr

        .K3D
;;;        femms
ASSERT_PTR(vD)
ASSERT_PTR(vA)
ASSERT_PTR(vB)

        mov     edx,vB                          ; Vector B
        mov     ecx,vA                          ; Vector A
        mov     eax,vD                          ; Vector Destination

        movd    mm0,(vmp3DVector PTR [edx]).x   ;vB.x   {0 Bx}
        movd    mm1,(vmp3DVector PTR [edx]).y   ;vB.y   {0 By}
        movd    mm2,(vmp3DVector PTR [edx]).z   ;vB.z   {0 Bz}

        movd    mm3,(vmp3DVector PTR [ecx]).x   ;vA.x   {0 Ax}
        movd    mm4,(vmp3DVector PTR [ecx]).y   ;vA.y   {0 Ay}
        movd    mm5,(vmp3DVector PTR [ecx]).z   ;vA.z   {0 Az}

        pfmul   mm4,mm0                         ;vB.xy  {00 AyBx}
        punpckldq mm0,mm1                               ;       {By Bx}
        movd    mm1,(vmp3DVector PTR [ecx]).y   ;vA.y   {0 Ay}
        punpckldq mm2,mm2                               ;       {Bz Bz}
        punpckldq mm3,mm1                               ;       {Ay Ax}
        punpckldq   mm5,mm5                     ;       {Az Az}

        pfmul   mm2,mm3                         ;*vA.xy {BzAy BzAx}
        movd    mm6,(vmp3DVector PTR [edx]).y   ;vB.y   {0 By}
        movq    mm7,mm2                         ;       {BzAy BzAx}

        pfmul   mm5,mm0                         ;*vB.xy {AzBy AzBx}
        pfsub   mm2,mm5                         ;       {BzAy-AzBy  BzAx-AzBx}
        pfsub   mm5,mm7                         ;y@     {AzBy-BzAy  AzBx-BzAx}

        pfmul   mm6,mm3                         ;vA.xy  {0Ay ByAx}
        movd    (vmp3DVector PTR [eax]).y,mm5   ;vD.y = AzBx-AxBz

        psrlq   mm2,32                          ;x@     {0 BzAy-AzBy}
        pfsub   mm6,mm4                         ;z@     {0-0 ByAx-AyBx}

        movd    (vmp3DVector PTR [eax]).x,mm2   ;vD.x = AyBz-AzBy
        movd    (vmp3DVector PTR [eax]).z,mm6   ;vD.z = AxBy-AyBx

;;;        femms
        ret
vmp_CrossProductAsm3DNow endp


;_________________________________________________________________________________________________

; *** NON-Aligned (128bit load/save) SSE version ***

        align   16
        public  vmp_CrossProductAsmSSE0
vmp_CrossProductAsmSSE0       proc    near , vD:ptr, vA:ptr, vB:ptr

        .XMM
;;;        emms
ASSERT_PTR(vD)
ASSERT_PTR(vA)
ASSERT_PTR(vB)

        mov     edx,vB                          ; Vector B
        mov     ecx,vA                          ; Vector A
        mov     eax,vD                          ; Vector Destination

        ;       z = AxBy-AyBx   y = AzBx-AxBz   x = AyBz-AzBy

        movups  xmm1,[edx]                      ;vB.xyz# {# Bz By Bx}
        movups  xmm0,[ecx]                      ;vA.xyz# {# Az Ay Ax}
        andps   xmm1,OWORD PTR lomsk96          ;        {0 Bz By Bx}
        andps   xmm0,OWORD PTR lomsk96          ;        {0 Az Ay Ax}
        movaps  xmm5,xmm1

        shufps  xmm1,xmm1,0d2h         ;11010010b  3 1 0 2 {0 By Bx Bz}
        movaps  xmm6,xmm0
        shufps  xmm0,xmm0,0c9h         ;11001001b  3 0 2 1 {0 Ax Az Ay}
        shufps  xmm6,xmm6,0d2h         ;11010010b  3 1 0 2 {0 Ay Ax Az}
        mulps   xmm1,xmm0

        shufps  xmm5,xmm5,0c9h         ;11001001b  3 0 2 1 {0 Bx Bz By}
        movups  xmm2,[eax]                      ; Get 128bit Destination

        mulps   xmm5,xmm6
        andps   xmm2,OWORD PTR himsk32          ;        {# 0 0 0} Retain dest upper 32bits
        subps   xmm1,xmm5                       ;        {# z y x}
        orps    xmm1,xmm2
        movups  [eax],xmm1                      ;vD.#xyz {# z y x}

;;;        emms
        ret
vmp_CrossProductAsmSSE0 endp

;_________________________________________________________________________________________________

;
; *** 16byte Aligned SSE version ***
; NOTE: If arguments not aligned, will exception fault!
;
        align   16
        public  vmp_CrossProductAsmSSE
vmp_CrossProductAsmSSE       proc    near , vD:ptr, vA:ptr, vB:ptr

        .XMM
;;;        emms
ASSERT_PTR16(vD)
ASSERT_PTR16(vA)
ASSERT_PTR16(vB)

        mov     edx,vB                          ; Vector B
        mov     ecx,vA                          ; Vector A
        mov     eax,vD                          ; Vector Destination

        ;       z = AxBy-AyBx   y = AzBx-AxBz   x = AyBz-AzBy

        movaps  xmm1,[edx]                      ;vB.xyz# {# Bz By Bx}
        movaps  xmm0,[ecx]                      ;vA.xyz# {# Az Ay Ax}
        andps   xmm1,OWORD PTR lomsk96          ;        {0 Bz By Bx}
        andps   xmm0,OWORD PTR lomsk96          ;        {0 Az Ay Ax}
        movaps  xmm5,xmm1
        shufps  xmm1,xmm1,0d2h         ;11010010b  3 1 0 2 {0 By Bx Bz}

        movaps  xmm6,xmm0
        shufps  xmm0,xmm0,0c9h         ;11001001b  3 0 2 1 {0 Ax Az Ay}
        shufps  xmm6,xmm6,0d2h         ;11010010b  3 1 0 2 {0 Ay Ax Az}
        mulps   xmm1,xmm0
        shufps  xmm5,xmm5,0c9h         ;11001001b  3 0 2 1 {0 Bx Bz By}
        movaps  xmm2,[eax]                      ; Get 128bit Destination
        mulps   xmm5,xmm6
        andps   xmm2,OWORD PTR himsk32          ;        {# 0 0 0} Retain dest upper 32bits
        subps   xmm1,xmm5                       ;        {# z y x}
        orps    xmm1,xmm2
        movaps  [eax],xmm1                      ;vD.#xyz {# z y x}
;;;        emms
        ret
vmp_CrossProductAsmSSE endp





;_________________________________________________________________________________________________




Vmd3DX86        ends
        end

