
//
//  QVsf3D.cpp - 3D (Quad) Vector Special Function
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 3 Oct 2001 - James Leiterman
//
//  Last Modification
//
//  stab=4
//



#include "KariType.h"                                   // Master Definitions

    //  3rd Party Libraries
    
#include    <assert.h>
#include    <iostream.h>
#include    <math.h>

    //  Book Development Library Modules

#include    "CpuAsm.h"                                  // CPU
#include    "vmp3D.h"
#include    "vmp3DEx.h"

#include    "QVsf3D.h"                                  // 3D Math
#include    "QVsf3DEx.h"



vmp_QVecSqrtProc            vmp_QVecSqrt;
vmp_QVecSqrtProc            vmp_QVecSqrt0;
vmp_QVecSqrtProc            vmp_QVecSqrtFast;
vmp_QVecSqrtProc            vmp_QVecSqrtFast0;

vmp_QVecMinProc             vmp_QVecMin;
vmp_QVecMinProc             vmp_QVecMin0;
vmp_QVecMaxProc             vmp_QVecMax;
vmp_QVecMaxProc             vmp_QVecMax0;

vmp_QVecMagnitudeProc       vmp_QVecMagnitude;
vmp_QVecMagnitudeProc       vmp_QVecMagnitude0;
vmp_QVecMagnitudeProc       vmp_QVecMagnitudeFast;
vmp_QVecMagnitudeProc       vmp_QVecMagnitudeFast0;

vmp_QVecNormalizeProc       vmp_QVecNormalize;
vmp_QVecNormalizeProc       vmp_QVecNormalize0;
vmp_QVecNormalizeProc       vmp_QVecNormalizeFast;
vmp_QVecNormalizeProc       vmp_QVecNormalizeFast0;





//
//  (Quad) Vector Square Root
//

void vmp_QVecSqrtGeneric(vmp3DQVector * const pvD, const vmp3DQVector * const pvA)
{
    ASSERT_PTR4(pvD);
    ASSERT_PTR4(pvA);

    ASSERT_FNEG(pvA->x);
    ASSERT_FNEG(pvA->y);
    ASSERT_FNEG(pvA->z);
    ASSERT_FNEG(pvA->w);

    pvD->x = sqrtf( pvA->x );
    pvD->y = sqrtf( pvA->y );
    pvD->z = sqrtf( pvA->z );
    pvD->w = sqrtf( pvA->w );
}



//
//  (Quad) Vector Minimum
//

void vmp_QVecMinGeneric( vmp3DQVector * const pvD, const vmp3DQVector * const pvA, const vmp3DQVector * const pvB )
{
    ASSERT_PTR4(pvD);
    ASSERT_PTR4(pvA);
    ASSERT_PTR4(pvB);

    pvD->x = (pvA->x < pvB->x) ? pvA->x : pvB->x;
    pvD->y = (pvA->y < pvB->y) ? pvA->y : pvB->y;
    pvD->z = (pvA->z < pvB->z) ? pvA->z : pvB->z;
    pvD->w = (pvA->w < pvB->w) ? pvA->w : pvB->w;
}



//
//  (Quad) Vector Maximum
//

void vmp_QVecMaxGeneric( vmp3DQVector * const pvD, const vmp3DQVector * const pvA, const vmp3DQVector * const pvB )
{
    ASSERT_PTR4(pvD);
    ASSERT_PTR4(pvA);
    ASSERT_PTR4(pvB);

    pvD->x = (pvA->x > pvB->x) ? pvA->x : pvB->x;
    pvD->y = (pvA->y > pvB->y) ? pvA->y : pvB->y;
    pvD->z = (pvA->z > pvB->z) ? pvA->z : pvB->z;
    pvD->w = (pvA->w > pvB->w) ? pvA->w : pvB->w;
}



//
//  (Quad) Vector Magnitude
//

void vmp_QVecMagnitudeGeneric(float * const pfD, const vmp3DQVector * const pvA)
{
    ASSERT_PTR4(pfD);
    ASSERT_PTR4(pvA);

    *pfD = sqrtf( pvA->x * pvA->x + pvA->y * pvA->y + pvA->z * pvA->z );
}



//
//  (Quad) Vector Normalization
//

void vmp_QVecNormalizeGeneric(vmp3DQVector * const pvD, const vmp3DQVector * const pvA)
{
    float fMag;

    ASSERT_PTR4(pvD);
    ASSERT_PTR4(pvA);

    fMag = sqrtf( pvA->x * pvA->x + pvA->y * pvA->y + pvA->z * pvA->z );

    if ( fMag < SINGLE_PRECISION )  // If too close to zero...
      {
        pvD->x = pvA->x;
        pvD->y = pvA->y;
        pvD->z = pvA->z;
      }
    else
      {
        fMag = 1.0f / fMag;
        pvD->x = pvA->x * fMag;
        pvD->y = pvA->y * fMag;
        pvD->z = pvA->z * fMag;
      }

    pvD->w = 0.0;
}







