
//
//  Bench.cpp - Vector 3D Matrix Function
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//      Hopefully by now you have realized the necessity of 16 byte aligned data
//      and so this program is written as such!
//
//      Because of all the Generic code there is an excessive amount of SIMD entry/exit
//      function calls. This gets dramatically reduced by
//
//  Created - 5 Oct 2001 - James Leiterman
//
//  Last Modification
//
//
//  stab=4
//

#include "KariType.h"                                   // Master Definitions

    //  3rd Party Libraries
    
#include    <assert.h>
#include    <iostream.h>
#include    <memory.h>

#include <D3dx8math.h>

    //  Book Development Library Modules

#include "CpuAsm.h"                                     // CPU Module
#include "vmp3D.h"
#include "Matrix3D.h"
#include "Matrix3DEx.h"

#include "vecBug.h"



//
//  Test GenerateModel with Matrix Alignment
//

static void testMatrix( void )
{
    float matrixbuf[16  *14 +3];    // enough space + 128bit alignment padding
    float fAngX, fAngY, fAngZ, fAngX2, fAngY2, fAngZ2, fStep, fGDet;
    vmp3DMatrix *pmGWorld, *pmWorld, *pmGWorld2, *pmWorld2;
    vmp3DMatrix *pmGRotX, *pmGRotY, *pmGRotZ;
    vmp3DMatrix *pmRotX, *pmRotY, *pmRotZ;
    vmp3DMatrix *pmGTmp, *pmTmp;
    vmp3DMatrix *pmGTrans, *pmTrans;
    float fPrecision;

    fPrecision = SINGLE_PRECISION;
    fStep = 30.0f;

        //  Adjust Alignment factor

    pmGRotX = (vmp3DMatrix*) ALIGN16((int)matrixbuf);

    pmGRotY = pmGRotX + 1;
    pmGRotZ = pmGRotY + 1;
    
    pmRotX = pmGRotZ + 1;
    pmRotY = pmRotX + 1;
    pmRotZ = pmRotY + 1;

    pmGWorld = pmRotZ + 1;
    pmWorld = pmGWorld + 1;

    pmGWorld2 = pmWorld + 1;
    pmWorld2 = pmGWorld2 + 1;

    pmGTrans = pmWorld2 + 1;
    pmTrans = pmGTrans + 1;

    pmGTmp = pmTrans + 1;
    pmTmp = pmGTmp + 1;

        //  Matrix Identity Check

    cout << "Identity Matrix Check" << endl;

    vmp_MatrixSetIdentityGeneric( *pmGTmp );

    vmp_SIMDEntry();    // MMX vvvvvvvvvvvvvvvvvvvvvvvvv

    vmp_MatrixSetIdentity( *pmTmp );

    if ( !vmp_IsMatrixEqual( *pmGTmp, *pmTmp, fPrecision ))
      {
        cout << " Identity failure!" << endl;

        vmp_MatrixPrint(*pmGTmp);
        vmp_MatrixPrint(*pmTmp);
        return;
      }

    vmp_SIMDExit();     // FPU ^^^^^^^^^^^^^^^^^^^^^^^^^

        //  Setup our test Source Matrix

    vmp_MatrixTestFill( *pmGWorld );

        //  ____________________________________________
        //  Matrix Copy Check

    cout << "Copy Matrix Check" << endl;

    vmp_MatrixSplat( *pmGTmp, 0.0f );
    vmp_MatrixSplat( *pmTmp,  0.0f );

    vmp_MatrixCopyGeneric( *pmGTmp, *pmGWorld );        

    vmp_SIMDEntry();    // MMX vvvvvvvvvvvvvvvvvvvvvvvvv

    vmp_MatrixCopy(        *pmTmp,  *pmGWorld );        

    if ( !vmp_IsMatrixEqual( *pmGTmp, *pmGWorld, fPrecision ))
      {
        cout << " Matrix Generic Copy failure!" << endl;
        vmp_MatrixPrint(*pmGTmp);
        vmp_MatrixPrint(*pmTmp);
        return;
      }

    if ( !vmp_IsMatrixEqual( *pmTmp, *pmGTmp, fPrecision ))
      {
        cout << " Matrix Copy failure!" << endl;
        vmp_MatrixPrint(*pmGTmp);
        vmp_MatrixPrint(*pmTmp);
        return;
      }

        //  ____________________________________________
        //  Matrix Set Translate Check

    cout << "Set Translate Matrix Check" << endl;

    vmp_MatrixSplat( *pmGTmp, 0.0f );
    vmp_MatrixSplat( *pmTmp,  0.0f );

    vmp_SIMDExit();     // FPU ^^^^^^^^^^^^^^^^^^^^^^^^^

    vmp_MatrixSetTranslateGeneric( *pmGTmp, 4.0f, 5.0f, 6.0f );     

    vmp_SIMDEntry();    // MMX vvvvvvvvvvvvvvvvvvvvvvvvv
    vmp_MatrixSetTranslate(        *pmTmp,  4.0f, 5.0f, 6.0f );     

    if ( !vmp_IsMatrixEqual( *pmTmp, *pmGTmp, fPrecision ))
      {
        cout << " Matrix Translate failure!" << endl;
        vmp_MatrixPrint(*pmGTmp);
        vmp_MatrixPrint(*pmTmp);
        return;
      }

        //  ____________________________________________
        //  Matrix Transpose Check

    cout << "Transpose Matrix Check" << endl;

    vmp_MatrixSplat( *pmGTmp, 0.0f );
    vmp_MatrixSplat( *pmTmp,  0.0f );
    vmp_SIMDExit();     // ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

    vmp_MatrixTransposeGeneric( *pmGTmp, *pmGWorld );   // diagonal - flip

    vmp_SIMDEntry();    // MMX vvvvvvvvvvvvvvvvvvvvvvvvv
    vmp_MatrixTranspose(        *pmTmp, *pmGWorld );    // diagonal - flip

    if (    vmp_IsMatrixEqual( *pmGTmp, *pmGWorld, fPrecision )
        ||  vmp_IsMatrixEqual( *pmGTmp, *pmGWorld, fPrecision )
        || !vmp_IsMatrixEqual( *pmGTmp, *pmTmp,    fPrecision ) )
      {
        cout << " Matrix Generic Transpose failure!" << endl;
        vmp_MatrixPrint(*pmGTmp);
        vmp_MatrixPrint(*pmTmp);
        return;
      }
    
    vmp_SIMDExit();     // FPU ^^^^^^^^^^^^^^^^^^^^^^^^^

    vmp_MatrixTransposeGeneric( *pmGTmp, *pmGTmp );     // diagonal - flop

    vmp_SIMDEntry();    // MMX vvvvvvvvvvvvvvvvvvvvvvvvv

    vmp_MatrixTranspose(        *pmTmp,  *pmTmp );      // diagonal - flop

    if (   !vmp_IsMatrixEqual( *pmGTmp, *pmGWorld, fPrecision )
        || !vmp_IsMatrixEqual( *pmGTmp, *pmTmp,    fPrecision ) )
      {
        cout << " Matrix Transpose failure!" << endl;
        vmp_MatrixPrint(*pmGTmp);
        vmp_MatrixPrint(*pmTmp);
        return;
      }

    vmp_SIMDExit();     // FPU ^^^^^^^^^^^^^^^^^^^^^^^^^

        //  ____________________________________________
        //  Matrix-Matrix Addition Check

    cout << "Matrix-Matrix Addition Check" << endl;

    vmp_MatrixAddGeneric( *pmGTmp, *pmGWorld, *pmGWorld );

    vmp_SIMDEntry();    // MMX vvvvvvvvvvvvvvvvvvvvvvvvv
    vmp_MatrixAdd(        *pmTmp,  *pmGWorld, *pmGWorld );

    if (    vmp_IsMatrixEqual( *pmGTmp, *pmGWorld, fPrecision )
        || !vmp_IsMatrixEqual( *pmGTmp, *pmTmp,    fPrecision ) )
      {
        cout << " Matrix-Matrix Addition failure!" << endl;
        vmp_MatrixPrint(*pmGTmp);
        vmp_MatrixPrint(*pmTmp);
        return;
      }

    vmp_SIMDExit();     // FPU ^^^^^^^^^^^^^^^^^^^^^^^^^

        //  ____________________________________________
        //  Matrix Scaling Check

    cout << "Matrix Scaling Check" << endl;

    vmp_MatrixSetScalingGeneric( *pmGTmp, 1.0, 2.0, 3.0 );

    vmp_SIMDEntry();    // MMX vvvvvvvvvvvvvvvvvvvvvvvvv
    vmp_MatrixSetScaling(        *pmTmp,  1.0, 2.0, 3.0 );

    if (!vmp_IsMatrixEqual( *pmGTmp, *pmTmp,    fPrecision ))
      {
        cout << " Set Matrix Scaling failure!" << endl;
        vmp_MatrixPrint(*pmGTmp);
        vmp_MatrixPrint(*pmTmp);
        return;
      }

        //  ____________________________________________
        //  Matrix-Matrix Multiplication Check

    cout << "Matrix-Matrix Multiplication Check" << endl;

    vmp_MatrixSplat( *pmGTmp, 0.0f );
    vmp_MatrixSplat( *pmTmp,  0.0f );

    vmp_SIMDExit();     // FPU ^^^^^^^^^^^^^^^^^^^^^^^^^

    vmp_MatrixMulGeneric( *pmGTmp, *pmGWorld, *pmGWorld );

    vmp_SIMDEntry();    // MMX vvvvvvvvvvvvvvvvvvvvvvvvv
    vmp_MatrixMul(        *pmTmp,  *pmGWorld, *pmGWorld );

    if (    vmp_IsMatrixEqual( *pmGTmp, *pmGWorld, fPrecision )
        || !vmp_IsMatrixEqual( *pmGTmp, *pmTmp,    fPrecision ) )
      {
        cout << " Matrix-Matrix Multiplication failure!" << endl;
        vmp_MatrixPrint(*pmGTmp);
        vmp_MatrixPrint(*pmTmp);
        return;
      }

    vmp_SIMDExit();     // FPU ^^^^^^^^^^^^^^^^^^^^^^^^^

        //  ____________________________________________
        //  Rotations Check

    cout << "Matrix XYZ Rotations Check" << endl;

    for ( fAngX = -360.0f; fAngX <= 360.0f; fAngX += fStep )
      {
        for ( fAngY = -180.0f; fAngY <= 180.0f; fAngY += fStep )
          {
            for ( fAngZ = -360.0f; fAngZ <= 360.0f; fAngZ += fStep )
              {
                fAngX2 = DEG2RAD( fAngX ); /// 182.04f;
                fAngY2 = DEG2RAD( fAngY ); /// 182.04f;
                fAngZ2 = DEG2RAD( fAngZ ); /// 182.04f;
            
                    //  Rotate Generic and custom

                vmp_MatrixSetRotateXGeneric( *pmGRotX, fAngX2 );
                vmp_MatrixSetRotateYGeneric( *pmGRotY, fAngY2 );
                vmp_MatrixSetRotateZGeneric( *pmGRotZ, fAngZ2 );
                vmp_MatrixSetTranslateGeneric( *pmGTrans, fAngX, fAngY, fAngZ );

                vmp_MatrixSetRotateX(   *pmRotX, fAngX2 );
                vmp_MatrixSetRotateY(   *pmRotY, fAngY2 );
                vmp_MatrixSetRotateZ(   *pmRotZ, fAngZ2 );

                vmp_SIMDEntry();    // MMX

                vmp_MatrixSetTranslate( *pmTrans, fAngX, fAngY, fAngZ );

                    //  Compare

                if ( !vmp_IsMatrixEqual( *pmGRotX, *pmRotX, fPrecision ))
                  {
                    cout << " Aligned Matrix Rot X failure!" << endl;
                    vmp_MatrixPrint(*pmGRotX);
                    vmp_MatrixPrint(*pmRotX);
                    continue;
                  }

                if ( !vmp_IsMatrixEqual( *pmGRotY, *pmRotY, fPrecision ))
                  {
                    cout << " Aligned Matrix Rot Y failure!" << endl;
                    vmp_MatrixPrint(*pmGRotY);
                    vmp_MatrixPrint(*pmRotY);
                    continue;
                  }

                if ( !vmp_IsMatrixEqual( *pmGRotZ, *pmRotZ, fPrecision ))
                  {
                    cout << " Aligned Matrix Rot Z failure!" << endl;
                    vmp_MatrixPrint(*pmGRotZ);
                    vmp_MatrixPrint(*pmRotZ);
                    continue;
                  }

                if ( !vmp_IsMatrixEqual( *pmGTrans, *pmTrans, fPrecision ))
                  {
                    cout << " Aligned Matrix Translation failure!" << endl;
                    vmp_MatrixPrint(*pmGTrans);
                    vmp_MatrixPrint(*pmTrans);
                    continue;
                  }

                vmp_SIMDExit();     // FPU

                    //  Addition Check

                vmp_MatrixAddGeneric( *pmGWorld, *pmGRotX,  *pmGRotY );

                vmp_SIMDEntry();    // MMX

                vmp_MatrixAdd(        *pmWorld,  *pmRotX,   *pmRotY );

                if ( !vmp_IsMatrixEqual( *pmGWorld, *pmWorld, fPrecision ))
                  {
                    cout << " Aligned Matrix Addition failure!" << endl;
                    vmp_MatrixPrint(*pmGWorld);
                    vmp_MatrixPrint(*pmWorld);
                    continue;
                  }

                vmp_SIMDExit();     // FPU

                    //  Matrix-Matrix Multiplication Check

                vmp_MatrixTestFill( *pmGTmp );

                vmp_SIMDEntry();    // MMX

                vmp_MatrixAdd(*pmTmp, *pmGTmp, *pmGTmp);

                vmp_SIMDExit();     // FPU

                vmp_MatrixMulGenericSlow( *pmGWorld, *pmGTmp,  *pmTmp );
                vmp_MatrixMulGeneric(     *pmWorld,  *pmGTmp,  *pmTmp );

                vmp_SIMDEntry();    // MMX

                if ( !vmp_IsMatrixEqual( *pmGWorld, *pmWorld, fPrecision ))
                  {
                    cout << " Aligned Generic Slow to Generic Matrix Mul failure!" << endl;
                    vmp_MatrixPrint(*pmGWorld);
                    vmp_MatrixPrint(*pmWorld);
                    continue;
                  }

                vmp_MatrixMul( *pmGWorld, *pmGTmp,  *pmTmp );

                if ( !vmp_IsMatrixEqual( *pmGWorld, *pmWorld, fPrecision ))
                  {
                    cout << " Aligned Generic Slow to Generic Matrix Mul failure!" << endl;
                    vmp_MatrixPrint(*pmGWorld);
                    vmp_MatrixPrint(*pmWorld);
                    continue;
                  }

                vmp_SIMDExit();     // FPU

                    //  Build up Matrix using each Rotation Matrix

                vmp_MatrixMulGeneric( *pmGWorld, *pmGRotX,  *pmGRotY );
                vmp_MatrixMulGeneric( *pmGWorld, *pmGWorld, *pmGRotZ );
                vmp_MatrixMulGeneric( *pmGWorld, *pmGWorld, *pmGTrans );

                vmp_SIMDEntry();    // MMX

                vmp_MatrixMul(        *pmWorld,  *pmRotX,   *pmRotY );
                vmp_MatrixMul(        *pmWorld,  *pmWorld,  *pmRotZ );
                vmp_MatrixMul(        *pmWorld,  *pmWorld,  *pmTrans );

                if ( !vmp_IsMatrixEqual( *pmGWorld, *pmWorld, fPrecision ))
                  {
                    cout << " Aligned Matrix Rot then Mul failure!" << endl;
                    vmp_MatrixPrint(*pmGWorld);
                    vmp_MatrixPrint(*pmWorld);
                    continue;
                  }

                    //  Matrix Inverse  D=A^-1

                vmp_MatrixSplat( *pmGTmp, -1.0f );
                vmp_MatrixSplat( *pmTmp, -8.0f );

                if (NULL==D3DXMatrixInverse( (D3DXMATRIX*)pmGTmp, &fGDet, (D3DXMATRIX*)pmGWorld))
                  {
                    cout << "No determinate!";
                  }
                else
                  {
                    vmp_MatrixInvGeneric( *pmTmp, *pmGWorld );
                    if (!vmp_IsMatrixEqual( *pmGTmp, *pmTmp, 0.001f ))  //fPrecision ))
                      {
                        cout << " MatrixInverse failure!" << endl;
                        vmp_MatrixPrint(*pmGWorld);
                        vmp_MatrixPrint(*pmGTmp);
                        vmp_MatrixPrint(*pmTmp);
                      }

                    vmp_MatrixSetIdentity( *pmWorld2 );

                    vmp_MatrixMul( *pmGTmp, *pmGTmp, *pmGWorld );
                    vmp_MatrixMul( *pmTmp,  *pmTmp,  *pmGWorld );

                    if (!vmp_IsMatrixEqual( *pmGTmp, *pmWorld2, 0.0001f ))
                      {
                        cout << " MatrixInverse G to Identity failure!" << endl;
                        vmp_MatrixPrint(*pmWorld2);
                        vmp_MatrixPrint(*pmGTmp);
                      }

                    if (!vmp_IsMatrixEqual( *pmTmp, *pmWorld2, 0.0001f ))
                      {
                        cout << " MatrixInverse to Identity failure!" << endl;
                        vmp_MatrixPrint(*pmGWorld2);
                        vmp_MatrixPrint(*pmTmp);
                      }           
                  }
                vmp_SIMDExit();     // FPU ^^^^^^^^^^^^^^^^^^^^^^^^^

                    //  Set a World View using Rotation Angles

                vmp_MatrixSetIdentity(        *pmWorld2 );

                vmp_MatrixRotateX(            *pmWorld2, *pmWorld2, fAngX2 );
                vmp_MatrixRotateY(            *pmWorld2, *pmWorld2, fAngY2 );
                vmp_MatrixRotateZ(            *pmWorld2, *pmWorld2, fAngZ2 );
                vmp_MatrixTranslate(          *pmWorld2, *pmWorld2, fAngX, fAngY, fAngZ );
                
                    //  Compare the two different methods of Rotation Matrix & Rotation Angles

                if ( !vmp_IsMatrixEqual( *pmWorld, *pmWorld2, fPrecision ))
                  {
                    cout << " Aligned Matrix Rot diff failure!" << endl;
                    vmp_MatrixPrint(*pmWorld);
                    vmp_MatrixPrint(*pmWorld2);
                    continue;
                  }

                vmp_SIMDExit();     // FPU
              }
          }
      }
}




int main( int argc, char *argv[] )
{
    CpuInfo cinfo;
    char szBuf[ CPU_SZBUF_MAX ];

    CpuDetect( &cinfo );                                // Detect CPU

    cout << "\nVector Matrix Test\n\n";
    cout << CpuInfoStr( szBuf, &cinfo ) << endl;

    if (!( CPUBITS_MMX & cinfo.nBits ))
      {
        cout << "ERROR! MMX Not supported!\n\n" << flush;
        return -1;
      }

    Matrix3DSetup( &cinfo );                                // Setup our 3D Vector stuff

    testMatrix();

    cout << "Fin!\n\n" << flush;

    return 0;
}

