
//
//  Matrix3D.cpp - 3D Vector Matrix Function
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 3 Oct 2001 - James Leiterman
//
//  Last Modification
//
//  stab=4
//


#include "KariType.h"                                   // Master Definitions

    //  3rd Party Libraries
    
#include    <assert.h>
#include    <iostream.h>
#include    <math.h>
#include <stdio.h>
#include <D3dx8math.h>

    //  Book Development Library Modules

#include    "vmp3D.h"
#include    "vmp3DEx.h"
#include    "Matrix3D.h"                                    // 3D Math
#include    "Matrix3DEx.h"




vmp_MatrixSetIdentityProc   vmp_MatrixSetIdentity;

vmp_MatrixSetScalingProc    vmp_MatrixSetScaling;

vmp_MatrixSetTranslateProc  vmp_MatrixSetTranslate;

vmp_MatrixSetRotateProc     vmp_MatrixSetRotateX;
vmp_MatrixSetRotateProc     vmp_MatrixSetRotateY;
vmp_MatrixSetRotateProc     vmp_MatrixSetRotateZ;

vmp_MatrixTransposeProc     vmp_MatrixTranspose;

vmp_MatrixCopyProc      vmp_MatrixCopy;

vmp_MatrixAddProc       vmp_MatrixAdd;

vmp_MatrixMulProc       vmp_MatrixMul;




//                            | 1 0 0 0 |
//  Matrix - Set Identity     | 0 1 0 0 |
//                            | 0 0 1 0 |
//                            | 0 0 0 1 |

void vmp_MatrixSetIdentityGeneric( vmp3DMatrix Mx )
{
    ASSERT_PTR4(Mx);

                  Mx[0][1]=     Mx[0][2]=     Mx[0][3] = 
    Mx[1][0]=                   Mx[1][2]=     Mx[1][3] = 
    Mx[2][0]=     Mx[2][1]=                   Mx[2][3] = 
    Mx[3][0]=     Mx[3][1]=     Mx[3][2]=                0.0f;

    Mx[0][0]=     Mx[1][1]=     Mx[2][2]=     Mx[3][3]=  1.0f;
}



//
//  Matrix - Scalar mD = r*mA
//

void vmp_MatrixScalarGeneric( vmp3DMatrix dMx, const vmp3DMatrix aMx, float r )
{
    ASSERT_PTR4(dMx);
    ASSERT_PTR4(aMx);

    dMx[0][0]=aMx[0][0]*r;  dMx[0][1]=aMx[0][1]*r;
    dMx[0][2]=aMx[0][2]*r;  dMx[0][3]=aMx[0][3]*r;

    dMx[1][0]=aMx[1][0]*r;  dMx[1][1]=aMx[1][1]*r;
    dMx[1][2]=aMx[1][2]*r;  dMx[1][3]=aMx[1][3]*r;

    dMx[2][0]=aMx[2][0]*r;  dMx[2][1]=aMx[2][1]*r;
    dMx[2][2]=aMx[2][2]*r;  dMx[2][3]=aMx[2][3]*r;

    dMx[3][0]=aMx[3][0]*r;  dMx[3][1]=aMx[3][1]*r;
    dMx[3][2]=aMx[3][2]*r;  dMx[3][3]=aMx[3][3]*r;
}



//                            | Sx 0  0  0 |
//  Matrix - Set Scaling      | 0  Sy 0  0 |
//                            | 0  0  Sz 0 |
//                            | 0  0  0  1 |

void vmp_MatrixSetScalingGeneric( vmp3DMatrix Mx, float fSx, float fSy, float fSz )
{
    ASSERT_PTR4(Mx);

                  Mx[0][1]=     Mx[0][2]=     Mx[0][3] = 
    Mx[1][0]=                   Mx[1][2]=     Mx[1][3] = 
    Mx[2][0]=     Mx[2][1]=                   Mx[2][3] =
    Mx[3][0]=     Mx[3][1]=     Mx[3][2]=                0.0f;

    Mx[0][0]=fSx; Mx[1][1]=fSy; Mx[2][2]=fSz; Mx[3][3] = 1.0f;
}




//                            | 1  0  0  0 |
//  Matrix - Set Translation  | 0  1  0  0 |
//                            | 0  0  1  0 |
//                            | Tx Ty Tz 1 |

void vmp_MatrixSetTranslateGeneric( vmp3DMatrix Mx, float fTx, float fTy, float fTz )
{
    ASSERT_PTR4(Mx);

                  Mx[0][1]=     Mx[0][2]=     Mx[0][3] = 
    Mx[1][0]=                   Mx[1][2]=     Mx[1][3] = 
    Mx[2][0]=     Mx[2][1]=                   Mx[2][3] = 0.0f;
    Mx[0][0]=     Mx[1][1]=     Mx[2][2]=     Mx[3][3] = 1.0f;

    Mx[3][0]=fTx; Mx[3][1]=fTy; Mx[3][2]=fTz;
}



//
//  Sine / Cosine
//

void vmp_FSinCosGeneric( float * const pfSin, float * const pfCos, float fRads )
{
    ASSERT_PTR4(pfSin);
    ASSERT_PTR4(pfCos);

    *pfSin = sinf(fRads);
    *pfCos = cosf(fRads);
}



//                            | 1   0   0   0 |
//  Matrix - Set X Rotation   | 0  cos sin  0 |
//                            | 0 -sin cos  0 |
//                            | 0   0   0   1 |

void vmp_MatrixSetRotateXGeneric( vmp3DMatrix Mx, float fRads )
{
    float fSin, fCos;

    ASSERT_PTR4(Mx);

    vmp_FSinCos(&fSin, &fCos, fRads);

    vmp_MatrixSetIdentityGeneric( Mx );

    Mx[1][2] =               fSin;
    Mx[2][2] =  Mx[1][1] =   fCos;
    Mx[2][1] = -Mx[1][2]; //-sinf
}



//                            | cos  0 -sin  0 |
//  Matrix - Set Y Rotation   |  0   1   0   0 |
//                            | sin  0  cos  0 |
//                            |  0   0   0   1 |

void vmp_MatrixSetRotateYGeneric( vmp3DMatrix Mx, float fRads )
{
    float fSin, fCos;

    ASSERT_PTR4(Mx);

    vmp_FSinCos(&fSin, &fCos, fRads);

    vmp_MatrixSetIdentityGeneric( Mx );

    Mx[2][0] =               fSin;
    Mx[2][2] =  Mx[0][0] =   fCos;
    Mx[0][2] = -Mx[2][0]; //-sinf
}



//                            |  cos sin  0   0 |
//  Matrix - Set Z Rotation   | -sin cos  0   0 |
//                            |   0   0   1   0 |
//                            |   0   0   0   1 |

void vmp_MatrixSetRotateZGeneric( vmp3DMatrix Mx, float fRads )
{
    float fSin, fCos;

    ASSERT_PTR4(Mx);

    vmp_FSinCos(&fSin, &fCos, fRads);

    vmp_MatrixSetIdentityGeneric( Mx );

    Mx[0][1] =               fSin;
    Mx[1][1] =  Mx[0][0] =   fCos;
    Mx[1][0] = -Mx[0][1]; //-sinf
}



//
//  Matrix - Copy  mD = mA
//

void vmp_MatrixCopyGeneric( vmp3DMatrix dMx, const vmp3DMatrix aMx )
{
    ASSERT_PTR4(dMx);
    ASSERT_PTR4(aMx);

    dMx[0][0]=aMx[0][0];  dMx[0][1]=aMx[0][1];  dMx[0][2]=aMx[0][2];  dMx[0][3]=aMx[0][3];
    dMx[1][0]=aMx[1][0];  dMx[1][1]=aMx[1][1];  dMx[1][2]=aMx[1][2];  dMx[1][3]=aMx[1][3];
    dMx[2][0]=aMx[2][0];  dMx[2][1]=aMx[2][1];  dMx[2][2]=aMx[2][2];  dMx[2][3]=aMx[2][3];
    dMx[3][0]=aMx[3][0];  dMx[3][1]=aMx[3][1];  dMx[3][2]=aMx[3][2];  dMx[3][3]=aMx[3][3];
}



//
//  Matrix - Addition  mD = mA+mB
//

void vmp_MatrixAddGeneric( vmp3DMatrix dMx, const vmp3DMatrix aMx, const vmp3DMatrix bMx )
{
    ASSERT_PTR4(dMx);
    ASSERT_PTR4(aMx);
    ASSERT_PTR4(bMx);

    dMx[0][0]=aMx[0][0]+bMx[0][0];  dMx[0][1]=aMx[0][1]+bMx[0][1];
    dMx[0][2]=aMx[0][2]+bMx[0][2];  dMx[0][3]=aMx[0][3]+bMx[0][3];

    dMx[1][0]=aMx[1][0]+bMx[1][0];  dMx[1][1]=aMx[1][1]+bMx[1][1];
    dMx[1][2]=aMx[1][2]+bMx[1][2];  dMx[1][3]=aMx[1][3]+bMx[1][3];

    dMx[2][0]=aMx[2][0]+bMx[2][0];  dMx[2][1]=aMx[2][1]+bMx[2][1];
    dMx[2][2]=aMx[2][2]+bMx[2][2];  dMx[2][3]=aMx[2][3]+bMx[2][3];

    dMx[3][0]=aMx[3][0]+bMx[3][0];  dMx[3][1]=aMx[3][1]+bMx[3][1];
    dMx[3][2]=aMx[3][2]+bMx[3][2];  dMx[3][3]=aMx[3][3]+bMx[3][3];
}



//
//  Matrix - Transpose
//

void vmp_MatrixTransposeGeneric( vmp3DMatrix dMx, const vmp3DMatrix aMx )
{
    float f[6];

    ASSERT_PTR4(dMx);
    ASSERT_PTR4(aMx);

        // Diagonal
    dMx[0][0]=aMx[0][0]; dMx[1][1]=aMx[1][1]; dMx[2][2]=aMx[2][2]; dMx[3][3]=aMx[3][3];

        // Use a tmp. buffer just incase source and destination are the same!
         f[0]=aMx[0][1];      f[1]=aMx[0][2];      f[2]=aMx[0][3];
                              f[3]=aMx[1][2];      f[4]=aMx[1][3];
                                                   f[5]=aMx[2][3];
    dMx[0][1]=aMx[1][0]; dMx[1][0]=f[0];
    dMx[0][2]=aMx[2][0]; dMx[2][0]=f[1];
    dMx[0][3]=aMx[3][0]; dMx[3][0]=f[2];  
    dMx[1][2]=aMx[2][1]; dMx[2][1]=f[3];
    dMx[1][3]=aMx[3][1]; dMx[3][1]=f[4]; 
    dMx[2][3]=aMx[3][2]; dMx[3][2]=f[5];
}




//
//  Matrix - Inverse   mD = mA^-1
//

bool vmp_MatrixInvGeneric( vmp3DMatrix dMx, const vmp3DMatrix aMx )
{
    vmp3DMatrix s;
    vmp3DQVector t0, t1, t2;
    float fDet;
    bool bRet;

    ASSERT_PTR4(dMx);
    ASSERT_PTR4(aMx);

    vmp_MatrixTransposeGeneric( s, aMx );
                                    //     A^T        A
    t0.x = s[2][2] * s[3][3];       //  = A2z*A3w   A2z*A3w
    t0.y = s[2][3] * s[3][2];       //  = A2w*A3z   A3z*A2w
    t0.z = s[2][1] * s[3][3];       //  = A2y*A3w   A1z*A3w
    t0.w = s[2][2] * s[3][1];       //  = A2z*A3y   A2z*A1w

    t1.x = s[2][3] * s[3][1];       // += A2w*A3y   A3z*A1w
    t1.y = s[2][0] * s[3][3];       // += A2x*A3w   A0z*A3w
    t1.z = s[2][3] * s[3][0];       // += A2w*A3x   A3z*A0w
    t1.w = s[2][0] * s[3][2];       // += A2x*A3z   A0z*Azw

    t2.x = s[2][1] * s[3][2];       // += A2y*A3z   A1z*Azw
    t2.y = s[2][2] * s[3][0];       // += A2z*A3x   A2z*A0w
    t2.z = s[2][0] * s[3][1];       // += A2x*A3y   A0z*A1w
    t2.w = s[2][1] * s[3][0];       // += A2y*A3x   A1z*A0w

    dMx[0][0]  = t0.x*s[1][1] + t1.x*s[1][2] + t2.x*s[1][3];
    dMx[0][0] -= t0.y*s[1][1] + t0.z*s[1][2] + t0.w*s[1][3];

    dMx[0][1]  = t0.y*s[1][0] + t1.y*s[1][2] + t2.y*s[1][3];
    dMx[0][1] -= t0.x*s[1][0] + t1.z*s[1][2] + t1.w*s[1][3];

    dMx[0][2]  = t0.z*s[1][0] + t1.z*s[1][1] + t2.z*s[1][3];
    dMx[0][2] -= t1.x*s[1][0] + t1.y*s[1][1] + t2.w*s[1][3];

    dMx[0][3]  = t0.w*s[1][0] + t1.w*s[1][1] + t2.w*s[1][2];
    dMx[0][3] -= t2.x*s[1][0] + t2.y*s[1][1] + t2.z*s[1][2];

    dMx[1][0]  = t0.y*s[0][1] + t0.z*s[0][2] + t0.w*s[0][3];
    dMx[1][0] -= t0.x*s[0][1] + t1.x*s[0][2] + t2.x*s[0][3];

    dMx[1][1]  = t0.x*s[0][0] + t1.z*s[0][2] + t1.w*s[0][3];
    dMx[1][1] -= t0.y*s[0][0] + t1.y*s[0][2] + t2.y*s[0][3];

    dMx[1][2]  = t1.x*s[0][0] + t1.y*s[0][1] + t2.w*s[0][3];
    dMx[1][2] -= t0.z*s[0][0] + t1.z*s[0][1] + t2.z*s[0][3];

    dMx[1][3]  = t2.x*s[0][0] + t2.y*s[0][1] + t2.z*s[0][2];
    dMx[1][3] -= t0.w*s[0][0] + t1.w*s[0][1] + t2.w*s[0][2];

                                    //     A^T        A
    t0.x = s[0][2]*s[1][3];         //  = A0y*A1w   A2x*A3y
    t0.y = s[0][3]*s[1][2];         //  = A0w*A1z   A2x*A2y
    t0.z = s[0][1]*s[1][3];         //  = A0y*A1w   A1x*A3y
    t0.w = s[0][2]*s[1][1];         //  = A0z*A1y   A2x*A1y

    t1.x = s[0][3]*s[1][1];         // += A0w*A1y   A3x*A1y
    t1.y = s[0][0]*s[1][3];         // += A0x*A1w   A0x*A3y
    t1.z = s[0][3]*s[1][0];         // += A0w*A1x   A3x*A0y
    t1.w = s[0][0]*s[1][2];         // += A0x*A1z   A0x*A2y

    t2.x = s[0][1]*s[1][2];         // += A0y*A1z   A1x*A2y
    t2.y = s[0][2]*s[1][0];         // += A0z*A1x   A2x*A0y
    t2.z = s[0][0]*s[1][1];         // += A0x*A1y   A0x*A1y
    t2.w = s[0][1]*s[1][0];         // += A0y*A1x   A1x*A0y

    dMx[2][0]  = t0.x*s[3][1] + t1.x*s[3][2] + t2.x*s[3][3];
    dMx[2][0] -= t0.y*s[3][1] + t0.z*s[3][2] + t0.w*s[3][3];

    dMx[2][1]  = t0.y*s[3][0] + t1.y*s[3][2] + t2.y*s[3][3];
    dMx[2][1] -= t0.x*s[3][0] + t1.z*s[3][2] + t1.w*s[3][3];

    dMx[2][2]  = t0.z*s[3][0] + t1.z*s[3][1] + t2.z*s[3][3];
    dMx[2][2] -= t1.x*s[3][0] + t1.y*s[3][1] + t2.w*s[3][3];

    dMx[2][3]  = t0.w*s[3][0] + t1.w*s[3][1] + t2.w*s[3][2];
    dMx[2][3] -= t2.x*s[3][0] + t2.y*s[3][1] + t2.z*s[3][2];

    dMx[3][0]  = t0.z*s[2][2] + t0.w*s[2][3] + t0.y*s[2][1];
    dMx[3][0] -= t1.x*s[2][2] + t2.x*s[2][3] + t0.x*s[2][1];

    dMx[3][1]  = t1.w*s[2][3] + t0.x*s[2][0] + t1.z*s[2][2];
    dMx[3][1] -= t2.y*s[2][3] + t0.y*s[2][0] + t1.y*s[2][2];

    dMx[3][2]  = t1.y*s[2][1] + t2.w*s[2][3] + t1.x*s[2][0];
    dMx[3][2] -= t1.z*s[2][1] + t2.z*s[2][3] + t0.z*s[2][0];

    dMx[3][3]  = t2.z*s[2][2] + t2.x*s[2][0] + t2.y*s[2][1];
    dMx[3][3] -= t2.w*s[2][2] + t0.w*s[2][0] + t1.w*s[2][1];

        // calculate determinant

    fDet = s[0][0]*dMx[0][0] + s[0][1]*dMx[0][1] + s[0][2]*dMx[0][2] + s[0][3]*dMx[0][3];
    bRet = (0.0f != fDet);

    if(bRet)
      {
        fDet = 1.0f / fDet;
        vmp_MatrixScalarGeneric( dMx, dMx, fDet );
      }

    return bRet;
}





void vmp_QVecApplyMatrixGeneric( vmp3DQVector *pvD, const vmp3DQVector * const pvA, const vmp3DMatrix bMx )
{
    vmp3DQVector vD;

    ASSERT_PTR4(pvD);
    ASSERT_PTR4(pvA);
    ASSERT_PTR4(bMx);

    vD.x   = (bMx[0][0] * pvA->x) + (bMx[1][0] * pvA->y) + (bMx[2][0] * pvA->z) + (bMx[3][0] * pvA->w);
    vD.y   = (bMx[0][1] * pvA->x) + (bMx[1][1] * pvA->y) + (bMx[2][1] * pvA->z) + (bMx[3][1] * pvA->w);
    vD.z   = (bMx[0][2] * pvA->x) + (bMx[1][2] * pvA->y) + (bMx[2][2] * pvA->z) + (bMx[3][2] * pvA->w);
    pvD->w = (bMx[0][3] * pvA->x) + (bMx[1][3] * pvA->y) + (bMx[2][3] * pvA->z) + (bMx[3][3] * pvA->w);
    pvD->x = vD.x;
    pvD->y = vD.y;
    pvD->z = vD.z;
}



//
//  Matrix Multiplication   mD = mA x mB
//
void vmp_MatrixMulGenericSlow( vmp3DMatrix dMx, const vmp3DMatrix aMx, const vmp3DMatrix bMx )
{
    vmp3DMatrix mx;                                     // 4 byte aligned is ok for Generic
    uint u, v;

    ASSERT_PTR4(dMx);
    ASSERT_PTR4(aMx);
    ASSERT_PTR4(bMx);

    for ( u=0; u < 4; u++ )
      {
        for ( v = 0; v < 4; v++ )
          {
            mx[u][v] = aMx[u][0] * bMx[0][v]
                     + aMx[u][1] * bMx[1][v]
                     + aMx[u][2] * bMx[2][v]
                     + aMx[u][3] * bMx[3][v];
          }
      }

    vmp_MatrixCopyGeneric(dMx, mx);
}



void vmp_MatrixMulGeneric( vmp3DMatrix dMx, const vmp3DMatrix aMx, const vmp3DMatrix bMx )
{
    vmp3DMatrix mx;                                     // 4 byte aligned is ok for Generic
    uint u;
    vmp3DQVector *pvA, *pvD;
    float *pB;

    ASSERT_PTR4(dMx);
    ASSERT_PTR4(aMx);
    ASSERT_PTR4(bMx);

    pB = (float *)bMx;
    pvA = (vmp3DQVector*)aMx;
    pvD = (vmp3DQVector*)mx;

    for ( u=0; u < 4; u++ )
      {
        pvD->x = (*(pB+0) * pvA->x) + (*(pB+4) * pvA->y) + (*(pB+8) * pvA->z) + (*(pB+12) * pvA->w);
        pvD->y = (*(pB+1) * pvA->x) + (*(pB+5) * pvA->y) + (*(pB+9) * pvA->z) + (*(pB+13) * pvA->w);
        pvD->z = (*(pB+2) * pvA->x) + (*(pB+6) * pvA->y) + (*(pB+10)* pvA->z) + (*(pB+14) * pvA->w);
        pvD->w = (*(pB+3) * pvA->x) + (*(pB+7) * pvA->y) + (*(pB+11)* pvA->z) + (*(pB+15) * pvA->w);
        pvA++;
        pvD++;
      }

    vmp_MatrixCopyGeneric(dMx, mx);
}



//
//  Matrix Apply to Vector mD = mA * vB
//

void vmp_MatrixApplyVecGeneric( vmp3DVector *pvD, vmp3DVector *pvA, const vmp3DMatrix bMx )
{
    vmp3DQVector vD;

    ASSERT_PTR4(pvD);
    ASSERT_PTR4(pvA);
    ASSERT_PTR4(bMx);

    vD.x = pvA->x*bMx[0][0] + pvA->y*bMx[1][0] + pvA->z*bMx[2][0] + bMx[3][0];
    vD.y = pvA->x*bMx[0][1] + pvA->y*bMx[1][1] + pvA->z*bMx[2][1] + bMx[3][1];
    vD.z = pvA->x*bMx[0][2] + pvA->y*bMx[1][2] + pvA->z*bMx[2][2] + bMx[3][2];
    vD.w = pvA->x*bMx[0][3] + pvA->y*bMx[1][3] + pvA->z*bMx[2][3] + bMx[3][3];
    
    if( fabs( vD.w ) < 0.0000001 )                      // Too close to zero?
      {
        pvD->x = vD.x;
        pvD->y = vD.y;
        pvD->z = vD.z;
      }
    else
      {
        pvD->x = vD.x / vD.w;
        pvD->y = vD.y / vD.w;
        pvD->z = vD.z / vD.w;
      }
}



//____________________________________________________________________________
//============================================================================



//                       | 00 01 02 03 |   | 1  0  0  0 |
//  Matrix - Translate   | 10 11 12 13 |   | 0  1  0  0 |
//                     = | 20 21 22 23 | * | 0  0  1  0 |
//                       | 30 31 32 33 |   | Tx Ty Tz 1 |
//
//  NOTE: Copy and Translate
//
//  Assume EMMS is MMX mode upon entry

void vmp_MatrixTranslate( vmp3DMatrix dMx, const vmp3DMatrix aMx, float fTx, float fTy, float fTz )
{
    ASSERT_PTR4(dMx);
    ASSERT_PTR4(aMx);

        // If not same buffer
    if (dMx != aMx)
      {
        vmp_MatrixCopy( dMx, aMx );             // copy the matrix
      }

        //  Now set T{x,y,z}

    vmp_SIMDEntry();    // FPU
    dMx[3][0] += fTx;
    dMx[3][1] += fTy;
    dMx[3][2] += fTz;
    vmp_SIMDExit();     // MMX
}




//                        | 00 01 02 03 |   | 1   0   0   0 |
//  Matrix - Rotation X   | 10 11 12 13 |   | 0  cos sin  0 |
//                      = | 20 21 22 23 | * | 0 -sin cos  0 |
//                        | 30 31 32 33 |   | 0   0   0   1 |
//
//  NOTE: Multiply two Matrices
//
//  Assume EMMS is MMX mode upon entry

void vmp_MatrixRotateX( vmp3DMatrix dMx, const vmp3DMatrix aMx, float fRads )
{
    float matrixbuf[(4*4) +3];  // enough space + 128bit alignment padding
    vmp3DMatrix *mx;

    ASSERT_PTR4(dMx);
    ASSERT_PTR4(aMx);

    mx = (vmp3DMatrix*)ALIGN16((uint)matrixbuf);

    vmp_SIMDEntry();    // FPU
    vmp_MatrixSetRotateX( *mx, fRads );
    vmp_SIMDExit();     // MMX

    vmp_MatrixMul( dMx, aMx, *mx );
}



//                        | 00 01 02 03 |   | cos  0 -sin  0 |
//  Matrix - Rotation Y   | 10 11 12 13 |   |  0   1   0   0 |
//                      = | 20 21 22 23 | * | sin  0  cos  0 |
//                        | 30 31 32 33 |   |  0   0   0   1 |
//
//  Assume EMMS is MMX mode upon entry

void vmp_MatrixRotateY( vmp3DMatrix dMx, const vmp3DMatrix aMx, float fRads )
{
    float matrixbuf[(4*4) +3];  // enough space + 128bit alignment padding
    vmp3DMatrix *mx;

    ASSERT_PTR4(dMx);
    ASSERT_PTR4(aMx);

    mx = (vmp3DMatrix*)ALIGN16((uint)matrixbuf);

    vmp_SIMDEntry();    // FPU
    vmp_MatrixSetRotateY( *mx, fRads );
    vmp_SIMDExit();     // MMX

    vmp_MatrixMul( dMx, aMx, *mx );
}



//                        | 00 01 02 03 |   |  cos sin  0   0 |
//  Matrix - Rotation Z   | 10 11 12 13 |   | -sin cos  0   0 |
//                      = | 20 21 22 23 | * |   0   0   1   0 |
//                        | 30 31 32 33 |   |   0   0   0   1 |
//
//  Assume EMMS is MMX mode upon entry

void vmp_MatrixRotateZ( vmp3DMatrix dMx, const vmp3DMatrix aMx, float fRads )
{
    float matrixbuf[(4*4) +3];  // enough space + 128bit alignment padding
    vmp3DMatrix *mx;

    ASSERT_PTR4(dMx);
    ASSERT_PTR4(aMx);

    mx = (vmp3DMatrix*)ALIGN16((uint)matrixbuf);

    vmp_SIMDEntry();    // FPU
    vmp_MatrixSetRotateZ( *mx, fRads );
    vmp_SIMDExit();     // MMX

    vmp_MatrixMul( dMx, aMx, *mx );
}




