
//
//  util.cpp - Utilities
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//      This is the memory, file, management, module header.
//
//  Win32 Console Application
//
//  Created - 13 Mar, 2002 - James Leiterman
//
//  Last Modification
//
//  stab=4
//



#include    "KariType.h"                                // Master Definitions

    //  3rd Party Files

#include    <windows.h>
#include    <assert.h>
#include    <stdio.h>
#include    <tchar.h>

    //  Project Files

#include    "ram.h"                                     // Memory Module
#include    "util.h"
#include    "Bench.h"



double exptbl[] =       // -e
{
    1.0,                    0.1,                    0.01,                   0.001,
    0.0001,                 0.00001,                0.000001,               0.0000001,
    0.00000001,             0.000000001,            0.0000000001,           0.00000000001,
    0.000000000001,         0.0000000000001,        0.00000000000001,       0.000000000000001,
    0.0000000000000001,     0.00000000000000001,    0.000000000000000001,   0.0000000000000000001,
    0.00000000000000000001,
};



//
//  ASCII to Double Conversion (X86)
//
//  FAST atof()
//
//  This function uses the BCD (Binary Coded Decimal) functionality of the X86 FPU chip
//  which automatically converts 10byte BCD into 10byte Floating Point. It is then written 
//  out as a Double-Precision which is handled automatically. The product of 10^-e with this
//  value calculates the final result!

double ASCIItoDouble( const char *pStr )
{
#ifdef CC_VMP_WIN32
    unsigned int dig[80], *pd;
    unsigned char bcd[12], *pb;
    double f;
    int     n, e;
    const char *p;

    ASSERT_PTR(pStr);

    *(((uint32*)bcd)+0) = 0;
    *(((uint32*)bcd)+1) = 0;
    *(((uint32*)bcd)+2) = 0;

        //  Collect negative/positive

    p = pStr;

    if ('-' == *p)
      {
        *(bcd+9) = 0x80;                                // Blend the Negative Bit into the BCD
        p++;
      }

        //  Collect digits & remember position of decimal point

    *dig = 0;
    e = n = 0;  
    pd = dig+1;

    while (('0' <= *p) && (*p <= '9'))
      {
        *pd++ = (*p++ - '0');                           // Collect a digit (leave in integer form for 'C' speed)
        n++;                                            // character count

        if ('.' == *p)
          {
            e = n;                                      // Remember decimal place position
            p++;
          }
      }

        //  Check for a really BIG (and thus ridiculous) number

    if (n > 18)                                         // More than 18 digits?
      {
        return atof(pStr);
      }

    if (e)      // 0=1.0  1=0.1  2=0.01  3=0.001, etc.
      {
        e = n - e;                                      // Get correct exponent!
      }

        //  repack into bcd (preset lead zeros)
        //  last to first digit

    n = ((n+1)>>1);                                     // Start in middle of bcd buffer
    pb = bcd;                                           // Calc. 1st bcd character position
    
    while(n--)                                          // loop for digit pairs
      {
        pd-=2;                                          // Roll back to last 2 digits
        *pb++ = (byte)((*(pd+0)<<4) | *(pd+1));         // blend two digits into BCD
      }

    __asm {
        fbld    tbyte ptr bcd                           ; Load (10 byte) BCD 
        fstp    f                                       ; Write 64bit Double-Precision Float
    }

//  return f * pow( 10.0, (double) -e );
    return f * exptbl[e];
#else
    ASSERT_PTR(pStr);

    return atof(pStr);
#endif
}



//
//  ASCII to Double Conversion (X86) w/Self Advancing Pointer
//
//  FASTER atof()
//
//  This function uses the BCD (Binary Coded Decimal) functionality of the X86 FPU chip
//  which automatically converts 10byte BCD into 10byte Floating Point. It is then written 
//  out as a Double-Precision which is handled automatically. The product of 10^-e with this
//  value calculates the final result!
//
//  NOTE: This would SCREAM if it was in Assembly Language!

double pASCIItoDouble( const char **ppStr )
{
#ifdef CC_VMP_WIN32
    unsigned int dig[80], *pd;
    unsigned char bcd[12], *pb;
    double f;
    int     n, e;
    const char *p;

    ASSERT_PTR(ppStr);
    ASSERT_PTR(*ppStr);

    *(((uint32*)bcd)+0) = 0;
    *(((uint32*)bcd)+1) = 0;
    *(((uint32*)bcd)+2) = 0;

        //  Collect negative/positive

    p = *ppStr;

    if ('-' == *p)
      {
        *(bcd+9) = 0x80;                                // Blend the Negative Bit into the BCD
        p++;
      }

        //  Collect digits & remember position of decimal point

    *dig = 0;
    e = n = 0;  
    pd = dig+1;

    while (('0' <= *p) && (*p <= '9'))
      {
        *pd++ = (*p++ - '0');                           // Collect a digit (leave in integer form for 'C' speed)
        n++;                                            // character count

        if ('.' == *p)
          {
            e = n;                                      // Remember decimal place position
            p++;
          }
      }

        //  Check for a really BIG (and thus ridiculous) number

    if (n > 18)                                         // More than 18 digits?
      {
        return atof(*ppStr);
      }

    *ppStr = p;                                         // Advanced pointer

    if (e)      // 0=1.0  1=0.1  2=0.01  3=0.001, etc.
      {
        e = n - e;                                      // Get correct exponent!
      }

        //  repack into bcd (preset lead zeros)
        //  last to first digit

    n = ((n+1)>>1);                                     // Start in middle of bcd buffer
    pb = bcd;                                           // Calc. 1st bcd character position
    
    while(n--)                                          // loop for digit pairs
      {
        pd-=2;                                          // Roll back to last 2 digits
        *pb++ = (byte)((*(pd+0)<<4) | *(pd+1));         // blend two digits into BCD
      }

    __asm {
        fbld    tbyte ptr bcd                           ; Load (10 byte) BCD 
        fstp    f                                       ; Write 64bit Double-Precision Float
    }

//  return f * pow( 10.0, (double) -e );
    return f * exptbl[e];
#else
    ASSERT_PTR(pStr);

    return atof(pStr);
#endif
}



//  Load file into memory - *** Win32 Version ***
//
//  Sets pointer and returns size of file loaded

bool fileLoad( const char * const pFName, byte **ppMem, ulong * const uipSize )
{
    HANDLE  hFile;
    DWORD   nSize, dw;
    bool    ret;
    
    ASSERT_PTR(pFName);
    ASSERT_PTR(ppMem);
    ASSERT_PTR(uipSize);

        // Open a file

    ret = false;
    *ppMem = NULL;
    *uipSize = 0;

    hFile = CreateFile( (const _TCHAR *) pFName, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );

    if (INVALID_HANDLE_VALUE != hFile)                  // Successful open
      {
        nSize = GetFileSize( hFile, NULL );

        if ( !nSize )
          {
            ret = true;                                 // * Success * but an empty file
          }
        else
          {
            if (ramGetClr(ppMem, nSize+1))              // one extra byte for a (string) terminator
              {
                if ( ReadFile( hFile, *ppMem, nSize, &dw, NULL ) && (nSize == dw) )
                  {
                    *((*ppMem)+nSize) = '\0';           // Set a terminator in extra byte allocated!

                    *uipSize = (uint) nSize;
                    ret = true;                         // *** SUCCESS ***
                  }
                else
                  {
                    ramRelease( ppMem );                // Release Memory if a file problem!
                    *ppMem = NULL;
                  }
              }
          }

        CloseHandle( hFile );
      }

    return ret;
}



//
//  Is Delimiter
//

bool isdelim( char c )
{
    if (isspace(c) || (',' == c))
      {
        return true;
      }
    else
      {
        return false;
      }
}



//
//  Skip White Space
//

char *skipdelim( const char *pStr )
{
    ASSERT_PTR(pStr);

    while (*pStr)
      {
        if (!isdelim(*pStr))                            // 0x09-0x0d, 0x20
          {
            break;
          }
            
        pStr++;                                         // Next character
      }

    return (char *)pStr;
}



//
//  Skip Word
//
char *skipword( const char *pStr )
{
    ASSERT_PTR(pStr);

    while (*pStr)
      {
        if (isdelim(*pStr)) // 0x09-0x0d, 0x20
          {
            break;
          }

        pStr++;                                         // Next character
      }

    return (char *)pStr;
}



//
//  Get Word
//
char *getword( char *pszBuf, const char *pszStr )
{
    ASSERT_PTR(pszBuf);
    ASSERT_PTR(pszStr);

    while (*pszStr)
      {
        if (isdelim(*pszStr))                           // 0x09-0x0d, 0x20,
          {
            break;
          }

        *pszBuf++ = *pszStr++;                          // Copy character
      }

    *pszBuf = '\0';                                     // Set terminator for substring in buffer

    return (char *)pszStr;                              // 1st char past word
}





