        TITLE   Book MCompX86.asm - Motion Compensation
        PAGE    53,160

;
; MCompX86.asm - 
;
; This document contains file that you may use at your own risk in your own programs.
;
; This module is designed to accomodate X86 processors in the book.
;
;       NOTE: Needs masm 6.14 or later!
;
;       Created - 21 Feb, 2002 - J.Leiterman
;
;       Modifications
;
;               
;
;       Tabs = 8



        .XLIST

        .387
        .686P
        .MMX
        .model flat, C

        include x86.inc
;        include CpuX86.inc
        include KariType.inc
;        include Prolog32.inc
 
        .LIST



        .data

ZeroD   dq      0000000000000000h
OneHD   dq      0001000100010001h



MCompX86  SEGMENT USE32 PUBLIC 'CODE'



;_________________________________________________________________________________________________
;=================================================================================================
;        8x8 Horizontal Motion Compensation w/Rounding
;
;       NOTE: Stride (+) is top down and (-) is bottom up orientation
;               Thus Src and Dst may be beginning of (+=)first or (-=)last scanline.
;
;void vmp_MCompHz8R( uint8 * pbDst, const uint8 * pbSrc, int iStride )
;{
;       uint dy, dx;
;
;       for (dy = 0; dy < 8; dy++)
;         {
;               for (dx = 0; dx < 8; dx++)
;                 {
;                       pbDst[dx] = (unsigned char) ((pbSrc[dx] + pbSrc[dx+1] + 1) >> 1);
;                 }
;
;               pbSrc += iStride;               // +- to beginning next scanline!
;               pbDst += iStride;
;       }
;}

;       *** MMX version *** 64bit (un)aligned

        align   16

        .MMX
        public  vmp_MCompHz8RAsmMMX
vmp_MCompHz8RAsmMMX       proc    near pbDst:ptr, pbSrc:ptr, iStride:DWORD
;;;        femms
ASSERT_PTR(pbDst)
ASSERT_PTR(pbSrc)
ASSERT_ZERO(iStride)

        mov     eax,pbSrc       ; Source
        mov     edx,pbDst       ; Destination
        mov     ecx,8           ; # of scanlines

        movq    mm7,ZeroD
        movq    mm6,OneHD

next:   movq    mm0,[eax+0]     ; [01234567]
        movq    mm1,[eax+1]     ; [12345678]

;;;     pavgusb mm0,mm1         ; (A + B + 1)>>1
        movq    mm2,mm0
        movq    mm3,mm1
        punpcklbw mm0,mm7       ; 0:A  Low
        punpcklbw mm1,mm7       ; 0:B  Low
        punpckhbw mm2,mm7       ; 0:A  High
        punpckhbw mm3,mm7       ; 0:B  High     

        paddusw mm0,mm1         ; A+B Low
        paddusw mm2,mm3         ; A+B High
        paddusw mm0,mm6         ; +1 Low
        paddusw mm2,mm6         ; +1 High
        psrlw   mm0,1           ; >>1
        psrlw   mm2,1           ; >>1
        packuswb mm0,mm2        ; H...L 

        movq    [edx],mm0       ; Save 64bit (8byte) result

        add     eax,iStride
        add     edx,iStride
        dec     ecx
        jne     next
        
;;;      femms
        ret
vmp_MCompHz8RAsmMMX endp



;_________________________________________________________________________________________________
;       *** 3DNow version *** 64bit (un)aligned

        align   16

        .K3D
        public  vmp_MCompHz8RAsm3DNow
vmp_MCompHz8RAsm3DNow       proc    near pbDst:ptr, pbSrc:ptr, iStride:DWORD
;;;        femms
ASSERT_PTR(pbDst)
ASSERT_PTR(pbSrc)
ASSERT_ZERO(iStride)

        mov     eax,pbSrc       ; Source
        mov     edx,pbDst       ; Destination
        mov     ecx,8           ; # of scanlines

next:   movq    mm0,[eax+0]     ; [01234567]
        movq    mm1,[eax+1]     ; [12345678]

        pavgusb mm0,mm1         ; (A + B + 1)>>1

        movq    [edx],mm0       ; Save 64bit (8byte) result

        add     eax,iStride
        add     edx,iStride
        dec     ecx
        jne     next
        
;;;      femms
        ret
vmp_MCompHz8RAsm3DNow endp



;_________________________________________________________________________________________________
;       *** SSE version *** 64bit (un)aligned

        align   16
        .XMM
        public  vmp_MCompHz8RAsmSSE
vmp_MCompHz8RAsmSSE       proc    near pbDst:ptr, pbSrc:ptr, iStride:DWORD
;;;        emms
ASSERT_PTR(pbDst)
ASSERT_PTR(pbSrc)
ASSERT_ZERO(iStride)

        mov     eax,pbSrc       ; Source
        mov     edx,pbDst       ; Destination
        mov     ecx,8           ; # of scanlines

next:   movq    mm0,[eax+0]     ; [01234567]
        movq    mm1,[eax+1]     ; [12345678]

        pavgb   mm0,mm1         ; (A + B + 1)>>1

        movq    [edx],mm0       ; Save 64bit (8byte) result

        add     eax,iStride
        add     edx,iStride
        dec     ecx
        jne     next
        
;;;        emms
        ret
vmp_MCompHz8RAsmSSE endp



;_________________________________________________________________________________________________
;=================================================================================================
;        16x16 Horizontal Motion Compensation w/Rounding
;
;       NOTE: Stride (+) is top down and (-) is bottom up orientation
;               Thus Src and Dst may be beginning of (+=)first or (-=)last scanline.
;
;void vmp_MCompHz16R( uint8 * pbDst, const uint8 * pbSrc, int iStride )
;{
;       uint dy, dx;
;
;       for (dy = 0; dy < 16; dy++)
;         {
;               for (dx = 0; dx < 16; dx++)
;                 {
;                       pbDst[dx] = (unsigned char) ((pbSrc[dx] + pbSrc[dx+1] + 1) >> 1);
;                 }
;
;               pbSrc += iStride;               // +- to beginning next scanline!
;               pbDst += iStride;
;       }
;}

;       *** MMX version *** 128bit (un)aligned

        align   16

        .MMX
        public  vmp_MCompHz16RAsmMMX
vmp_MCompHz16RAsmMMX       proc    near pbDst:ptr, pbSrc:ptr, iStride:DWORD
;;;        femms
ASSERT_PTR(pbDst)
ASSERT_PTR(pbSrc)
ASSERT_ZERO(iStride)

        mov     eax,pbSrc       ; Source
        mov     edx,pbDst       ; Destination
        mov     ecx,16          ; # of scanlines

        movq    mm7,ZeroD
        movq    mm6,OneHD

                ;       Lower 64bits

next:   movq    mm0,[eax+0]     ; [01234567]
        movq    mm1,[eax+1]     ; [12345678]

                ; (A + B + 1)>>1
        movq    mm2,mm0
        movq    mm3,mm1
        punpcklbw mm0,mm7       ; 0:A  Low
        punpcklbw mm1,mm7       ; 0:B  Low
        punpckhbw mm2,mm7       ; 0:A  High
        punpckhbw mm3,mm7       ; 0:B  High     

        paddusw mm0,mm1         ; A+B Low
        paddusw mm2,mm3         ; A+B High

        paddusw mm0,mm6         ; +1 Low
        paddusw mm2,mm6         ; +1 High

                ;       Upper 64bits
                
        movq    mm4,[eax+8]     ; [89ABCDEF]
        movq    mm5,[eax+8+1]   ; [9ABCDEF 10]

                ; (A + B + 1)>>1
        movq    mm1,mm4
        movq    mm3,mm5
        punpcklbw mm4,mm7       ; 0:A  Low
        punpcklbw mm5,mm7       ; 0:B  Low
        punpckhbw mm1,mm7       ; 0:A  High
        punpckhbw mm3,mm7       ; 0:B  High     

        paddusw mm4,mm5         ; A+B Low
        paddusw mm1,mm3         ; A+B High

        paddusw mm4,mm6         ; +1 Low
        paddusw mm1,mm6         ; +1 High

        psrlw   mm0,1           ; >>1
        psrlw   mm2,1           ; >>1
        psrlw   mm4,1           ; >>1
        psrlw   mm1,1           ; >>1

        packuswb mm0,mm2        ; H...L 
        packuswb mm4,mm1        ; H...L 

        movq    [edx+0],mm0     ; Save lower 64bit (8byte) result
        movq    [edx+8],mm4     ; Save upper 64bit (8byte) result

        add     eax,iStride
        add     edx,iStride
        dec     ecx
        jne     next
        
;;;      femms
        ret
vmp_MCompHz16RAsmMMX endp



;_________________________________________________________________________________________________
;       *** 3DNow version *** 128bit (un)aligned

        align   16

        .K3D
        public  vmp_MCompHz16RAsm3DNow
vmp_MCompHz16RAsm3DNow       proc    near pbDst:ptr, pbSrc:ptr, iStride:DWORD
;;;        femms
ASSERT_PTR(pbDst)
ASSERT_PTR(pbSrc)
ASSERT_ZERO(iStride)

        mov     eax,pbSrc       ; Source
        mov     edx,pbDst       ; Destination
        mov     ecx,16          ; # of scanlines

next:   movq    mm0,[eax+0]     ; [01234567]
        movq    mm2,[eax+1]     ; [12345678]
        movq    mm1,[eax+8]     ; [89ABCDEF]
        movq    mm3,[eax+8+1]   ; [9ABCDEF 10]

        pavgusb mm0,mm2         ; (A + B + 1)>>1
        pavgusb mm1,mm3         ; (A + B + 1)>>1

        movq    [edx+0],mm0     ; Save lower 64bits (8byte) result
        movq    [edx+8],mm1     ; Save upper 64bits (8byte) result

        add     eax,iStride
        add     edx,iStride
        dec     ecx
        jne     next
        
;;;      femms
        ret
vmp_MCompHz16RAsm3DNow endp



;_________________________________________________________________________________________________
;       *** SSE version *** 128bit (un)aligned

        align   16
        .XMM
        public  vmp_MCompHz16RAsmSSE
vmp_MCompHz16RAsmSSE       proc    near pbDst:ptr, pbSrc:ptr, iStride:DWORD
;;;        emms
ASSERT_PTR(pbDst)
ASSERT_PTR(pbSrc)
ASSERT_ZERO(iStride)

        mov     eax,pbSrc       ; Source
        mov     edx,pbDst       ; Destination
        mov     ecx,16          ; # of scanlines

next:   movq    mm0,[eax+0]     ; [01234567]
        movq    mm2,[eax+1]     ; [12345678]
        movq    mm1,[eax+8]     ; [89ABCDEF]
        movq    mm3,[eax+8+1]   ; [9ABCDEF 10]

        pavgb   mm0,mm2         ; (A + B + 1)>>1
        pavgb   mm1,mm3         ; (A + B + 1)>>1

        movq    [edx+0],mm0     ; Save lower 64bits (8byte) result
        movq    [edx+8],mm1     ; Save upper 64bits (8byte) result

        add     eax,iStride
        add     edx,iStride
        dec     ecx
        jne     next
        
;;;        emms
        ret
vmp_MCompHz16RAsmSSE endp



;_________________________________________________________________________________________________
;       *** SSE2 version *** 128bit (un)aligned

        align   16
        .XMM
        public  vmp_MCompHz16RAsmSSE2
vmp_MCompHz16RAsmSSE2       proc    near pbDst:ptr, pbSrc:ptr, iStride:DWORD
;;;        emms
ASSERT_PTR(pbDst)
ASSERT_PTR(pbSrc)
ASSERT_ZERO(iStride)

        mov     eax,pbSrc       ; Source
        mov     edx,pbDst       ; Destination
        mov     ecx,16          ; # of scanlines

next:   movdqu  xmm0,[eax+0]    ; [0123456789ABCDEF]
        movdqu  xmm1,[eax+1]    ; [123456789ABCDEF 10]
        add     eax,iStride

        pavgb   xmm0,xmm1       ; (A + B + 1)>>1

        movdqu  [edx],xmm0      ; Save 128bit (16byte) result
        add     edx,iStride
        dec     ecx
        jne     next
        
;;;        emms
        ret
vmp_MCompHz16RAsmSSE2 endp



;_________________________________________________________________________________________________
;=================================================================================================
;        8x8 Horizontal Motion Compensation w/Rounding
;
;       NOTE: Stride (+) is top down and (-) is bottom up orientation
;               Thus Src and Dst may be beginning of (+=)first or (-=)last scanline.
;
;void vmp_MCompVt8R( uint8 * pbDst, const uint8 * pbSrc, int iStride )
;{
;       uint dy, dx;
;
;       for (dy = 0; dy < 8; dy++)
;         {
;               for (dx = 0; dx < 8; dx++)
;                 {
;                       pbDst[dx] = (unsigned char) ((pbSrc[dx] + pbSrc[dx+iStride] + 1) >> 1);
;                 }
;
;               pbSrc += iStride;               // +- to beginning next scanline!
;               pbDst += iStride;
;       }
;}

;       *** MMX version *** 64bit (un)aligned

        align   16

        .MMX
        public  vmp_MCompVt8RAsmMMX
vmp_MCompVt8RAsmMMX       proc    near pbDst:ptr, pbSrc:ptr, iStride:DWORD
;;;        femms
ASSERT_PTR(pbDst)
ASSERT_PTR(pbSrc)
ASSERT_ZERO(iStride)

        movq    mm7,ZeroD       ; {0 0 0 0}
        movq    mm6,OneHD       ; {1 1 1 1}

        mov     eax,pbSrc       ; Source
        mov     edx,pbDst       ; Destination
        mov     ecx,8           ; # of scanlines

next:   movq    mm0,[eax]       ; [01234567]
        add     eax,iStride
        movq    mm1,[eax]       ; [01234567]

                ; (A + B + 1)>>1
        movq    mm2,mm0
        movq    mm3,mm1
        punpcklbw mm0,mm7       ; 0:A  Low
        punpcklbw mm1,mm7       ; 0:B  Low
        punpckhbw mm2,mm7       ; 0:A  High
        punpckhbw mm3,mm7       ; 0:B  High
        paddusw mm0,mm1         ; A+B Low
        paddusw mm2,mm3         ; A+B High
        paddusw mm0,mm6         ; +1 Low
        paddusw mm2,mm6         ; +1 High
        psrlw   mm0,1           ; >>1
        psrlw   mm2,1           ; >>1
        packuswb mm0,mm2        ; H...L 

        movq    [edx],mm0       ; Save 64bit (8byte) result
        add     edx,iStride

        dec     ecx
        jne     next
        
;;;      femms
        ret
vmp_MCompVt8RAsmMMX endp



;_________________________________________________________________________________________________
;       *** 3DNow version *** 64bit (un)aligned

        align   16

        .K3D
        public  vmp_MCompVt8RAsm3DNow
vmp_MCompVt8RAsm3DNow       proc    near pbDst:ptr, pbSrc:ptr, iStride:DWORD
;;;        femms
ASSERT_PTR(pbDst)
ASSERT_PTR(pbSrc)
ASSERT_ZERO(iStride)

        mov     eax,pbSrc       ; Source
        mov     edx,pbDst       ; Destination

        mov     ecx,8/2         ; # of scanlines

        movq    mm0,[eax]       ; [01234567]
        add     eax,iStride

next:   movq    mm1,[eax]       ; [01234567]
        add     eax,iStride

        pavgusb mm0,mm1         ; (A + B + 1)>>1

        movq    [edx],mm0       ; Save 64bit (8byte) result
        add     edx,iStride

        movq    mm0 ,[eax]
        add     eax,iStride

        pavgusb mm1,mm0         ; (A + B + 1)>>1

        movq    [edx],mm1       ; Save 64bit (8byte) result
        add     edx,iStride

        dec     ecx
        jne     next
        
;;;      femms
        ret
vmp_MCompVt8RAsm3DNow endp



;_________________________________________________________________________________________________
;       *** SSE version *** 64bit (un)aligned

        align   16
        .XMM
        public  vmp_MCompVt8RAsmSSE
vmp_MCompVt8RAsmSSE       proc    near pbDst:ptr, pbSrc:ptr, iStride:DWORD
;;;        emms
ASSERT_PTR(pbDst)
ASSERT_PTR(pbSrc)
ASSERT_ZERO(iStride)

        mov     eax,pbSrc       ; Source
        mov     edx,pbDst       ; Destination
        mov     ecx,8/2         ; # of scanlines

        movq    mm0,[eax]       ; [01234567]
        add     eax,iStride

next:   movq    mm1,[eax]       ; [01234567]
        add     eax,iStride

        pavgb   mm0,mm1         ; (A + B + 1)>>1

        movq    [edx],mm0       ; Save 64bit (8byte) result
        add     edx,iStride

        movq    mm0 ,[eax]
        add     eax,iStride

        pavgb   mm1,mm0         ; (A + B + 1)>>1

        movq    [edx],mm1       ; Save 64bit (8byte) result
        add     edx,iStride
                
        dec     ecx
        jne     next
        
;;;        emms
        ret
vmp_MCompVt8RAsmSSE endp



;_________________________________________________________________________________________________
;=================================================================================================
;        16x16 Vertical Motion Compensation w/Rounding
;
;       NOTE: Stride (+) is top down and (-) is bottom up orientation
;               Thus Src and Dst may be beginning of (+=)first or (-=)last scanline.
;
;void vmp_MCompVt16R( uint8 * pbDst, const uint8 * pbSrc, int iStride )
;{
;       uint dy, dx;
;
;       for (dy = 0; dy < 16; dy++)
;         {
;               for (dx = 0; dx < 16; dx++)
;                 {
;                       pbDst[dx] = (unsigned char) ((pbSrc[dx] + pbSrc[dx+iStride] + 1) >> 1);
;                 }
;
;               pbSrc += iStride;               // +- to beginning next scanline!
;               pbDst += iStride;
;       }
;}

;       *** MMX version *** 128bit (un)aligned

        align   16

        .MMX
        public  vmp_MCompVt16RAsmMMX
vmp_MCompVt16RAsmMMX       proc    near pbDst:ptr, pbSrc:ptr, iStride:DWORD
;;;        femms
ASSERT_PTR(pbDst)
ASSERT_PTR(pbSrc)
ASSERT_ZERO(iStride)

        movq    mm7,ZeroD       ; {0 0 0 0}
        movq    mm6,OneHD       ; {1 1 1 1}

        mov     eax,pbSrc       ; Source
        mov     edx,pbDst       ; Destination
        mov     ecx,16          ; # of scanlines

next:   movq    mm0,[eax+0]     ; [01234567]
        movq    mm1,[eax+8]     ; [89ABCDEF]
        add     eax,iStride
        movq    mm3,[eax+0]     ; [01234567]
        movq    mm5,[eax+8]     ; [89ABCDEF]

                ; (A + B + 1)>>1
        movq    mm2,mm0
        movq    mm4,mm3
        punpcklbw mm0,mm7       ; 0:AL  Low
        punpcklbw mm3,mm7       ; 0:BL  Low
        punpckhbw mm2,mm7       ; 0:AL  High
        punpckhbw mm4,mm7       ; 0:BL  High

        paddusw mm0,mm3         ; AL+BL Low
        movq    mm3,mm1
        paddusw mm2,mm4         ; AH+BH High
        movq    mm4,mm5

        punpcklbw mm1,mm7       ; 0:AH  Low
        punpcklbw mm5,mm7       ; 0:BH  Low
        punpckhbw mm3,mm7       ; 0:AH  High
        punpckhbw mm4,mm7       ; 0:BH  High
        paddusw mm1,mm5         ; AH+BH Low
        paddusw mm3,mm4         ; AH+BH High

        paddusw mm0,mm6         ; +1 Low (Lower 64bits)
        paddusw mm2,mm6         ; +1 High
        paddusw mm1,mm6         ; +1 Low (Upper 64bits)
        paddusw mm3,mm6         ; +1 High

        psrlw   mm0,1           ; >>1 Low
        psrlw   mm2,1           ; >>1 High
        psrlw   mm1,1           ; >>1 Low
        psrlw   mm3,1           ; >>1 High

        packuswb mm0,mm2        ; H...L 
        packuswb mm1,mm3        ; H...L 

        movq    [edx+0],mm0     ; Save Lower 64bit (8byte) result
        movq    [edx+8],mm1     ; Save Upper 64bit (8byte) result
        add     edx,iStride

        dec     ecx
        jne     next
        
;;;      femms
        ret
vmp_MCompVt16RAsmMMX endp



;_________________________________________________________________________________________________
;       *** 3DNow version *** 128bit (un)aligned

        align   16

        .K3D
        public  vmp_MCompVt16RAsm3DNow
vmp_MCompVt16RAsm3DNow       proc    near pbDst:ptr, pbSrc:ptr, iStride:DWORD
;;;        femms
ASSERT_PTR(pbDst)
ASSERT_PTR(pbSrc)
ASSERT_ZERO(iStride)

        mov     eax,pbSrc       ; Source
        mov     edx,pbDst       ; Destination
        mov     ecx,16/2                ; # of scanlines /2

        movq    mm0,[eax+0]     ; [01234567]
        movq    mm2,[eax+8]     ; [01234567]
        add     eax,iStride

next:   movq    mm1,[eax+0]     ; [01234567]
        movq    mm3,[eax+8]     ; [01234567]
        add     eax,iStride

        pavgusb mm0,mm1         ; (A + B + 1)>>1
        pavgusb mm2,mm3         ; (A + B + 1)>>1

        movq    [edx+0],mm0     ; Save lower 64bit (8byte) result
        movq    [edx+8],mm2     ; Save upper 64bit (8byte) result
        add     edx,iStride

        movq    mm0 ,[eax+0]
        movq    mm2 ,[eax+8]
        add     eax,iStride

        pavgusb mm1,mm0         ; (A + B + 1)>>1
        pavgusb mm3,mm2         ; (A + B + 1)>>1

        movq    [edx+0],mm1     ; Save lower 64bit (8byte) result
        movq    [edx+8],mm3     ; Save upper 64bit (8byte) result
        add     edx,iStride
                
        dec     ecx
        jne     next
        
;;;      femms
        ret
vmp_MCompVt16RAsm3DNow endp



;_________________________________________________________________________________________________
;       *** SSE version *** 128bit (un)aligned

        align   16
        .XMM
        public  vmp_MCompVt16RAsmSSE
vmp_MCompVt16RAsmSSE       proc    near pbDst:ptr, pbSrc:ptr, iStride:DWORD
;;;        emms
ASSERT_PTR(pbDst)
ASSERT_PTR(pbSrc)
ASSERT_ZERO(iStride)

        mov     eax,pbSrc       ; Source
        mov     edx,pbDst       ; Destination
        mov     ecx,16/2                ; # of scanlines

        movq    mm0,[eax+0]     ; [01234567]
        movq    mm2,[eax+8]     ; [01234567]
        add     eax,iStride

next:   movq    mm1,[eax+0]     ; [01234567]
        movq    mm3,[eax+8]     ; [01234567]
        add     eax,iStride

        pavgb   mm0,mm1         ; (A + B + 1)>>1
        pavgb   mm2,mm3         ; (A + B + 1)>>1

        movq    [edx+0],mm0     ; Save lower 64bit (8byte) result
        movq    [edx+8],mm2     ; Save upper 64bit (8byte) result
        add     edx,iStride

        movq    mm0 ,[eax+0]
        movq    mm2 ,[eax+8]
        add     eax,iStride

        pavgb   mm1,mm0         ; (A + B + 1)>>1
        pavgb   mm3,mm2         ; (A + B + 1)>>1

        movq    [edx+0],mm1     ; Save lower 64bit (8byte) result
        movq    [edx+8],mm3     ; Save upper 64bit (8byte) result
        add     edx,iStride
                
        dec     ecx
        jne     next
        
;;;        emms
        ret
vmp_MCompVt16RAsmSSE endp



;_________________________________________________________________________________________________
;       *** SSE2 version *** 128bit (un)aligned

        align   16
        .XMM
        public  vmp_MCompVt16RAsmSSE2
vmp_MCompVt16RAsmSSE2       proc    near pbDst:ptr, pbSrc:ptr, iStride:DWORD
;;;        emms
ASSERT_PTR(pbDst)
ASSERT_PTR(pbSrc)
ASSERT_ZERO(iStride)

        mov     eax,pbSrc       ; Source
        mov     edx,pbDst       ; Destination
        mov     ecx,16/2        ; # of scanlines

        movdqu  xmm0,[eax]      ; [0123456789ABCDEF]
        add     eax,iStride

next:   movdqu  xmm1,[eax]      ; [0123456789ABCDEF]
        add     eax,iStride

        pavgb   xmm0,xmm1       ; (A + B + 1)>>1

        movdqu  [edx],xmm0      ; Save 128bit (16byte) result
        add     edx,iStride

        movdqu  xmm0,[eax]      ; [0123456789ABCDEF]
        add     eax,iStride

        pavgb   xmm1,xmm0       ; (A + B + 1)>>1

        movdqu  [edx],xmm1      ; Save 128bit (16byte) result
        add     edx,iStride

        dec     ecx
        jne     next
        
;;;        emms
        ret
vmp_MCompVt16RAsmSSE2 endp



MCompX86  ends
        end

