        TITLE   Book yuv12X86.asm - YUV12 to RGB
        PAGE    53,160

;
; yuv12X86.asm - 
;
; This document contains file that you may use at your own risk in your own programs.
;
; This module is designed to accomodate X86 processors in the book.
;
;       NOTE: Needs masm 6.14 or later!
;
;       Created - 25 Feb, 2002 - J.Leiterman
;
;       Modifications
;
;               
;
;       Tabs = 8



        .XLIST

        .387
        .686P
        .MMX
        .model flat, C

        include x86.inc
;        include CpuX86.inc
        include KariType.inc
;        include Prolog32.inc
 
        .LIST



        .data

ZeroD   dq      00000000000000000h,00000000000000000h

ColorY  dq      02568256825682568h,02568256825682568h   ; 2568h
ColorUR dq      03343334333433343h,03343334333433343h   ; 3343h
ColorUG dq      0E5E2E5E2E2E2E2E2h,0E5E2E5E2E5E2E5E2h   ; E5E2h   (-1a1eh)
ColorVG dq      0F36EF36EF36EF36Eh,0F36EF36EF36EF36Eh   ; F36Eh   (-0c92h)
ColorVB dq      040CF40CF40CF40CFh,040CF40CF40CF40CFh   ; 40CFh

DiffUV  dq      0FC00FC00FC00FC00h,0FC00FC00FC00FC00h   ; -128x8
DiffY   dq      0FF80FF80FF80FF80h,0FF80FF80FF80FF80h   ; -16x8  




yuv12X86  SEGMENT USE32 PUBLIC 'CODE'



;_________________________________________________________________________________________________
;=================================================================================================
;
;       YUV12 to RGB Conversion
;
;void vmp_yuv12toRGB32Generic(
;       void *pImg, uint nWidth, uint nHeight, int iStride,
;       const uint8 * pBaseY, uint nStrideY, 
;       const uint8 * pBaseU, const uint8 * pBaseV, uint nStrideUV)
;

;       *** MMX version *** 64bit (un)aligned

        align   16

        .MMX
        public  vmp_yuv12toRGB32AsmMMX
vmp_yuv12toRGB32AsmMMX  proc    near,
                pImg:ptr, nWidth:DWORD, nHeight:DWORD, iStride:DWORD,
                pBaseY:ptr,             nStrideY:DWORD, 
                pBaseU:ptr, pBaseV:ptr, nStrideUV:DWORD
                        push    ebx
        push    esi
        push    edi

;;;        femms
ASSERT_PTR16(pBaseY)
ASSERT_ZERO(nStrideY)
ASSERT_PTR16(pBaseU)
ASSERT_PTR16(pBaseV)
ASSERT_ZERO(nStrideUV)
ASSERT_PTR16(pImg)
ASSERT_ZERO(nWidth)
ASSERT_ZERO(nHeight)
ASSERT_ZERO(iStride)

        mov     eax,nWidth
        shr     nWidth,3                ; nWidth >>= 3;   n/8

        sub     nStrideY,eax    ; nStrideY -= nWidth
        mov     ebx,pImg
        mov     esi,pBaseY
        shl     eax,2           ; x4
        mov     edi,pBaseU
        mov     edx,pBaseV
        sub     iStride,eax     ; iStride -= (nWidth*4)
        mov     eax,0           ; (Flip Flip)

        ; Vertical Loop

Vrt:    mov     ecx,nWidth

        ; Horizontal Loop

Hrz:    movq    mm0,[edi]       ; [u7 u6 u5 u4 u3 u2 u1 u0] *pU
        add     edi,4
        movq    mm1,[edx]       ; [v7 v6 v5 v4 v3 v2 v1 v0] *pV
        add     edx,4

        punpcklbw mm0,ZeroD     ; [ 0 u3  0 u2  0 u1  0 u0] 0...255
        punpcklbw mm1,ZeroD     ; [ 0 v3  0 v2  0 v1  0 v0]

        movq    mm2,[esi]       ; [y7 y6 y5 y4 y3 y2 y1 y0]
        add     esi,8

        psllw   mm0,3           ; [ U3 U2 U1 U0 ] 0...2040  8*(0...255)
        psllw   mm1,3           ; [ V3 V2 V1 V0 ]

        movq    mm3,mm2         ; [y7 y6 y5 y4 y3 y2 y1 y0]

        paddsb  mm0,DiffUV      ; [ U3 U2 U1 U0 ] (8*u)-(8*128) = 8(u-128)
        paddsb  mm1,DiffUV      ; [ V3 V2 V1 V0 ]     = {-1024...1016}

        punpcklbw mm2,ZeroD     ; [ 0 y3  0 y2  0 y1  0 y0]
        punpckhbw mm3,ZeroD     ; [ 0 y7  0 y6  0 y5  0 y4]
        
        movq    mm6,mm0
        pmulhw  mm0,ColorUR     ; [UR3 UR2 UR1 UR0]  ur =  0x3343 * 8(u-128)
        pmulhw  mm6,ColorUG     ; [UG3 UG2 UG1 UG0]  ug = -0x1a1e * 8(u-128)

        movq    mm4,mm0         ; [UR3 UR2 UR1 UR0]
        movq    mm7,mm1         ; [ V3  V2  V1  V0]

        punpcklwd mm0,mm0               ; [UR1 UR1 UR0 UR0]
        punpckhwd mm4,mm4               ; [UR3 UR3 UR2 UR2]

        psllw   mm2,3           ; [ Y3  Y2  Y1  Y0] (8*y)
        psllw   mm3,3           ; [ Y7  Y6  Y5  Y4]

        pmulhw  mm1,ColorVB     ; [VB3 VB2 VB1 VB0]  vb =  0x40cf * 8(v-128)
        pmulhw  mm7,ColorVG     ; [VG3 VG2 VG1 VG0]  vg = -0x0c92 * 8(v-128)

        movq    mm5,mm1         ; [VB3 VB2 VB1 VB0]
        paddsw  mm2,DiffY       ; [ Y3  Y2  Y1  Y0] (8*y)-(8*16) = 8(y-16)
        paddsw  mm3,DiffY       ; [ Y7  Y6  Y5  Y4] 

        paddsw  mm7,mm6         ; [UG3+VG3 UG2+VG2 UG1+VG1 UG0+VG0]

        punpcklwd mm1,mm1               ; [VB1 VB1 VB0 VB0]
        punpckhwd mm5,mm5               ; [VB3 VB3 VB2 VB2]

        movq    mm6,mm7         ; [UG3+VG3 UG2+VG2 UG1+VG1 UG0+VG0]

        pmulhw  mm2,ColorY      ; [ Y3  Y2  Y1  Y0]  y = 0x2568 * 8(y-16)
        pmulhw  mm3,ColorY      ; [ Y7  Y6  Y5  Y4]  y = 0x2568 * 8(y-16)

        punpcklwd mm7,mm7               ; [ UG1+VG1 UG1+VG1 UG0+VG0 UG0+VG0 ]
        punpckhwd mm6,mm6               ; [ UG3+VG3 UG3+VG3 UG2+VG2 UG2+VG2 ]

                ; ((n<<3)>>16 == >>13

        paddsw  mm0,mm2         ; r=[#r3 #r2 #r1 #r0]   = [UR1+Y3 UR1+Y2 UR0+Y1 UR0+Y0]
        paddsw  mm4,mm3         ; r=[#r7 #r6 #r5 #r4]   = [UR3+Y7 UR3+Y6 UR2+Y5 UR2+Y4]
        paddsw  mm1,mm2         ; b=[#b3 #b2 #b1 #b0]   = [VB1+Y3 VB1+Y2 VB0+Y1 VB0+Y0]
        paddsw  mm5,mm3         ; b=[#b7 #b6 #b5 #b4]   = [VB3+Y7 VB3+Y6 VB2+Y5 VB2+Y4]

        paddsw  mm2,mm7         ; g=[#g3 #g2 #g1 #g0] =[Y3+UG1+VG1 Y2+UG1+VG1 Y1+UG0+VG0 Y0+UG0+VG0]
        paddsw  mm3,mm6         ; g=[#g7 #g6 #g5 #g4] =[Y7+UG3+VG3 Y6+UG3+VG3 Y5+UG2+VG2 Y4+UG2+VG2]

        packuswb mm0,mm4                ;r=[r7 r6 r5 r4 r3 r2 r1 r0]
        packuswb mm1,mm5                ;b=[b7 b6 b5 b4 b3 b2 b1 b0]
        packuswb mm2,mm3                ;g=[g7 g6 g5 g4 g3 g2 g1 g0]
        
                ; Interlace [3]=0 [2]=b [1]=g [0]=r
            
        movq    mm4,mm0         ;r
        movq    mm3,mm2         ;g

        punpcklbw mm4,mm1               ;br   [b3 r3 b2 r2 b1 r1 b0 r0]  r3...0 |= b3...0
        punpckhbw mm0,mm1               ;     [b7 r7 b6 r6 b5 r5 b4 r4]  b7...4 |= r7...0
        punpcklbw mm3,ZeroD     ;0g   [ 0 g3  0 g2  0 g1  0 g0]
        punpckhbw mm2,ZeroD     ;     [ 0 g7  0 g6  0 g5  0 g4] 

        movq    mm5,mm4         ;br   [b3 r3 b2 r2 b1 r1 b0 r0]
        movq    mm1,mm0

        punpcklbw mm4,mm3               ;0bgr [ 0 b1 g1 r1  0 b0 g0 r0]
        punpckhbw mm5,mm3               ;     [ 0 b3 g3 r3  0 b2 g2 r2]
        punpcklbw mm0,mm2               ;0bgr [ 0 b5 g5 r5  0 b4 g4 r4]
        punpckhbw mm1,mm2               ;     [ 0 b7 g7 r7  0 b6 g6 r6]

        movq    [ebx+0],mm4     ;bgr1 bgr0
        movq    [ebx+8],mm5     ;bgr3 bgr2
        movq    [ebx+16],mm0    ;bgr5 bgr4
        movq    [ebx+24],mm1    ;bgr7 bgr6
        add     ebx,32

        dec     ecx
        jnz     Hrz             ; Loop for width of image

                ; Odd/Even

        mov     ecx,nStrideUV   ; UV stride
        mov     edi,pBaseU      ; pU (old)
        mov     edx,pBaseV      ; pV (old)
        and     ecx,eax         ; 0 : stride
        xor     eax,0FFFFFFFFh  ; Flip mask

        add     esi,nStrideY    ; pY   += stride adj.
        add     ebx,iStride     ; pImg += stride adj.

        add     edi,ecx         ; pU += (odd) ? 0 : next
        add     edx,ecx         ; pV +=

        mov     pBaseU,edi      ; pU (Save even=old or odd=new pointer)
        mov     pBaseV,edx      ; pV

        dec     nHeight
        jne     Vrt             ; Loop for height of image

        pop     edi
        pop     esi
        pop     ebx
;;;      femms
        ret
vmp_yuv12toRGB32AsmMMX endp



;_________________________________________________________________________________________________
;
;       *** SSE2 version *** 128bit (un)aligned

        align   16

        .XMM
        public  vmp_yuv12toRGB32AsmSSE2
vmp_yuv12toRGB32AsmSSE2 proc    near,
                pImg:ptr, nWidth:DWORD, nHeight:DWORD, iStride:DWORD,
                pBaseY:ptr,             nStrideY:DWORD, 
                pBaseU:ptr, pBaseV:ptr, nStrideUV:DWORD
        push    ebx
        push    esi
        push    edi

ASSERT_PTR16(pBaseY)
ASSERT_ZERO(nStrideY)
ASSERT_PTR16(pBaseU)
ASSERT_PTR16(pBaseV)
ASSERT_ZERO(nStrideUV)
ASSERT_PTR16(pImg)
ASSERT_ZERO(nWidth)
ASSERT_ZERO(nHeight)
ASSERT_ZERO(iStride)

        mov     eax,nWidth
        shr     nWidth,4                ; nWidth >>= 4;   n/16

        sub     nStrideY,eax    ; nStrideY -= nWidth
        mov     ebx,pImg
        mov     esi,pBaseY
        shl     eax,2           ; x4
        mov     edi,pBaseU
        mov     edx,pBaseV
        sub     iStride,eax     ; iStride -= (nWidth*4)
        mov     eax,0           ; (Flip Flip)

        ; Vertical Loop

Vrt:    mov     ecx,nWidth

        ; Horizontal Loop

Hrz:    movdqu  xmm0,[edi]      ; [uF uE uD uC uB uA u9 u8 u7 u6 u5 u4 u3 u2 u1 u0] *pU
        add     edi,8
        movdqu  xmm1,[edx]      ; [vF vE vD vC vB vA v9 v8 v7 v6 v5 v4 v3 v2 v1 v0] *pV
        add     edx,8

        punpcklbw xmm0,ZeroD    ; [ 0 u7  0 u6  0 u5  0 u4  0 u3  0 u2  0 u1  0 u0] 0...255
        punpcklbw xmm1,ZeroD    ; [ 0 v7  0 v6  0 v5  0 v4  0 v3  0 v2  0 v1  0 v0]

        movdqa  xmm2,[esi]      ; [yF yE yD yC yB yA y9 y8 y7 y6 y5 y4 y3 y2 y1 y0]
        add     esi,16

        psllw   xmm0,3          ; [ U7 U6 U5 U4 U3 U2 U1 U0 ] 0...2040  8*(0...255)
        psllw   xmm1,3          ; [ V7 V6 V5 V4 V3 V2 V1 V0 ]

        movdqa  xmm3,xmm2       ; [yF yE yD yC yB yA y9 y8 y7 y6 y5 y4 y3 y2 y1 y0]

        paddsb  xmm0,DiffUV     ; [ U7 U6 U5 U4 U3 U2 U1 U0 ] (8*u)-(8*128) = 8(u-128)
        paddsb  xmm1,DiffUV     ; [ V7 V6 V5 V4 V3 V2 V1 V0 ]     = {-1024...1016}

        punpcklbw xmm2,ZeroD    ; [ 0 y7  0 y6  0 y5  0 y4  0 y3  0 y2  0 y1  0 y0]
        punpckhbw xmm3,ZeroD    ; [ 0 yF  0 yE  0 yD  0 yC  0 yB  0 yA  0 y9  0 y8]
        
        movdqa  xmm6,xmm0
        pmulhw  xmm0,ColorUR    ; [UR7 UR6 UR5 UR4 UR3 UR2 UR1 UR0]  ur =  0x3343 * 8(u-128)
        pmulhw  xmm6,ColorUG    ; [UG7 UG6 UG5 UG4 UG3 UG2 UG1 UG0]  ug = -0x1a1e * 8(u-128)

        movdqa  xmm4,xmm0       ; [UR7 UR6 UR5 UR4 UR3 UR2 UR1 UR0]
        movdqa  xmm7,xmm1       ; [ V7  V6  V5  V4  V3  V2  V1  V0]

        punpcklwd xmm0,xmm0     ; [UR3 UR3 UR2 UR2 UR1 UR1 UR0 UR0]
        punpckhwd xmm4,xmm4     ; [UR7 UR7 UR6 UR6 UR5 UR5 UR4 UR4]

        psllw   xmm2,3          ; [ Y7  Y6  Y5  Y4  Y3  Y2  Y1  Y0] (8*y)
        psllw   xmm3,3          ; [ YF  YE  YD  YC  YB  YA  Y9  Y8]

        pmulhw  xmm1,ColorVB    ; [VB7 VB6 VB5 VB4 VB3 VB2 VB1 VB0]  vb =  0x40cf * 8(v-128)
        pmulhw  xmm7,ColorVG    ; [VG7 VG6 VG5 VG4 VG3 VG2 VG1 VG0]  vg = -0x0c92 * 8(v-128)

        movdqa  xmm5,xmm1       ; [VB7 VB6 VB5 VB4 VB3 VB2 VB1 VB0]
        paddsw  xmm2,DiffY      ; [ Y7  Y6  Y5  Y4  Y3  Y2  Y1  Y0] (8*y)-(8*16) = 8(y-16)
        paddsw  xmm3,DiffY      ; [ YF  YE  YD  YC  YB  YA  Y9  Y8] 

        paddsw  xmm7,xmm6       ; [UG7+VG7 UG6+VG6 UG5+VG5 UG4+UG4 UG3+VG3 UG2+VG2 UG1+VG1 UG0+VG0]

        punpcklwd xmm1,xmm1     ; [VB3 VB3 VB2 VB2 VB1 VB1 VB0 VB0]
        punpckhwd xmm5,xmm5     ; [VB7 VB7 VB6 VB6 VB5 VB5 VB4 VB4]

        movdqa  xmm6,xmm7       ; [UG7+VG7 UG6+VG6 UG5+VG5 UG4+VG4 UG3+VG3 UG2+VG2 UG1+VG1 UG0+VG0]

        pmulhw  xmm2,ColorY     ; [ Y7  Y6  Y5  Y4  Y3  Y2  Y1  Y0]  y = 0x2568 * 8(y-16)
        pmulhw  xmm3,ColorY     ; [ YF  YE  YD  YC  YB  YA  Y9  Y8]  y = 0x2568 * 8(y-16)

        punpcklwd xmm7,xmm7     ; [ UG3+VG3 UG3+VG3 UG2+VG2 UG2+VG2 UG1+VG1 UG1+VG1 UG0+VG0 UG0+VG0 ]
        punpckhwd xmm6,xmm6     ; [ UG7+VG7 UG7+VG7 UG6+VG6 UG6+VG6 UG5+VG5 UG5+VG5 UG4+VG4 UG4+VG4 ]

                ; ((n<<3)>>16 == >>13

        paddsw  xmm0,xmm2       ; r=[#r7 #r6 #r5 #r4 #r3 #r2 #r1 #r0]   = [UR1+Y3 UR1+Y2 UR0+Y1 UR0+Y0]
                                ;                                         [UR3+Y7 UR3+Y6 UR2+Y5 UR2+Y4]
        paddsw  xmm4,xmm3       ; r=[#rF #rE #rD #rC #rB #rA #r9 #r8]   = [UR5+YB UR5+YA UR4+Y9 UR4+Y8]
                                ;                                         [UR7+YF UR7+YE UR6+YD UR6+YC]
        paddsw  xmm1,xmm2       ; b=[#b7 #b6 #b5 #b4 #b3 #b2 #b1 #b0]   = [VB1+Y3 VB1+Y2 VB0+Y1 VB0+Y0]
                                ;                                         [VB3+Y7 VB3+Y6 VB2+Y5 VB2+Y4]
        paddsw  xmm5,xmm3       ; b=[#bF #bE #bD #bC #bB #bA #b9 #b8]   = [VB5+YB VB5+YA VB4+Y9 VB4+Y8]
                                ;                                         [VB7+YF VB7+YE VB6+YD VB6+YC]

        paddsw  xmm2,xmm7       ; g=[#g7 #g6 #g5 #g4 #g3 #g2 #g1 #g0] =[Y3+UG1+VG1 Y2+UG1+VG1 Y1+UG0+VG0 Y0+UG0+VG0]
                                ;                                      [Y7+UG3+VG3 Y6+UG3+VG3 Y5+UG2+VG2 Y4+UG2+VG2]
        paddsw  xmm3,xmm6       ; g=[#gF #gE #gD #gC #gB #gA #g9 #g8] =[YB+UG5+VG5 YA+UG5+VG5 Y9+UG4+VG4 Y8+UG3+VG3]
                                ;                                      [YF+UG7+VG7 YE+UG7+VG7 YD+UG6+VG6 YC+UG6+VG6]
        packuswb xmm0,xmm4      ;r=[rF rE rD rC rB rA r9 r8 r7 r6 r5 r4 r3 r2 r1 r0]
        packuswb xmm1,xmm5      ;b=[bF bE bD bC bB bA b9 b8 b7 b6 b5 b4 b3 b2 b1 b0]
        packuswb xmm2,xmm3      ;g=[gF gE gD gC gB gA g9 g8 g7 g6 g5 g4 g3 g2 g1 g0]
        
                ; Interlace [3]=0 [2]=b [1]=g [0]=r
            
        movdqa  xmm4,xmm0       ;r
        movdqa  xmm3,xmm2       ;g

        punpcklbw xmm4,xmm1     ;br   [b7 r7 b6 r6 b5 r5 b4 r4 b3 r3 b2 r2 b1 r1 b0 r0]  r3...0 |= b3...0
        punpckhbw xmm0,xmm1     ;     [bF rF bE rE bD rD bC rC bB rB bA rA b9 r9 b8 r8]  b7...4 |= r7...0
        punpcklbw xmm3,ZeroD    ;0g   [ 0 g7  0 g6  0 g5  0 g4  0 g3  0 g2  0 g1  0 g0]
        punpckhbw xmm2,ZeroD    ;     [ 0 gF  0 gE  0 gD  0 gC  0 gB  0 gA  0 g9  0 g8] 

        movdqa  xmm5,xmm4       ;br   [b7 r7 b6 r6 b5 r5 b4 r4 b3 r3 b2 r2 b1 r1 b0 r0]
        movdqa  xmm1,xmm0       ;     [bF rF bE rE bD rD bC rC bB rB bA rA b9 r9 b8 r8]

        punpcklbw xmm4,xmm3     ;0bgr [ 0 b3 g3 r2  0 b2 g2 r2  0 b1 g1 r1  0 b0 g0 r0]
        punpckhbw xmm5,xmm3     ;     [ 0 b7 g7 r7  0 b6 g6 r6  0 b5 g5 r5  0 b4 g4 r4]
        punpcklbw xmm0,xmm2     ;0bgr [ 0 bB gB rB  0 bA gA rA  0 b9 g9 r9  0 b8 g8 r8]
        punpckhbw xmm1,xmm2     ;     [ 0 bF gF rF  0 bE gE rE  0 bD gD rD  0 bC gC rC]

        movdqa  [ebx+0],xmm4    ;bgr3 bgr2 bgr1 bgr0
        movdqa  [ebx+16],xmm5   ;bgr7 bgr6 bgr5 bgr4
        movdqa  [ebx+32],xmm0   ;bgrB bgrA bgr9 bgr8
        movdqa  [ebx+48],xmm1   ;bgrF bgrE bgrD bgrC
        add     ebx,64

        dec     ecx
        jnz     Hrz             ; Loop for width of image

                ; Odd/Even

        mov     ecx,nStrideUV   ; UV stride
        mov     edi,pBaseU      ; pU (old)
        mov     edx,pBaseV      ; pV (old)
        and     ecx,eax         ; 0 : stride
        xor     eax,0FFFFFFFFh  ; Flip mask

        add     esi,nStrideY    ; pY   += stride adj.
        add     ebx,iStride     ; pImg += stride adj.

        add     edi,ecx         ; pU += (odd) ? 0 : next
        add     edx,ecx         ; pV +=

        mov     pBaseU,edi      ; pU (Save even=old or odd=new pointer)
        mov     pBaseV,edx      ; pV

        dec     nHeight
        jne     Vrt             ; Loop for height of image

        pop     edi
        pop     esi
        pop     ebx
        ret
vmp_yuv12toRGB32AsmSSE2 endp



yuv12X86  ends
        end

